/*
 * Copyright (c) 1994,1997 by Sun Microsystems, Inc.
 * All rights reserved.
 */

#ifndef	_PCNET_H
#define	_PCNET_H

#pragma ident	"@(#)pcnet.h	1.10	97/04/04 SMI"

#ifdef	__cplusplus
extern "C" {
#endif

/*
 * Declarations and definitions specific to the Am79C970
 * PCNET Ethernet Local Area Network Controller.
 *
 * Header file dependencies:
 *	sys/types.h
 */

/*
 * The PCNET chip saves address pins by accessing
 * several registers with one address pin by first writing the
 * register address to an internal address register, then reading
 * or writing a data register.  To use this safely, care must be
 * taken that the driver isn't reentered between the writing
 * of the address register and the access to the data register,
 * lest the reentered code try to touch the registers and
 * screw up the sequence.
 *
 * There are a variety of registers accessible with this scheme; we take
 * special note of CSR0, CSR1, CSR2, and CSR3.  In normal operation, only
 * CSR0 can be or needs to be accessed, so the driver normally leaves a 0
 * in the Register Address Port, allowing CSR0 to be accessed simply by
 * accessing the Register Data Port.  During the initialization
 * sequence, when the other CSRs need to be accessed, the appropriate
 * CSR address is written into the address port, and afterwards
 * the 0 is put back in the address port.
 */

/*
 * There are several fundamental choices when defining structures for
 * the am79c970. The basic ones we make here are:
 *    - These datastructures live in little-endian memory (see note below)
 *    - SSIZE32=1; The chip will request memory in 32-bit chunks. This
 *                 redefines a number of initialization structures.
 *    - DWIO mode is ON. This is a concession to i86_subr.s, which
 *                 is unable to perform 16-bit I/O space writes.
 *
 * Little-endian note: These structures are designed assuming little-endian
 * byte ordering. As a result, any memory involving these datastructure must
 * be allocated with ddi_dma_mem_alloc, with argument accflags set to
 * DDI_LE_STRUCTURE_ACC. On pentiums, this has no effect, on sparcs, this
 * causes the IE (Invert Endian) bit to be set in the MMU entry for the page.
 */

#define	LE_RAP_OFFSET 0x14	/* Register Address Port */
#define	LE_RDP_OFFSET 0x10	/* Register Data Port */
struct lanceregs {
	volatile uint32_t	lance_aprom1;	/* APROM register */
	volatile uint32_t	lance_aprom2;	/* APROM register */
	volatile uint32_t	lance_aprom3;	/* APROM register */
	volatile uint32_t	lance_aprom4;	/* APROM register */
	volatile uint32_t	lance_rdp;	/* Register Data Port */
	volatile uint32_t	lance_rap;	/* Register Address Port */
	volatile uint32_t	 lance_reset;    /* Reset register */
	volatile uint32_t	 lance_bdp;	/* Bus cfg data port */
};
#define	lance_csr lance_rdp

#define	LANCE_CSR0		0
#define	LANCE_CSR1		1
#define	LANCE_CSR2		2
#define	LANCE_CSR3		3
#define	LANCE_CSR58		58

/*
 * Control and status bits for CSR0.
 * These behave somewhat strangely, but the net effect is that
 * bit masks may be written to the register which affect only
 * those functions for which there is a one bit in the mask.
 * The exception is the interrupt enable, which must be explicitly
 * set to the correct value in each mask that is used.
 *
 * RO - Read Only, writing has no effect
 * RC - Read, Clear.  Writing 1 clears, writing 0 has no effect
 * RW - Read, Write.
 * W1 - Write with 1 only.  Writing 1 sets, writing 0 has no effect.
 *	Reading gives unpredictable data but doesn't hurt anything.
 * RW1 - Read, Write with 1 only.  Writing 1 sets, writing 0 has no effect.
 */

#define	LANCE_ERR	0x8000		/* RO BABL | CERR | MISS | MERR */
#define	LANCE_BABL	0x4000		/* RC transmitted too many bits */
#define	LANCE_CERR	0x2000		/* RC No Heartbeat */
#define	LANCE_MISS	0x1000		/* RC Missed an incoming packet */
#define	LANCE_MERR	0x0800		/* RC Memory Error; no acknowledge */
#define	LANCE_RINT	0x0400		/* RC Received packet Interrupt */
#define	LANCE_TINT	0x0200		/* RC Transmitted packet Interrupt */
#define	LANCE_IDON	0x0100		/* RC Initialization Done */
#define	LANCE_INTR	0x0080		/* RO BABL|MISS|MERR|RINT|TINT|IDON */
#define	LANCE_INEA	0x0040		/* RW Interrupt Enable */

#define	LANCE_STAT	0xff00		/* RC All status bits */

#define	LANCE_RXON	0x0020		/* RO Receiver On */
#define	LANCE_TXON	0x0010		/* RO Transmitter On */
#define	LANCE_TDMD	0x0008		/* W1 Transmit Demand (send it now) */
#define	LANCE_STOP	0x0004		/* RW1 Stop */
#define	LANCE_STRT	0x0002		/* RW1 Start */
#define	LANCE_INIT	0x0001		/* RW1 Initialize */

/*
 * CSR1 is the low 16 bits of the address of the initialization block
 * CSR2 is the high 16 bits of the address of the initialization block
 *             (in 32-bit mode).
 * CSR3 mode bits: None are used by this driver. All should be zero.
 */
#define	LANCE_BSWP	0x4	/* Byte Swap (off for spitfire and pentium) */

/* CSR58 bits. Alias of BCR20. */
#define	LANCE_CSRPCNET  0x200	/* RO PCNET-ISA cfg bit. Should be on. */
#define	LANCE_SSIZE32   0x100   /* RO Use ILACC (32-bit) mode. should be ON */
#define	LANCE_SWSTYLE   0x0FF	/* RW Mask for style defined below */
#define	LANCE_PCNETPCI  0x02	/* RW set PCNet-PCI mode. */

/* Note: The below should probably be a paddr_t, but paddr_t may become a */
/*	64-bit value in some future release. It's now defined as a u_int32 */
/*	to ensure it stays a 32-bit value, which the hardware requires. */

struct lancering {		/* Descriptor Ring Pointer */
	uint32_t  lr_addr;	/* ring address */
};

/*
 * Initialization Block.  This structure is constructed in memory,
 * and it's address is written into the chip during initialization.
 * The chip then fetches it's initialization info from the structure.
 */
struct lance_init_block {
	/* In the normal mode, these 16 bits are all 0 */
	uint16_t	mode;

	/* The below two values occupy the high four bits in the eight bit */
	/* fields. I find it cleaner to describe them as chars rather than */
	/* bit fields. The low order 4 bits of both fields must be zero.   */

	uint8_t		ib_rlen;	/* size of rcv ring. Upper 4 bits only. */
	uint8_t		ib_tlen;	/* size of xmit ring. Upper 4 bits only. */

	uint8_t		ib_padr[6];	/* physical ethernet address. */
	uint8_t		ib_resv[2];	/* reserve */
	uint16_t	ib_ladrf[4];	/* Multicast logical address filter */

	struct lancering ib_rdrp;	/* Receive Descriptor Ring Pointer */
	struct lancering ib_tdrp;	/* Transmit Descriptor Ring Pointer */
};

struct	lmd {				/* Message Descriptor */
	uint32_t	lmd_addr;	/* Address of data */
	uint16_t	lmd_bcnt;	/* Buffer Byte Count (maximum length) */
	uint8_t		lmd_res1;	/* unused */
	uint8_t		lmd_flags;	/* own through enp */
	uint16_t	lmd_mcnt;	/* Message Byte Count (actual length) */
					/* On transmit: retry count. */
	uint16_t	lmd_rccrpc;	/* runt packet count and rcv */
					/* collision count */
					/* On transmit: tdr counter */
#define	lmd_flags3 lmd_rccrpc		/* On transmit: buff, uflo, exdef, lcol, */
					/* lcar, rtry, and tdr */
	uint32_t	lmd_tmd3;	/* r/tmd3. unused, aligns to 16 bytes. */
};

/* Bits common to both rmds and tmds */
#define	LMD_OWN		0x80	/* Chip owns the descriptor */
#define	LMD_ERR		0x40	/* Error occurred */
#define	LMD_STP		0x02	/* Start of Packet */
#define	LMD_ENP		0x01	/* End of Packet */

/* Bits in rmd flags */
#define	RMD_FRAM	0x20	/* Framing error */
#define	RMD_OFLO	0x10	/* Internal Silo Overflowed. Valid if !ENP */
#define	RMD_CRC		0x08	/* CRC Error */
#define	RMD_BUFF	0x04	/* Didn't have a buffer for the packet */
/* bits in tmd flags */
#define	TMD_RES		0x20	/* Reserved, lance writes this with a zero */
#define	TMD_MORE	0x10	/* More than one retry was needed */
#define	TMD_ONE  	0x08	/* Exactly One Retry, valid only if !LCOL */
#define	TMD_DEF  	0x04	/* Deferred (net was initially busy) */

/* Bits for lmd_errflags */
#define	TMD_BUFF 0x8000		/* Buffer Error (imples underflow too) */
#define	TMD_UFLO 0x4000		/* Underflow Error */
#define	TMD_LCOL 0x1000		/* Late Collision */
#define	TMD_LCAR 0x0800		/* Loss of Carrier */
#define	TMD_RTRY 0x0400		/* More than 16 Retry's */
#define	TMD_TDR	 0x03ff		/* Time Domain Reflectometry counter mask */

/* Handy combo errflags */
#define	TMD_ANYERROR	(TMD_BUFF|TMD_UFLO|TMD_LCOL|TMD_LCAR|TMD_RTRY)

#define	DESCRING_ALIGN	(16)

#ifdef	__cplusplus
}
#endif

#endif	/* _PCNET_H */
