/*
 * Copyright (c) 1989-1997, by Sun Microsytems, Inc.
 * All rights reserved.
 */
#pragma ident	"@(#)bst.c	1.45	97/10/03 SMI"

/*
 * bst.c - Block SCSI Target driver; an example block SCSA target
 *	   driver for Solaris 2.x.
 *
 * This file is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify this file without charge, but are not authorized to
 * license or distribute it to anyone else except as part of a product
 * or program developed by the user.
 *
 * THIS FILE IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 *
 * This file is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 *
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY THIS FILE
 * OR ANY PART THEREOF.
 *
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even
 * if Sun has been advised of the possibility of such damages.
 *
 * Sun Microsystems, Inc.
 * 901 San Antonio Rd
 * Palo Alto, California 94303
 *
 * ---------------------------------------------------------------------------
 *
 * This driver is intended as an example of programming a block SCSA
 * target driver for SVR4; it is not intended for any particular
 * direct access device, although it has been tested on a hard disk drive.
 *
 * This driver looks for a Sun-style disk label (struct dk_label, as defined
 * in <sys/dklabel.h>) on sector zero of the device. The label contains the
 * offsets and lengths of each data partition on the device. The driver
 * creates a minor device node for each partition, and uses the partition
 * info from the label to calculate the physical block number from the
 * logical block requested.
 *
 * This driver also supports the "DKIO" ioctls necessary to format and
 * partition the disk (defined in <sys/dkio.h>).
 *
 * The other major difference between this and the character-only driver
 * (sst) is that this driver supports I/O request queueing.
 *
 * Terminology in this driver:
 *	"instance" - this device, e.g. the disk drive
 *	"partition" - the part of the device ("slice") that's being accessed
 * The instance and partition are coded into the minor device number; the
 * partition is the low three bits (i.e. 8 partitions), and the instance is
 * the rest (i.e. bits 3 - 15).
 *
 * NOTE: Areas where you may need to change this code or add your own to
 *	deal with your specific device are marked "Note".
 * Other warnings are marked "WARNING"
 *
 * To compile:
 *	% cc -D_KERNEL -c bst.c
 *	% ld -r bst.o -o bst
 * WARNING: Only the -Xa and -Xt (the default) will work! The header files
 *	    don't conform to strict ANSI C; -Xs and -Xc will fail.
 *
 * To install:
 * 1. Copy the module (bst) and config file (bst.conf) into /kernel/drv
 * 2. Run add_drv(1M).
 *
 * Note: you only need to run add_drv the very first time you introduce
 *	 the driver to the system. Thereafter you can use modunload to
 *	 unload the driver and just reference it (e.g. open(2) it) to
 *	reload.
 *
 * Setting variables
 *	Variables can be explicitly set from the /etc/system file, by
 *	adding an entry of the form
 *		"set bst:<variable name>=<value>"
 *	Alternatively, you can use adb to set variables and debug as
 *	follows:
 *		# adb -kw /dev/ksyms /dev/mem
 *	The /etc/system file is read only once at boot time, if you change
 *	it you must reboot for the change to take effect.
 */


/*
 * Includes, Declarations and Local Data
 */
#include <sys/note.h>
#include <sys/scsi/scsi.h>
#include <sys/dkio.h>
#include <sys/vtoc.h>
#include <sys/stat.h>
#include <sys/modctl.h>
#include <sys/file.h>

#include <sys/ddi.h>
#include <sys/sunddi.h>

#include "fdisk.h"
#include "bst_def.h"

/*
 * Local Function Prototypes
 */
static int bst_open(dev_t *dev_p, int flag, int otyp, cred_t *cred_p);
static int bst_close(dev_t dev, int flag, int otyp, cred_t *cred_p);
static int bst_strategy(struct buf *bp);
static int bst_ioctl(dev_t, int, intptr_t, int, cred_t *, int *);
static int bst_read(dev_t dev, struct uio *uio, cred_t *cred_p);
static int bst_write(dev_t dev, struct uio *uio, cred_t *cred_p);
static int bst_aread(dev_t dev, struct aio_req *aio, cred_t *cred_p);
static int bst_awrite(dev_t dev, struct aio_req *aio, cred_t *cred_p);
static int bst_dump(dev_t dev, caddr_t addr, daddr_t blkno, int nblk);

static int bst_rw(dev_t dev, struct uio *uio, int rw);
static int bst_ioctl_cmd(dev_t, struct uscsi_cmd *, enum uio_seg, enum uio_seg);
static void bst_checklabel(struct bst_soft *tgt);
static void bst_uselabel(struct bst_soft *tgt, struct dk_label *l);


static void bst_dkioc_gvtoc(struct bst_soft *tgt, struct vtoc *vtoc);
static int bst_dkioc_svtoc(struct vtoc *vtoc, struct bst_soft *tgt);
static int bst_write_label(dev_t dev);

static int bst_read_fdisk(struct bst_soft *tgt, int (*f)());
static int bst_iopb_read_block(struct bst_soft *tgt, daddr_t blkno,
	struct buf *bp, int (*f)());
static int bst_iopb_write_block(struct bst_soft *tgt, daddr_t blkno,
	struct buf *bp, int (*f)());
static void bst_clear_cont_alleg(struct bst_soft *tgt, int (*f)());
static int bst_scsi_poll(struct bst_soft *tgt, struct scsi_pkt *pkt);

static void bst_getmap(struct bst_soft *tgt, struct dk_map *map);
static void bst_chgmap(struct bst_soft *tgt, struct dk_map *map);

static int bst_ready(struct bst_soft *tgt, dev_t dev);
static void bst_read_capacity(struct bst_soft *tgt, dev_t dev);
static int bst_simple(dev_t dev, int cmd, int flag);
static void bst_start(struct bst_soft *tgt);
static int bst_runout(caddr_t arg);
static void bst_done(struct bst_soft *tgt, struct buf *bp);
static int bst_next_xfer(struct bst_soft *tgt, struct buf *bp,
	struct scsi_pkt *pkt, struct bst_pkt_private *prv);
static void bst_make_cmd(struct bst_soft *tgt, struct buf *bp, int (*f)());
static void bst_restart(caddr_t arg);
static void bst_callback(struct scsi_pkt *pkt);
static int bst_handle_incomplete(struct bst_soft *tgt, struct buf *bp);
static int bst_handle_arq(struct scsi_pkt *pktp, struct bst_soft *tgt,
    struct buf *bp);
static int bst_handle_sense(struct bst_soft *tgt, struct buf *bp);
static int bst_check_error(struct bst_soft *tgt, struct buf *bp);
static void bst_log(struct scsi_device *devp, int level, const char *fmt, ...);
static void bst_hex_print(char *msg, char *cp, int len);
static void bst_inq_fill(char *p, int l, char *s);
static void bst_dump_cdb(struct bst_soft *tgt, struct scsi_pkt *pkt,
	int cdblen);
static int bst_prop_op(dev_t, dev_info_t *, ddi_prop_op_t, int,
    char *, caddr_t, int *);
static int bst_power(dev_info_t *, int, int);

#ifdef	DEBUG
static void dump(char *hdr, u_char *buf, int nbytes);
#endif

/*
 * Local Static Data
 */
static int	bst_io_time		= BST_IO_TIME;
static int	bst_retry_count		= BST_RETRY_COUNT;
static void	*bst_state;

static char	*bst_label		= "bst";

/*
 * Array of commands supported by the device, suitable for scsi_errmsg(9f)
 * Note: Add or remove commands here as appropriate for your device.
 */
static struct scsi_key_strings bst_cmds[] = {
	0x00, "test unit ready",
	0x01, "rezero",
	0x03, "request sense",
	0x04, "format",
	0x05, "read block limits",
	0x07, "reassign",
	0x08, "read",
	0x0a, "write",
	0x0b, "seek",
	0x0f, "read reverse",
	0x12, "inquiry",
	0x13, "verify",
	0x14, "recover buffered data",
	0x15, "mode select",
	0x16, "reserve",
	0x17, "release",
	0x18, "copy",
	0x1a, "mode sense",
	0x1b, "start/stop/load",
	0x1e, "door lock",
	0x28, "write (10)",
	0x2a, "read (10)",
	0x37, "read defect data",
	-1, NULL,
};

/*
 * Errors at or above this level will be reported
 */
static int bst_error_reporting = SCSI_ERR_RETRYABLE;



/*
 * Debug message control
 * Debug Levels:
 *	0 = no messages
 *	1 = Errors
 *	2 = Subroutine calls & control flow
 *	3 = I/O Data (verbose!)
 * Can be set with adb or in the /etc/system file with
 * "set bst:bst_debug=<value>"
 * Defining DEBUG on the compile line (-DDEBUG) will enable debugging
 * statements in this driver, and will also enable the ASSERT statements.
 */
#ifdef DEBUG
static int	bst_debug	= 0;
static int	bst_debug_cdb	= 0;
#else
static int	bst_debug	= 0;
static int	bst_debug_cdb	= 0;
#endif /* DEBUG */

#define	BST_DUMP_CDB(tgt, pkt, cdblen)				\
	if (bst_debug_cdb) {					\
		bst_dump_cdb((tgt), (pkt), (cdblen));		\
	}

/*
 * -----------------------------------------------------------------------------
 *	Module Loading/Unloading and Autoconfiguration Routines
 */

/*
 * Device driver ops vector - cb_ops(9s)
 * Device switch table fields (equivalent to the old 4.x cdevsw and bdevsw).
 * Unsupported entry points (e.g. for xxprint() and xxdump()) are set to
 * nodev, except for the poll routine, which is set to nochpoll(), a routine
 * that returns ENXIO.
 * XXX The xxdump entry point should be supported if the device is to be used
 * to dump the system statefile for SUSPEND/RESUME.
 */
static struct cb_ops bst_cb_ops = {
	bst_open,		/* b/c open */
	bst_close,		/* b/c close */
	bst_strategy,		/* b strategy */
	nodev,			/* b print */
	bst_dump,			/* b dump */
	bst_read,		/* c read */
	bst_write,		/* c write */
	bst_ioctl,		/* c ioctl */
	nodev,			/* c devmap */
	nodev,			/* c mmap */
	nodev,			/* c segmap */
	nochpoll,		/* c poll */
	bst_prop_op,		/* cb_prop_op */
	NULL,			/* streamtab  */
	D_MP | D_NEW | D_HOTPLUG,	/* Driver compatibility flag */
	CB_REV,			/* cb_ops revision */
	bst_aread,		/* c aread */
	bst_awrite		/* c awrite */
};


/*
 * dev_ops(9S) structure
 * Device Operations table, for autoconfiguration
 */
static int bst_info(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
    void **result);
static int bst_probe(dev_info_t *devi);
static int bst_attach(dev_info_t *devi, ddi_attach_cmd_t cmd);
static int bst_detach(dev_info_t *devi, ddi_detach_cmd_t cmd);
static int bst_doattach(dev_info_t *devi);
static int bst_dodetach(dev_info_t *devi);

static struct dev_ops bst_ops = {
	DEVO_REV,		/* devo_rev, */
	0,			/* refcnt  */
	bst_info,		/* info */
	nulldev,		/* identify */
	bst_probe,		/* probe */
	bst_attach,		/* attach */
	bst_detach,		/* detach */
	nodev,			/* reset */
	&bst_cb_ops,		/* driver operations */
	NULL,			/* bus operations */
	bst_power		/* power */
};


static struct modldrv modldrv = {
	&mod_driverops,			/* Type of module (driver). */
	"Sample Block SCSI Target Driver 1.45",	/* Description */
	&bst_ops,			/* driver ops */
};

static struct modlinkage modlinkage = {
	MODREV_1, &modldrv, NULL
};

/*
 * Tell the system that we depend on the general scsi support routines,
 * i.e the scsi "misc" module must be loaded
 */

char _depends_on[] = "misc/scsi";


/*
 * -----------------------------------------------------------------------------
 *	Module Loading/Installation Routines
 */

/*
 * Module Installation
 * Install the driver and initialize private data structure allocation
 */
int
_init(void)
{
	int e;

	if ((e = ddi_soft_state_init(&bst_state, sizeof (struct bst_soft),
								0)) != 0) {
		BST_LOG(0, BST_CE_DEBUG2,
			"_init ddi_soft_state_init failed: 0x%x\n", e);
		return (e);
	}

	if ((e = mod_install(&modlinkage)) != 0) {
		BST_LOG(0, BST_CE_DEBUG2,
				"_init mod_install failed: 0x%x\n", e);
		ddi_soft_state_fini(&bst_state);
	}

	BST_LOG(0, BST_CE_DEBUG2, "_init succeeded\n");
	return (e);
}


/*
 * Module removal
 */
int
_fini(void)
{
	int e;

	if ((e = mod_remove(&modlinkage)) != 0) {
		BST_LOG(0, BST_CE_DEBUG1, "_fini mod_remove failed, 0x%x\n", e);
		return (e);
	}

	ddi_soft_state_fini(&bst_state);

	BST_LOG(0, BST_CE_DEBUG2, "_fini succeeded\n");
	return (e);
}

/*
 * Return module info
 */
int
_info(struct modinfo *modinfop)
{
	BST_LOG(0, BST_CE_DEBUG2, "_info\n");
	return (mod_info(&modlinkage, modinfop));
}


/*
 * -----------------------------------------------------------------------------
 *	Autoconfiguration Routines
 */

/*
 * probe(9e) - Check that we're talking to the right device on the SCSI bus.
 * Calls scsi_probe() to see if there's a device at our target id. If
 * there is, scsi_probe will fill in the sd_inq struct (in the devp
 * scsi device struct) with the inquiry data. Validate the data here,
 * allocate a request sense packet, and start filling in the private data
 * structure.
 * Note: Probe should have no side-effects. Just check that we have the right
 *	 device, don't set up any data or initialize the device here. Also,
 *	 it's a Sun convention to probe quietly; send messages to the log
 *	 file only, not to the console.  Finally, call scsi_unprobe to undo
 *	 the effects of scsi_probe.
 * Note: The host adapter driver sets up the scsi_device structure and puts
 *	 it into the dev_info structure with ddi_set_driver_private().
 * Note: No need to allow for probing/attaching in the open() routine because
 *	 of the loadability - the first reference to the device will auto-load
 *	 it, i.e. will call this routine.
 */
#define	VIDSZ		8	/* Vendor Id length in Inquiry Data */
#define	PIDSZ		16	/* Product Id length in Inquiry Data */

static int
bst_probe(dev_info_t *dip)
{
	struct scsi_device	*devp;
	int			rval = DDI_PROBE_FAILURE;
	int			err;
	char			vpid[VIDSZ+PIDSZ+1];

	devp = (struct scsi_device *)ddi_get_driver_private(dip);
	BST_LOG(devp, BST_CE_DEBUG2, "bst_probe: <%d,%d>\n",
	    devp->sd_address.a_target, devp->sd_address.a_lun);

	/*
	 * scsi_probe(9f) uses the SCSI Inquiry command to  test for the
	 * presence of the device. If it's successful, it will fill in
	 * the sd_inq field in the scsi_device structure.
	 */
	switch (err = scsi_probe(devp, SLEEP_FUNC)) {
	case SCSIPROBE_NORESP:
		BST_LOG(devp, BST_CE_DEBUG1,
		    "%s%d: <%d,%d> no response\n",
		    ddi_get_name(devp->sd_dev),
		    ddi_get_instance(devp->sd_dev),
		    devp->sd_address.a_target, devp->sd_address.a_lun);
		rval = DDI_PROBE_FAILURE;
		break;

	case SCSIPROBE_NONCCS:
	case SCSIPROBE_NOMEM:
	case SCSIPROBE_FAILURE:
	default:
		BST_LOG(devp, BST_CE_DEBUG1,
		    "%s%d: <%d,%d> probe failed 0x%x\n",
		    ddi_get_name(devp->sd_dev),
		    ddi_get_instance(devp->sd_dev),
		    devp->sd_address.a_target, devp->sd_address.a_lun,
			err);
		rval = DDI_PROBE_FAILURE;
		break;

	case SCSIPROBE_EXISTS:
		/*
		 * Inquiry succeeded, devp->sd_inq is now filled in
		 * Note: Check inq_dtype, inq_vid, inq_pid and any other
		 *	 fields to make sure the target/unit is what's
		 *	 expected (sd_inq is a struct scsi_inquiry,
		 *	 defined in scsi/generic/inquiry.h).
		 * Note: Put device-specific checking into the appropriate
		 *	 case statement, and delete the rest.
		 * Note: The DTYPE_* defines are from <scsi/generic/inquiry.h>,
		 *	 this is the full list as of "now", check it for new
		 *	 types.
		 */
		switch (devp->sd_inq->inq_dtype) {
		case DTYPE_OPTICAL:
		case DTYPE_DIRECT:
		case DTYPE_WORM:
		case DTYPE_RODIRECT:
			/*
			 * Print what was found on the console.
			 * Send the 'found' message to the system
			 * log file only, by inserting an exclamation
			 * point, "!", as the	first character of the
			 * message - see cmn_err(9f).
			 */
			bcopy((caddr_t)devp->sd_inq->inq_vid,
			    (caddr_t)vpid, VIDSZ);
			bcopy((caddr_t)devp->sd_inq->inq_pid,
			    (caddr_t)vpid+VIDSZ, PIDSZ);
			vpid[VIDSZ+PIDSZ] = 0;
			cmn_err(CE_CONT, "!%s%d: <%d,%d> %s %s\n",
			    ddi_get_name(devp->sd_dev),
			    ddi_get_instance(devp->sd_dev),
			    devp->sd_address.a_target,
			    devp->sd_address.a_lun,
			    scsi_dname((int)devp->sd_inq->inq_dtype),
			    vpid);
#ifdef	DEBUG
			cmn_err(CE_CONT, "%s%d: <%d,%d> %s %s\n",
			    ddi_get_name(devp->sd_dev),
			    ddi_get_instance(devp->sd_dev),
			    devp->sd_address.a_target,
			    devp->sd_address.a_lun,
			    scsi_dname((int)devp->sd_inq->inq_dtype),
			    vpid);
#endif
			rval = DDI_PROBE_SUCCESS;
			break;

		case DTYPE_NOTPRESENT:
#ifdef	DEBUG
			cmn_err(CE_CONT, "%s%d: <%d,%d> no device present\n",
			    ddi_get_name(devp->sd_dev),
			    ddi_get_instance(devp->sd_dev),
			    devp->sd_address.a_target,
			    devp->sd_address.a_lun);
#endif
			rval = DDI_PROBE_FAILURE;
			break;

		default:
#ifdef	DEBUG
			cmn_err(CE_CONT, "%s%d: <%d,%d> %s not supported\n",
			    ddi_get_name(devp->sd_dev),
			    ddi_get_instance(devp->sd_dev),
			    devp->sd_address.a_target,
			    devp->sd_address.a_lun,
			    scsi_dname((int)devp->sd_inq->inq_dtype));
#endif
			rval = DDI_PROBE_FAILURE;
			break;
		}
	}
	scsi_unprobe(devp);
	return (rval);
}


/*
 * Attach
 *	- create minor device nodes for all 8 partitions
 *	- initialize per-unit mutex's & conditional variables
 *	- if it's a fixed-disk device, read & validate the label now; if
 *	  it's removable, wait until open().
 */
static int
bst_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
	int instance;
	struct bst_soft	*tgt;

	instance = ddi_get_instance(dip);
	tgt = ddi_get_soft_state(bst_state, instance);

	switch (cmd) {
	case DDI_ATTACH:
		if (bst_doattach(dip) == DDI_FAILURE)
			return (DDI_FAILURE);
		return (DDI_SUCCESS);

	case DDI_PM_RESUME:
		/*
		 * There is no hardware state to restart and no
		 * timeouts to restart since we didn't PM_SUSPEND
		 * with active cmds or active timeouts
		 * We just need to unblock waiting threads and restart
		 * I/O
		 *
		 * the code for DDI_RESUME is almost identical except
		 * it uses the suspend flag rather than pm_suspend flag
		 */
		if (tgt == NULL)
			return (DDI_FAILURE);

		mutex_enter(BST_MUTEX(tgt));

		/* allow new operations and unblock any waiting threads */
		tgt->pm_suspended = 0;
		cv_broadcast(&tgt->suspend_cv);

		mutex_exit(BST_MUTEX(tgt));

		/* restart I/O, there should not be an active cmd */
		ASSERT(tgt->iorq.b_forw == NULL);

		bst_start(tgt);

		return (DDI_SUCCESS);

	case DDI_RESUME:
		if (tgt == NULL)
			return (DDI_FAILURE);

		mutex_enter(BST_MUTEX(tgt));

		/* unblock waiting threads */
		tgt->suspended = 0;
		cv_broadcast(&tgt->suspend_cv);

		/* are we still pm_suspended? */
		if (tgt->pm_suspended) {
			mutex_exit(BST_MUTEX(tgt));
			return (DDI_SUCCESS);
		}
		mutex_exit(BST_MUTEX(tgt));

		/* restart I/O, there should not be an active cmd */
		ASSERT(tgt->iorq.b_forw == NULL);

		bst_start(tgt);


		return (DDI_SUCCESS);

	default:
		return (DDI_FAILURE);
	}

}


static int
bst_doattach(dev_info_t *dip)
{
	int			instance, slice, minor;
	struct	scsi_pkt	*rqpkt;
	struct bst_soft		*tgt;
	struct scsi_device	*devp;
	struct buf		*bp;
	char			name[8];


	/*
	 * Find which instance we are, and create a data structure for
	 * the per-instance data. Cross-link the scsi_device struct (devp)
	 * with the per-instance structure (tgt).
	 */
	devp = (struct scsi_device *)ddi_get_driver_private(dip);
	instance = ddi_get_instance(dip);
	BST_LOG(devp, BST_CE_DEBUG2, "bst_attach: <%d,%d>\n",
		devp->sd_address.a_target, devp->sd_address.a_lun);

	/*
	 * Re-probe the device to get the Inquiry data; it's used
	 * elsewhere in the driver. The Inquiry data was validated in
	 * bst_probe so there's no need to look at it again here.
	 */
	if (scsi_probe(devp, SLEEP_FUNC) != SCSIPROBE_EXISTS) {
		BST_LOG(0, BST_CE_DEBUG1,
		    "bst_attach: <%d,%d> probe failed\n",
		    devp->sd_address.a_target, devp->sd_address.a_lun);
		scsi_unprobe(devp);
		return (DDI_FAILURE);
	}

	if (ddi_soft_state_zalloc(bst_state, instance) != DDI_SUCCESS) {
		scsi_unprobe(devp);
		return (DDI_FAILURE);
	}
	tgt = ddi_get_soft_state(bst_state, instance);
	devp->sd_private = (opaque_t)tgt;
	tgt->devp = devp;

	/*
	 * Allocate a buffer for 'special' (driver internal) commands
	 */
	tgt->sbufp = getrbuf(KM_SLEEP);
	if (tgt->sbufp == NULL) {
		goto error;
	}

	/*
	 * Set auto-rqsense, per-target; record whether it's allowed
	 * in arq
	 */
	tgt->arq = ((scsi_ifsetcap(ROUTE(tgt),
	    "auto-rqsense", 1, 1) == 1) ? 1 : 0);

	BST_LOG(devp, BST_CE_DEBUG2, "bst_attach: Auto Sensing %s\n",
	    tgt->arq ? "enabled" : "disabled");

	if (!tgt->arq) {
		/*
		 * No auto sensing, have to do it ourselves. Set up a packet
		 * for it now so we don't have to possibly wait for resources
		 * for it when a command fails.
		 */
		bp = scsi_alloc_consistent_buf(&devp->sd_address,
		    NULL, SENSE_LENGTH, B_READ, SLEEP_FUNC, NULL);
		if (!bp) {
			goto error;
		}

		rqpkt = scsi_init_pkt(&devp->sd_address, NULL, bp,
		    CDB_GROUP0, 1, sizeof (struct bst_pkt_private),
		    PKT_CONSISTENT, SLEEP_FUNC, NULL);
		if (!rqpkt) {
			goto error;
		}
		devp->sd_sense = (struct scsi_extended_sense *)bp->b_un.b_addr;

		(void) scsi_setup_cdb((union scsi_cdb *)rqpkt->pkt_cdbp,
			SCMD_REQUEST_SENSE,
			0, SENSE_LENGTH, 0);

		rqpkt->pkt_comp = bst_callback;
		rqpkt->pkt_time = bst_io_time;
		rqpkt->pkt_flags = FLAG_SENSING;
		tgt->rqbp = bp;
		tgt->rqs = rqpkt;
	}

	/*
	 * Create the minor nodes; one per logical partition. See the man
	 * page for ddi_create_minor_node(9f).
	 * The 2nd parameter is the minor node name; drvconfig(1M) appends
	 * it to the /devices entry, after the colon.
	 * The 4th parameter ('instance') is the actual minor number, put
	 * into the /devices entry's inode and passed to the driver.
	 * The 5th parameter ("DDI_NT_BLOCK_CHAN") is the node type; it's
	 * used by disks(1M) to create the links from /dev to /devices.
	 */
	for (slice = 0; slice < BST_NMAP; slice++) {
		minor = BST_MINOR(instance, slice);
		(void) sprintf(name, "%c", 'a' + slice);
		if (ddi_create_minor_node(dip, name, S_IFBLK,
			    minor, DDI_NT_BLOCK_CHAN, 0) == DDI_FAILURE) {
			BST_LOG(devp, BST_CE_DEBUG1,
			    "Create block minor failed\n");
			goto error;
		}

		(void) sprintf(name, "%c,raw", 'a' + slice);
		if (ddi_create_minor_node(dip, name, S_IFCHR,
			    minor, DDI_NT_BLOCK_CHAN, 0) == DDI_FAILURE) {
			BST_LOG(devp, BST_CE_DEBUG1,
			    "Create raw minor failed\n");
			goto error;
		}
	}

	/*
	 * Initialize power management bookkeeping.
	 */
	if (pm_create_components(dip, 1) == DDI_SUCCESS) {
		if (pm_idle_component(dip, 0) == DDI_FAILURE) {
			BST_LOG(devp, BST_CE_DEBUG1,
				"Idle compnent 0 failed\n");
			goto error;
		}
		pm_set_normal_power(dip, 0, 1);
		tgt->power_level = 1;
	} else {
		BST_LOG(devp, BST_CE_DEBUG1, "Create components failed\n");
		goto error;
	}

	/*
	 * Since this driver manages devices with "remote"
	 * hardware, i.e. the devices themselves have no
	 * "reg" property, the SUSPEND/RESUME commands in
	 * detach/attach will not be called by the power
	 * management framework unless we request it by creating
	 * a "pm-hardware-state" property and setting it
	 * to value "needs-suspend-resume".
	 */
	if (ddi_prop_update_string(DDI_DEV_T_NONE, dip,
	    "pm-hardware-state", "needs-suspend-resume") !=
			DDI_PROP_SUCCESS) {
		BST_LOG(devp, BST_CE_DEBUG1,
		    "ddi_prop_update(\"pm-hardware-state\")failed\n");
		pm_destroy_components(dip);
		goto error;
	}

	/*
	 * Initialize the conditional variables.
	 * Note: We don't need to initialize the driver instance mutex
	 *	 because it's actually devp->sd_mutex (in the struct
	 *	 scsi_device) and is initialized by our parent, the
	 *	 host adapter driver.
	 */
	cv_init(&tgt->sbuf_cv, "targ_cv", CV_DRIVER, NULL);
	cv_init(&tgt->disk_busy_cv, "disk_busy_cv", CV_DRIVER, NULL);
	cv_init(&tgt->suspend_cv, "suspend_cv", CV_DRIVER, NULL);

	/*
	 * Initialize the open/close semaphore. This allows open() and
	 * close() calls to be serialized (i.e. a current open/close
	 * thread will complete before another is allowed).
	 */
	sema_init(&tgt->semaphore, 1, "bst_sem", SEMA_DRIVER, NULL);

	ddi_report_dev(dip);

	BST_LOG(devp, BST_CE_DEBUG2, "Attached bst driver\n");

	return (DDI_SUCCESS);

error:
	if (tgt->rqbp) {
		scsi_free_consistent_buf(tgt->rqbp);
	}
	if (tgt->rqs) {
		scsi_destroy_pkt(tgt->rqs);
	}
	if (tgt->sbufp) {
		freerbuf(tgt->sbufp);
	}
	ddi_remove_minor_node(dip, NULL);	/* remove all minor nodes */
	ddi_soft_state_free(bst_state, instance);
	devp->sd_private = (opaque_t)NULL;
	devp->sd_sense = (struct scsi_extended_sense *)NULL;
	scsi_unprobe(devp);
	return (DDI_FAILURE);
}


/*
 * Detach
 * Free resources allocated in bst_attach
 * Note: If you implement a timeout routine in this driver, cancel it
 *	 here. Note that scsi_init_pkt is called with SLEEP_FUNC, so it
 *	 will wait for resources if non are available. Changing it to
 *	 a callback constitutes a timeout
 */
static int
bst_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	int instance;
	struct bst_soft	*tgt;
	struct diskhd *dp;

	instance = ddi_get_instance(dip);
	tgt = ddi_get_soft_state(bst_state, instance);
	dp = &tgt->iorq;

	switch (cmd) {
	case DDI_DETACH:
		return (bst_dodetach(dip));

	case DDI_SUSPEND:
		/*
		 * to process DDI_SUSPEND, we must do the following:
		 *
		 * - wait till outstanding operations completed and
		 *   block new operations
		 * - cancel pending timeouts
		 *
		 * In this driver there is no hardware state to save.
		 *
		 * If a callback is outstanding which cannot be cancelled
		 * then either wait for the callback to complete or fail
		 * the suspend
		 *
		 * Note that there are two flags, suspend and pm_suspend
		 * for suspend/resume and power management, respectively
		 */
		mutex_enter(BST_MUTEX(tgt));

		/* this shouldn't happen */
		if (tgt->suspended) {
			mutex_exit(BST_MUTEX(tgt));
			return (DDI_SUCCESS);
		}

		/* stop new operations */
		tgt->suspended = 1;

		/*
		 * is the disk already suspended thru DDI_PM_SUSPEND?
		 *
		 * A successful DDI_PM_SUSPEND ensures that the device
		 * isn't active and no timeouts are outstanding
		 */
		if (tgt->pm_suspended) {
			ASSERT(tgt->timeout_id == 0);
			mutex_exit(BST_MUTEX(tgt));
			return (DDI_SUCCESS);
		}

		/*
		 * wait till current operation completed. If we are
		 * in the resource wait state (with a callback outstanding)
		 * then we need to wait till the callback completes and
		 * starts the next cmd. Note that if the request
		 * were never to be satisfied, we would hang forever
		 */
		while (dp->b_forw || (tgt->state == BST_STATE_RWAIT)) {
			cv_wait(&tgt->disk_busy_cv, BST_MUTEX(tgt));
		}

		/* cancel outstanding timeouts, there shouldn't be any */
		if (tgt->timeout_id) {
			int tmp_id = tgt->timeout_id;
			tgt->timeout_id = 0;
			mutex_exit(BST_MUTEX(tgt));
			(void) untimeout(tmp_id);
		} else {
			mutex_exit(BST_MUTEX(tgt));
		}

		return (DDI_SUCCESS);

	case DDI_PM_SUSPEND:
		ASSERT(tgt->suspended == 0);

		mutex_enter(BST_MUTEX(tgt));

		/*
		 * if the device is busy then we fail the PM_SUSPEND.
		 *
		 * the only reason that timeouts might be outstanding
		 * is that an error recovery is in progress so the
		 * device is really still busy
		 */
		if (dp->b_forw || dp->av_forw ||
		    (tgt->state == BST_STATE_RWAIT)) {
			mutex_exit(BST_MUTEX(tgt));
			return (DDI_FAILURE);
		}
		tgt->pm_suspended = 1;
		mutex_exit(BST_MUTEX(tgt));
		return (DDI_SUCCESS);

	default:
		return (DDI_FAILURE);
	}
}

static int
bst_dodetach(dev_info_t *dip)
{
	int			instance;
	struct scsi_device	*devp;
	struct bst_soft		*tgt;

	devp = (struct scsi_device *)ddi_get_driver_private(dip);
	instance = ddi_get_instance(dip);
	BST_LOG(devp, BST_CE_DEBUG2, "bst_detach: unit %d\n", instance);

	if ((tgt = ddi_get_soft_state(bst_state, instance)) == NULL) {
		BST_LOG(devp, CE_WARN, "No Target Struct for bst%d\n",
		    instance);
		return (DDI_FAILURE);
	}

	mutex_enter(BST_MUTEX(tgt));
	if ((tgt->state == BST_STATE_RWAIT) || tgt->timeout_id) {
		/* a callback/timeout is outstanding, we cannot detach */
		mutex_exit(BST_MUTEX(tgt));
		return (DDI_FAILURE);
	}
	mutex_exit(BST_MUTEX(tgt));

	/*
	 * Remove other data structures allocated in bst_attach()
	 */
	cv_destroy(&tgt->sbuf_cv);
	cv_destroy(&tgt->disk_busy_cv);
	cv_destroy(&tgt->suspend_cv);

	if (tgt->rqbp) {
		scsi_free_consistent_buf(tgt->rqbp);
	}
	if (tgt->rqs) {
		scsi_destroy_pkt(tgt->rqs);
	}
	if (tgt->sbufp)
		freerbuf(tgt->sbufp);
	sema_destroy(&tgt->semaphore);
	ddi_soft_state_free(bst_state, instance);
	pm_destroy_components(dip);

	devp->sd_private = (opaque_t)NULL;
	devp->sd_sense = (struct scsi_extended_sense *)NULL;
	ddi_remove_minor_node(dip, NULL);
	scsi_unprobe(devp);
	return (DDI_SUCCESS);
}

/*
 * Property operations.
 * Return the number of blocks for the partition
 * in question or forward the request to the propery facilities.
 */
static int
bst_prop_op(
	dev_t		dev,
	dev_info_t	*dip,
	ddi_prop_op_t	prop_op,
	int		mod_flags,
	char		*name,
	caddr_t		valuep,
	int		*lengthp)
{
	struct bst_soft	*tgt;
	int		nblocks;
	caddr_t		buffer;

	if ((dev != DDI_DEV_T_ANY) && (strcmp(name, "nblocks") == 0)) {

		tgt = ddi_get_soft_state(bst_state, ddi_get_instance(dip));
		if (tgt == NULL || tgt->geom_ok == 0) {
			return (DDI_PROP_NOT_FOUND);
		}

		/*
		* Requests for the property length are easy.
		*/
		if (prop_op == PROP_LEN) {
			*lengthp = sizeof (int);
			return (DDI_PROP_SUCCESS);
		}

		mutex_enter(BST_MUTEX(tgt));
		nblocks = (int)tgt->map[BST_SLICE(dev)].p_size;
		mutex_exit(BST_MUTEX(tgt));

		switch (prop_op)  {
		case PROP_LEN_AND_VAL_ALLOC:
			/*
			 * Allocate buffer for caller
			 */
			buffer = kmem_alloc((size_t)sizeof (int),
				((mod_flags & DDI_PROP_CANSLEEP) ?
				KM_SLEEP : KM_NOSLEEP));
			if (buffer == NULL)  {
				scsi_log(DEVINFO(tgt), bst_label, CE_WARN,
				    "no memory for 'nblocks' "
				    "property allocation\n");
				return (DDI_PROP_NO_MEMORY);
			}
			*(caddr_t *)valuep = buffer; /* Set callers buf ptr */
			break;

		case PROP_LEN_AND_VAL_BUF:
			/*
			 * Use caller's buffer
			 */
			if (sizeof (int) > (*lengthp))
				return (DDI_PROP_BUF_TOO_SMALL);
			buffer = valuep; /* get callers buf ptr */
			break;
		}
		*lengthp = sizeof (int);
		*((int *)buffer) = nblocks;
		return (DDI_PROP_SUCCESS);
	}

	/*
	* Property request not mine - pass it on.
	*/
	return (ddi_prop_op(dev, dip, prop_op, mod_flags,
		name, valuep, lengthp));
}

/*
 * -----------------------------------------------------------------------------
 *	Unix Entry Points
 *
 * bst_dump is used to dump a port of virtual address space
 * directly to a device in the case of system failure. It is also used
 * to copy the state of the kernel out to disk during a checkpoint
 * operation (see cpr(7)). This function should not use strategy
 * to perform the write since interrupts are disabled during dumping
 */
static int
bst_dump(dev_t dev, caddr_t addr, daddr_t blkno, int nblk)
{
	struct bst_soft	*tgt;
	struct buf	*bp;
	struct partition *dkp;
	int		rval;
	struct scsi_pkt	*pkt;

	if ((tgt = ddi_get_soft_state(bst_state, BST_INST(dev))) == NULL) {
		BST_LOG(0, BST_CE_DEBUG1, "dump - unattached instance %d\n",
		    BST_INST(dev));
		return (ENXIO);
	}
	_NOTE(NO_COMPETING_THREADS_NOW);

	BST_LOG(tgt->devp, BST_CE_DEBUG1,
		"bst_dump: addr = %x, blkno = %x nblk = %x\n",
		addr, blkno, nblk);

	if (tgt->geom_ok == 0) {
		BST_LOG(tgt->devp, BST_CE_DEBUG1, "dump - invalid geometry\n");
		/*LINTED*/
		_NOTE(COMPETING_THREADS_NOW);
		return (ENXIO);
	}

	if (tgt->suspended || tgt->pm_suspended) {
		(void) ddi_dev_is_needed(DEVINFO(tgt), 0, 1);
	}

	dkp = &tgt->map[BST_SLICE(dev)];
	if ((blkno < 0) || ((blkno + nblk) > dkp->p_size)) {
		BST_LOG(tgt->devp, BST_CE_DEBUG1,
			"dump - invalid block number\n");
		/*LINTED*/
		_NOTE(COMPETING_THREADS_NOW);
		return (EINVAL);
	}

	/*
	 * add starting blk # of slice (partition)
	 */
	blkno += tgt->offset[BST_SLICE(dev)];

	BST_LOG(tgt->devp, BST_CE_DEBUG1, "bst_dump: blkno = %x\n", blkno);

	bp = getrbuf(KM_NOSLEEP);
	if (bp == NULL) {
		/*LINTED*/
		_NOTE(COMPETING_THREADS_NOW);
		return (EIO);
	}

	bp->b_un.b_addr = addr;
	bp->b_edev = dev;
	bp->b_bcount = nblk * DEV_BSIZE;
	bp->b_flags = B_WRITE | B_BUSY;
	bp->b_blkno = blkno;

	pkt = scsi_init_pkt(ROUTE(tgt), NULL, bp, CDB_GROUP1,
	    (tgt->arq ? sizeof (struct scsi_arq_status) : 1),
	    sizeof (struct bst_pkt_private), 0, NULL_FUNC, NULL);

	if (pkt == (struct scsi_pkt *)NULL) {
		freerbuf(bp);
		/*LINTED*/
		_NOTE(COMPETING_THREADS_NOW);
		return (EIO);
	}
	(void) scsi_setup_cdb((union scsi_cdb *)pkt->pkt_cdbp,
			SCMD_WRITE_G1, blkno, nblk, 0);
	BST_DUMP_CDB(tgt, pkt, CDB_GROUP1);

	BST_LOG(tgt->devp, BST_CE_DEBUG1,
		"starting scsi_poll, pkt = %x\n", pkt);

	/*
	 * while dumping in polled mode, other cmds might
	 * complete and these should not be resubmitted. we
	 * set the dumping flag here which prevents bst_callback
	 * to requeue cmds
	 */
	tgt->dumping = 1;
	rval = bst_scsi_poll(tgt, pkt);
	tgt->dumping = 0;

	BST_LOG(tgt->devp, BST_CE_DEBUG1,
		"scsi_poll done, rval = %d pkt_reason = %x\n",
		rval, pkt->pkt_reason);

	scsi_destroy_pkt(pkt);
	freerbuf(bp);

	/*LINTED*/
	_NOTE(COMPETING_THREADS_NOW);
	if (rval != DDI_SUCCESS) {
		rval = EIO;
	}

	return (rval);
}


/*
 * open(9e)
 * Called for each open(2) call on the device.
 * There are no mutexes in the open or close routines here. Synchronization
 * and locking are done with a semaphore. This serializes open and close
 * calls: if one is in progress, another will block until it completes. If
 * you add access to data that may also be accessed by other than open/close,
 * you may need to add a mutex.
 *
 * Note: Credp can be used to restrict access to root, by calling drv_priv(9f);
 *	 see also cred(9s).
 *	 Flag shows the access mode (read/write). Check it if the device is
 *	 read or write only, or has modes where this is relevant.
 *	 Otyp is an open type flag, see open.h.
 */
/*ARGSUSED*/
static int
bst_open(dev_t *dev_p, int flag, int otyp, cred_t *cred_p)
{
	struct bst_soft	*tgt;
	dev_t		dev = *dev_p;
	int		slice;
	int		firstopen;

	if ((tgt = ddi_get_soft_state(bst_state, BST_INST(dev))) == NULL) {
		BST_LOG(0, BST_CE_DEBUG1, "open - unattached instance %d\n",
		    BST_INST(dev));
		return (ENXIO);		/* causes deferred attach */
	}
	slice = BST_SLICE(dev);	/* slice being opened */

	if ((otyp != OTYP_CHR) && (otyp != OTYP_BLK)) {
		BST_LOG(tgt->devp, BST_CE_DEBUG1,
		    "Unsupported open type %d", otyp);
		return (EINVAL);
	}
	BST_LOG(tgt->devp, BST_CE_DEBUG2, "open - %s slice%d\n",
	    (otyp == OTYP_CHR) ? "raw" : "block", slice);

	/*
	 * Decrement semaphore to lock out other open/close calls while
	 * this thread is active (see semaphore(9F)).
	 */
	sema_p(&tgt->semaphore);

	/*
	 * all device accesses go thru bst_strategy where we check
	 * on suspend status
	 */

	/*
	 * Find out if this is the first open. Don't need the mutex because
	 * this is read only access, also raw_map/block_map are accessed only
	 * in bst_open and bst_close which are protected by the semaphore.
	 */
	firstopen = (tgt->raw_map == 0) && (tgt->block_map == 0);

	/*
	 * If this is the first open or the device hasn't been responding,
	 * test to make sure it's still powered on and ready by sending
	 * the SCSI Test Unit Ready command.
	 */
	if (firstopen || (tgt->state == BST_STATE_NORESP)) {
		if (!bst_ready(tgt, dev)) {
			bst_log(tgt->devp, CE_NOTE, "Device not ready\n");
			sema_v(&tgt->semaphore);
			return (EIO);
		}
	}

	/*
	 * Check for exclusive open - exclusivity affects the whole disk,
	 * not just the slice being opened.
	 */
	if (flag & FEXCL) {
		if (!firstopen) {
			BST_LOG(tgt->devp, BST_CE_DEBUG1,
			    "(excl) slice already open\n");
			sema_v(&tgt->semaphore);
			return (EBUSY);
		}
		tgt->excl = 1;
	} else if (tgt->excl) {
		BST_LOG(tgt->devp, BST_CE_DEBUG1,
		    "already opened exclusively\n");
		sema_v(&tgt->semaphore);
		return (EBUSY);
	}

	/*
	 * Validate the slice being opened
	 */
	if (tgt->geom_ok == 0) {
		mutex_enter(BST_MUTEX(tgt));
		bst_read_capacity(tgt, dev);
		bst_checklabel(tgt);
		mutex_exit(BST_MUTEX(tgt));
	}
	if (tgt->geom_ok == 0 || tgt->map[slice].p_size == 0) {
		/*
		 * Fail unless we're opened for formatting
		 * (raw device opened NDELAY).
		 */
		if ((flag & (FNDELAY|FNONBLOCK)) == 0 || otyp != OTYP_CHR) {
			sema_v(&tgt->semaphore);
			return (ENXIO);
		}
	}


	/*
	 * Mark the slice open in the appropriate map
	 */
	if (otyp == OTYP_CHR) {
		tgt->raw_map |= 1 << slice;
	} else {
		tgt->block_map |= 1 << slice;
	}

	/*
	 * For removable disk drives, send the SCSI lock command on
	 * the first open
	 */
	if (firstopen && REMOVABLE(tgt)) {
		(void) bst_simple(dev, SCMD_DOORLOCK, BST_LOCK);
	}

	sema_v(&tgt->semaphore);
	return (0);
}


/*
 * close(9e)
 * Called on final close only, i.e. the last close(2) call.
 */
/*ARGSUSED*/
static int
bst_close(dev_t dev, int flag, int otyp, cred_t *cred_p)
{
	struct bst_soft *tgt;
	int	slice;


	if ((tgt = ddi_get_soft_state(bst_state, BST_INST(dev))) == NULL) {
		BST_LOG(0, BST_CE_DEBUG1, "bst_close bad instance %d\n",
		    BST_INST(dev));
		return (ENXIO);
	}
	slice = BST_SLICE(dev);

	BST_LOG(tgt->devp, BST_CE_DEBUG2, "close - %s slice%d\n",
	    (otyp == OTYP_CHR) ? "raw" : "block", slice);

	sema_p(&tgt->semaphore);
	_NOTE(NO_COMPETING_THREADS_NOW);

	/*
	 * Mark the slice closed in the appropriate map
	 */
	if (otyp == OTYP_CHR) {
		tgt->raw_map &= ~(1 << slice);
	} else {
		tgt->block_map &= ~(1 << slice);
	}

	tgt->excl = 0;		/* turn off exclusivity */

	/*
	 * If this was the last close (i.e. no partitions are open),
	 * and it's a removable disk, unlock the drive and invalidate
	 * the geometry
	 */


	if (!tgt->raw_map && !tgt->block_map &&
	    REMOVABLE(tgt)) {
		(void) bst_simple(dev, SCMD_DOORLOCK, BST_UNLOCK);
		tgt->geom_ok = 0;
	}

	sema_v(&tgt->semaphore);

	/*LINTED*/
	_NOTE(COMPETING_THREADS_NOW);

	return (0);
}


/*
 * Device Configuration Routine
 * link instance number (unit) with dev_info structure
 */
/*ARGSUSED*/
static int
bst_info(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg, void **result)
{
	struct bst_soft	*tgt;
	int		instance, error;

	BST_LOG(0, BST_CE_DEBUG2, "bst_info\n");

	switch (infocmd) {
	case DDI_INFO_DEVT2DEVINFO:
		instance = BST_INST((dev_t)arg);
		if ((tgt = ddi_get_soft_state(bst_state, instance)) == NULL)
			return (DDI_FAILURE);
		*result = (void *) tgt->devp->sd_dev;
		error = DDI_SUCCESS;
		break;
	case DDI_INFO_DEVT2INSTANCE:
		*result = (void *) BST_INST((dev_t)arg);
		error = DDI_SUCCESS;
		break;
	default:
		error = DDI_FAILURE;
	}
	return (error);
}


/*
 * Character (raw) read and write routines, called via read(2) and
 * write(2). These routines perform "raw" (i.e. unbuffered) i/o.
 * Since they're so similar, there's actually one 'rw' routine for both,
 * these devops entry points just call the general routine with the
 * appropriate flag.
 */
/* ARGSUSED2 */
static int
bst_read(dev_t dev, struct uio *uio, cred_t *cred_p)
{
	return (bst_rw(dev, uio, B_READ));
}

/* ARGSUSED2 */
static int
bst_write(dev_t dev, struct uio *uio, cred_t *cred_p)
{
	return (bst_rw(dev, uio, B_WRITE));
}


/*
 * Asynchronous character read and write routines
 */

/* ARGSUSED2 */
static int
bst_aread(dev_t dev, struct aio_req *aio, cred_t *cred_p)
{
	return (aphysio(bst_strategy, anocancel, dev, B_READ,
	    minphys, aio));
}

/* ARGSUSED2 */
static int
bst_awrite(dev_t dev, struct aio_req *aio, cred_t *cred_p)
{
	return (aphysio(bst_strategy, anocancel, dev, B_WRITE,
	    minphys, aio));
}


/*
 * General raw read/write routine
 * Just verify the unit number and transfer offset & length, and call
 * strategy via physio. Physio(9f) will take care of address mapping
 * and locking, and will split the transfer if ncessary, based on minphys,
 * possibly calling the strategy routine multiple times.
 */
static int
bst_rw(dev_t dev, struct uio *uio, int flag)
{
	struct bst_soft	*tgt;

	if ((tgt = ddi_get_soft_state(bst_state, BST_INST(dev))) == NULL) {
		return (ENXIO);
	}
	BST_LOG(tgt->devp, BST_CE_DEBUG2, "%s\n",
	    (flag == B_READ) ? "Read" : "Write");

	/*
	 * Verify the offset/length of the request - disk drives will
	 * accept only multiples of sectors.
	 */
	if (uio->uio_offset & (DEV_BSIZE-1)) {
		BST_LOG(tgt->devp, BST_CE_DEBUG1,
		    "Offset %d is not an integral number of sectors\n",
		    uio->uio_offset);
		return (EINVAL);
	} else if (uio->uio_iov->iov_len & (DEV_BSIZE-1)) {
		BST_LOG(tgt->devp, BST_CE_DEBUG1,
		    "Transfer length %d is not an integral number of sectors\n",
		    uio->uio_iov->iov_len);
		return (EINVAL);
	}

	return (physio(bst_strategy, NULL, dev, flag, minphys, uio));
}


/*
 * strategy
 * Main routine for commands to the device.  Called from bst_rw(),
 * bst_ioctl_cmd, and directly from the file system.
 * The mutex prevents this routine from being called simultaneously by
 * two threads.
 */
static int
bst_strategy(struct buf *bp)
{
	struct bst_soft *tgt;
	struct diskhd *dp;

	tgt = ddi_get_soft_state(bst_state, BST_INST(bp->b_edev));
	if (tgt == NULL) {
		bp->b_resid = bp->b_bcount;
		bioerror(bp, ENXIO);
		biodone(bp);
		return (0);
	}

	BST_LOG(tgt->devp, BST_CE_DEBUG2, "bst_strategy - %s\n",
	    (bp == tgt->sbufp) ? "special" : "regular");

	/*
	 * mark busy so we won't get powered down again while
	 * trying to resume.
	 * Note that there must be one pm_idle_components() for each
	 * pm_busy_components().
	 */
	(void) pm_busy_component(DEVINFO(tgt), 0);

	mutex_enter(BST_MUTEX(tgt));

	do {
		/*
		 * block cmds if/while device is suspended via
		 * DDI_SUSPEND
		 */
		while (tgt->suspended) {
			cv_wait(&tgt->suspend_cv, BST_MUTEX(tgt));
		}

		if (tgt->power_level == 0) {
			/*
			 * drop mutex because ddi_dev_is_needed()
			 * will result in a call to bst_power()
			 */
			mutex_exit(BST_MUTEX(tgt));
			(void) ddi_dev_is_needed(DEVINFO(tgt), 0, 1);
			mutex_enter(BST_MUTEX(tgt));
		}

		/*
		 * block cmds if device is still suspended with
		 * DDI_PM_SUSPEND; because we had to drop the mutex
		 * to call ddi_dev_is_needed we may be executing in a
		 * thread that came in after the power level was raised
		 * but before attach was called with DDI_PM_RESUME
		 */
		while (tgt->pm_suspended) {
			cv_wait(&tgt->suspend_cv, BST_MUTEX(tgt));
		}
	} while (tgt->suspended || (tgt->power_level == 0));

	mutex_exit(BST_MUTEX(tgt));

	if (bp != tgt->sbufp) {		/* Regular read/write */
		daddr_t		bn = bp->b_blkno;
		struct partition *dkp;
		int		blks;


		/*
		 * First check that the device is still there
		 */
		if (tgt->state == BST_STATE_NORESP) {
			if (!bst_ready(tgt, bp->b_edev)) {
				bp->b_resid = bp->b_bcount;
				bioerror(bp, ENXIO);
				biodone(bp);
				(void) pm_idle_component(DEVINFO(tgt), 0);
				return (0);
			}
		}

		/*
		 * bp->b_blkno is the offset into the disk slice.
		 * The offset plus the requested length must not be
		 * beyond the end of the slice.
		 */
		dkp = &tgt->map[BST_SLICE(bp->b_edev)];
		blks = bp->b_bcount >> DEV_BSHIFT;
		if ((bn < 0) || (bn + blks > dkp->p_size)) {
			bst_log(tgt->devp, CE_WARN,
			    "block %d out of range (len=%d, count=%d)\n",
			    bn, dkp->p_size, blks);
			bp->b_resid = bp->b_bcount;
			bioerror(bp, EINVAL);
			biodone(bp);
			(void) pm_idle_component(DEVINFO(tgt), 0);
			return (0);
		}

		/*
		 * Calculate the physical block number for disksort(9F);
		 * it uses b_resid as the sort key
		 */
		bp->b_resid = tgt->offset[BST_SLICE(bp->b_edev)] + bn;
	} else {
		/*
		 * Special command, put it at the front of the queue
		 */
		bp->b_resid = 0;
	}

	/*
	 * Add the new request into the i/o queue, pointed to by
	 * tgt->iorq. The queue is chained through av_forw
	 * and dp->b_forw holds the currently active request.
	 * Requests are added to the queue by disksort(9F), which
	 * uses the value in b_resid as the sort key. I'm using
	 * the physical block number here, although the cylinder
	 * number can also be used.
	 */
	dp = &tgt->iorq;		/* i/o request queue for this device */

	bp->av_forw = NULL;
	bp->av_back = NULL;		/* av_back is BST_BP_PKT */

	/*
	 * LOCK the tgt struct so we can fiddle with the queue pointers.
	 * We need to hold it for a bit because we don't want anyone to
	 * change dp->b_forw or dp->av_forw while we're accessing them.
	 */
	mutex_enter(BST_MUTEX(tgt));

	disksort(dp, bp);		/* add request into queue */

	if (dp->b_forw == NULL) { /* is device inactive? */
		mutex_exit(BST_MUTEX(tgt));
		bst_start(tgt);
	} else if (BST_BP_PKT(dp->av_forw) == 0) {
		/*
		 * There's an active request, so bst_done will call bst_start
		 * when it's completed. Try to pre-allocate resources for
		 * the next request in line. Don't care if it fails.
		 */
		bst_make_cmd(tgt, dp->av_forw, NULL_FUNC);
		mutex_exit(BST_MUTEX(tgt));
	} else {
		mutex_exit(BST_MUTEX(tgt));
	}

	return (0);
}


/*
 * ioctl calls.
 * Ioctls are device specific. This driver supports all of the DKIOC commands
 * and the USCSI (CDB pass through) command, used by format(1M).
 */
/* ARGSUSED3 */
static int
bst_ioctl(dev_t dev, int cmd, intptr_t arg, int flag,
	cred_t *cred_p, int *rval_p)
{
	int			err = 0;
	int			len;
	struct scsi_device	*devp;
	struct bst_soft		*tgt;
	char			data[DEV_BSIZE];
	u_char			otheropen;
	int			target, lun;
	struct dk_cinfo		*ctlr;
	struct vtoc		*vtocp;

	if ((tgt = ddi_get_soft_state(bst_state, BST_INST(dev))) == NULL) {
		return (ENXIO);		/* invalid minor number */
	}
	devp = tgt->devp;
	bzero((caddr_t)data, DEV_BSIZE);

	BST_LOG(devp, BST_CE_DEBUG2, "bst_ioctl: cmd = 0x%x\n", cmd);

	/*
	 * all device accesses go thru bst_strategy where we check
	 * on suspend status
	 */

	switch (cmd) {

	case USCSICMD:
		/*
		 * Check root permissions
		 */
		if (drv_priv(cred_p) != 0) {
			return (EPERM);
		}

		BST_LOG(devp, BST_CE_DEBUG3, "USCSICMD\n");
		/*
		 * Send an arbitrary SCSI command to the device. The user
		 * has specified the CDB, and and address/length for data
		 * (if needed). We do no validation of the passed uscsi_cmd
		 * structure (defined in <sys/scsi/impl/uscsi.h>).
		 */
		if (ddi_copyin((caddr_t)arg, data,
		    sizeof (struct uscsi_cmd), flag)) {
			return (EFAULT);
		}
		err = bst_ioctl_cmd(dev, (struct uscsi_cmd *)data,
		    UIO_USERSPACE, UIO_USERSPACE);
		if (ddi_copyout(data, (caddr_t)arg,
		    sizeof (struct uscsi_cmd), flag)) {
			err = EFAULT;
		}
		break;

	/*
	 * Ioctl commands for general disk access (formatting etc)
	 */
	case DKIOCINFO:		/* return controller info (struct dk_cinfo) */
		target = devp->sd_address.a_target;
		lun = devp->sd_address.a_lun;

		ctlr = (struct dk_cinfo *)data;
		(void) strcpy(ctlr->dki_cname,
		    ddi_get_name(ddi_get_parent(devp->sd_dev)));
		if (tgt->iscd) {
			ctlr->dki_ctype = DKC_CDROM;
		} else {
			ctlr->dki_ctype = DKC_SCSI_CCS;
		}
		ctlr->dki_flags = DKI_FMTVOL;
		ctlr->dki_cnum = ddi_get_instance(ddi_get_parent(devp->sd_dev));
		(void) strcpy(ctlr->dki_dname, ddi_get_name(devp->sd_dev));

		ctlr->dki_unit = ddi_get_instance(devp->sd_dev);
		ctlr->dki_slave = (target << 3) | lun;
		ctlr->dki_partition = BST_SLICE(dev);

		/*
		 * Max Transfer size of this device.
		 * We pick an arbitrarily low value that
		 * can be supported on all platforms
		 * for the particular architecture.
		 */
		ctlr->dki_maxtransfer = (56 * 1024) / DEV_BSIZE;

		BST_LOG(tgt->devp, BST_CE_DEBUG3,
		    "DKIOCINFO: cname=%s, cnum=%d, unit=%d, slave=%d, ptn=%d\n",
		    ctlr->dki_cname, ctlr->dki_cnum, ctlr->dki_unit,
		    ctlr->dki_slave, ctlr->dki_partition);

		if (ddi_copyout(data, (caddr_t)arg,
		    sizeof (struct dk_cinfo), flag)) {
			err = EFAULT;
		}
		break;

	case DKIOCGGEOM:	/* get geometry (struct dk_geom) */
		BST_LOG(tgt->devp, BST_CE_DEBUG3, "DKIOCGGEOM\n");

		mutex_enter(BST_MUTEX(tgt));
		bcopy((caddr_t)&tgt->geom, (caddr_t)data,
		    sizeof (struct dk_geom));
		mutex_exit(BST_MUTEX(tgt));

		if (ddi_copyout((caddr_t)data, (caddr_t)arg,
		    sizeof (struct dk_geom), flag)) {
			err = EFAULT;
		}
		break;

	case DKIOCSGEOM:	/* set geometry (struct dk_geom) */
		BST_LOG(tgt->devp, BST_CE_DEBUG3, "DKIOCSGEOM\n");

		if (ddi_copyin((caddr_t)arg, (caddr_t)data,
		    sizeof (struct dk_geom), flag)) {
			err = EFAULT;
		} else {
			mutex_enter(BST_MUTEX(tgt));
			bcopy((caddr_t)data, (caddr_t)&tgt->geom,
			    sizeof (struct dk_geom));
			mutex_exit(BST_MUTEX(tgt));
			if (bst_debug > 0) {
				cmn_err(CE_CONT, "DKIOCSGEOM:  ncyl=%d	"
				    "acyl=%d  nhead=%d	nsect=%d\n",
				    tgt->geom.dkg_ncyl, tgt->geom.dkg_acyl,
				    tgt->geom.dkg_nhead, tgt->geom.dkg_nsect);
			}
		}
		break;

	case DKIOCGAPART:	/* get partition map  (struct dk_map[NDKMAP]) */
		BST_LOG(tgt->devp, BST_CE_DEBUG3, "DKIOCGAPART\n");
		mutex_enter(BST_MUTEX(tgt));
		bst_getmap(tgt, (struct dk_map *)data);
		mutex_exit(BST_MUTEX(tgt));

		if (ddi_copyout((caddr_t)data, (caddr_t)arg,
		    sizeof (struct dk_map) * NDKMAP, flag)) {
			err = EFAULT;
		}
		break;

	case DKIOCSAPART:	/* set partition map  (struct dk_map[NDKMAP]) */
		BST_LOG(tgt->devp, BST_CE_DEBUG3, "DKIOCSAPART\n");
		len = sizeof (struct dk_map) * NDKMAP;
		if (ddi_copyin((caddr_t)arg, data, len, flag)) {
			err = EFAULT;
		} else {
			mutex_enter(BST_MUTEX(tgt));
			bst_chgmap(tgt, (struct dk_map *)data);
			mutex_exit(BST_MUTEX(tgt));
		}
		break;

	case DKIOCEJECT:	/* eject removable media */
		BST_LOG(tgt->devp, BST_CE_DEBUG3, "DKIOCEJECT\n");
		otheropen = tgt->raw_map & ~(1 << BST_SLICE(dev));
		if (!REMOVABLE(tgt)) {
			err = ENOTTY;
		} else if (tgt->block_map || otheropen) {
			err = EBUSY;	/* someone else has the device open */
		} else if (!bst_simple(dev, SCMD_DOORLOCK, BST_UNLOCK) ||
		    !bst_simple(dev, SCMD_START_STOP, BST_EJECT)) {
			err = EIO;
		} else {


	_NOTE(NO_COMPETING_THREADS_NOW);
			/*
			 * Don't need to mutex here because we know we're
			 * the only thread in the driver (there are no other
			 * opens). If the user's forked they're on their own!
			 */
			tgt->geom_ok = 0;		/* disk ejected */
	/*LINTED*/
	_NOTE(COMPETING_THREADS_NOW);

		}
		break;

	case DKIOCGVTOC:	/* get volume table of contents (struct vtoc) */
		BST_LOG(tgt->devp, BST_CE_DEBUG3, "DKIOCGVTOC\n");

		if (tgt->geom_ok == 0) {
			err = EINVAL;
		} else {
			mutex_enter(BST_MUTEX(tgt));
			bst_dkioc_gvtoc(tgt, (struct vtoc *)data);
			mutex_exit(BST_MUTEX(tgt));
			if (ddi_copyout(data, (caddr_t)arg,
			    sizeof (struct vtoc), flag)) {
				err = EFAULT;
			}
		}
		break;

	case DKIOCSVTOC:	/* set volume table of contents (struct vtoc) */
		BST_LOG(tgt->devp, BST_CE_DEBUG3, "DKIOCSVTOC\n");

		if (tgt->geom_ok == 0) {
			err = EINVAL;
		} else if (ddi_copyin((caddr_t)arg, data,
		    sizeof (struct vtoc), flag)) {
			err = EFAULT;
		} else {
			vtocp = (struct vtoc *)data;
			mutex_enter(BST_MUTEX(tgt));
			if ((err = bst_dkioc_svtoc(vtocp, tgt)) == 0) {
				if ((err = bst_write_label(dev)) == 0) {
					bst_checklabel(tgt);
				}
			}
			mutex_exit(BST_MUTEX(tgt));
		}

		break;

#if defined(_SUNOS_VTOC_16)
	/*
	 * retrieve the HBA geometry
	 */
	case DKIOCG_VIRTGEOM:
	    {
		int		i;
		struct dk_geom	*gp;

		gp = (struct dk_geom *)data;

		i = scsi_ifgetcap(ROUTE(tgt), "geometry", 1);
		gp->dkg_nhead = (i >> 16) & 0xff;
		gp->dkg_nsect = i & 0xff;
		gp->dkg_ncyl  = tgt->capacity /
				(gp->dkg_nhead * gp->dkg_nsect);
		gp->dkg_acyl  = 0;
		gp->dkg_pcyl  = gp->dkg_ncyl + gp->dkg_acyl;
		if (ddi_copyout((caddr_t)gp, (caddr_t)arg,
		    sizeof (struct dk_geom), flag)) {
			err = EFAULT;
		}
		break;
	    }

	/*
	 * retrieve the physical geometry of the disk
	 */
	case DKIOCG_PHYGEOM:
	    {
		struct dk_geom	*gp;

		gp = (struct dk_geom *)data;

		gp->dkg_nhead = tgt->geom.dkg_nhead;
		gp->dkg_nsect = tgt->geom.dkg_nsect;
		gp->dkg_ncyl  = tgt->capacity /
				(gp->dkg_nhead * gp->dkg_nsect);
		gp->dkg_acyl  = 0;
		gp->dkg_pcyl  = gp->dkg_ncyl + gp->dkg_acyl;
		if (ddi_copyout((caddr_t)gp, (caddr_t)arg,
		    sizeof (struct dk_geom), flag)) {
			err = EFAULT;
		}
		break;
	    }
#endif

	default:

		err = ENOTTY;
		break;
	}
	return (err);
}


/*
 * Run a command for user (from bst_ioctl) or from someone else in the driver.
 *
 * cdbspace is for address space of cdb; dataspace is for address space
 * of the buffer - user or kernel.
 */
static int
bst_ioctl_cmd(dev_t dev, struct uscsi_cmd *scmd,
    enum uio_seg cdbspace, enum uio_seg dataspace)
{
	caddr_t	cdb, user_cdbp;
	int	err, rw;
	struct	buf *bp;
	struct	scsi_pkt *pkt;
	struct	bst_soft *tgt;
	int	flag = 0;

	if ((tgt = ddi_get_soft_state(bst_state, BST_INST(dev))) == NULL) {
		BST_LOG(0, BST_CE_DEBUG1,
		    "bst_ioctl_cmd: invalid instance %d\n", BST_INST(dev));
		return (ENXIO);
	}
	ASSERT(mutex_owned(BST_MUTEX(tgt)) == 0);

	BST_LOG(tgt->devp, BST_CE_DEBUG2, "bst_ioctl_cmd\n");

	/*
	 * The uscsi structure itself is already in kernel space (copied
	 * in by bst_ioctl, or declared there by our caller); but
	 * we need to copy in the cdb here.
	 */
	cdb = kmem_zalloc(scmd->uscsi_cdblen, KM_SLEEP);
	if (cdbspace == UIO_SYSSPACE) {
		flag |= FKIOCTL;
	}
	if (ddi_copyin(scmd->uscsi_cdb, cdb,
	    (u_int)scmd->uscsi_cdblen, flag)) {
		kmem_free(cdb, (size_t)scmd->uscsi_cdblen);
		BST_LOG(tgt->devp, BST_CE_DEBUG1,
		    "bst_ioctl_cmd: ddi_copyin failed\n");
		return (EFAULT);
	}

	/*
	 * The cdb pointer in the structure passed by the user is pointing
	 * to user space. We've just copied the cdb into a local buffer,
	 * so point uscsi_cdb to it now. We'll restore hte user's pointer
	 * at the end.
	 */
	user_cdbp = scmd->uscsi_cdb;	/* save the user's pointer */
	scmd->uscsi_cdb = cdb;		/* point to the local cdb buffer */
	rw = (scmd->uscsi_flags & USCSI_READ) ? B_READ : B_WRITE;

	/*
	 * Get the 'special command' buffer.
	 * First lock the tgt struct; if the buffer's busy, wait for
	 * it to become free. Once we get the buffer, mark it busy to
	 * lock out other requests for it.
	 * Note	cv_wait will release the mutex, allowing other parts
	 *	of the driver to acquire it.
	 * Note	Once we have the special buffer, we can safely release
	 *	the mutex; the buffer is now busy and another thread will
	 *	block in the cv_wait until we release it. All the code
	 *	from here until we unset the busy flag is non-reentrant,
	 *	including the physio/strategy/start/callback/done thread.
	 */
	mutex_enter(BST_MUTEX(tgt));
	while (tgt->sbuf_busy)
		cv_wait(&tgt->sbuf_cv, BST_MUTEX(tgt));
	tgt->sbuf_busy = 1;

	bp = tgt->sbufp;
	mutex_exit(BST_MUTEX(tgt));

	if (scmd->uscsi_buflen) {
		/*
		 * We're sending/receiving data; create a uio structure and
		 * call physio to do the right things.
		 */
		struct iovec aiov;
		struct uio auio;
		struct uio *uio = &auio;

		BST_LOG(tgt->devp, BST_CE_DEBUG2,
		    "Data transfer (%d bytes)\n", scmd->uscsi_buflen);

		bzero((caddr_t)&auio, sizeof (struct uio));
		bzero((caddr_t)&aiov, sizeof (struct iovec));
		aiov.iov_base = scmd->uscsi_bufaddr;
		aiov.iov_len = scmd->uscsi_buflen;
		uio->uio_iov = &aiov;

		uio->uio_iovcnt = 1;
		uio->uio_resid = scmd->uscsi_buflen;
		uio->uio_segflg = dataspace;
		uio->uio_offset = 0;
		uio->uio_fmode = 0;

		/*
		 * Let physio do the rest...
		 */
		bp->av_back = NULL;		/* av_back is BST_BP_PKT */
		bp->b_private = (void *) scmd;
		err = physio(bst_strategy, bp, dev, rw, minphys, uio);
	} else {
		BST_LOG(tgt->devp, BST_CE_DEBUG2, "No data\n");
		/*
		 * No data transfer, we can call bst_strategy directly
		 */
		bp->av_back = NULL;		/* av_back is BST_BP_PKT */
		bp->b_private = (void *) scmd;
		bp->b_flags = B_BUSY | rw;
		bp->b_edev = dev;
		bp->b_bcount = bp->b_blkno = 0;
		(void) bst_strategy(bp);
		err = biowait(bp);
	}

	/*
	 * get the status block, if any, and
	 * release any resources that we had.
	 */
	scmd->uscsi_status = 0;
	if ((pkt = BST_BP_PKT(bp)) != NULL) {
		scmd->uscsi_status = SCBP_C(pkt);
		scsi_destroy_pkt(pkt);
		bp->av_back = NULL;		/* av_back is BST_BP_PKT */
	}
	BST_LOG(tgt->devp, BST_CE_DEBUG2,
	    "SCSI Command status = 0x%x\n", scmd->uscsi_status);

	/*
	 * Tell anybody who cares that the special buffer is now available.
	 * Again, lock the tgt struct first, then mark the buffer not
	 * busy and signal anyone who may be waiting for it.
	 */
	mutex_enter(BST_MUTEX(tgt));		/* LOCK the tgt struct */
	tgt->sbuf_busy = 0;
	cv_signal(&tgt->sbuf_cv);
	mutex_exit(BST_MUTEX(tgt));		/* UNLOCK the tgt struct */

	kmem_free(scmd->uscsi_cdb, (size_t)scmd->uscsi_cdblen);
	scmd->uscsi_cdb = user_cdbp;	/* restore the user's pointer */
	BST_LOG(tgt->devp, BST_CE_DEBUG2,
	    "bst_ioctl_cmd returning 0x%x\n", err);
	return (err);
}


/*
 * Unit start and Completion
 * Get the next command from the queue, allocate resources if they
 * haven't already been allocated, and send off the command.
 */
static void
bst_start(struct bst_soft *tgt)
{
	struct buf	*bp;
	struct diskhd	*dp;

	BST_LOG(tgt->devp, BST_CE_DEBUG2, "bst_start\n");

	dp = &tgt->iorq;

	/*
	 * LOCK the tgt struct, we're going to fiddle with the
	 * queue pointers
	 */
	mutex_enter(BST_MUTEX(tgt));

	/*
	 * If b_forw is non-NULL there's an active request already,
	 * if av_forw is NULL there are no requests queued. In either
	 * case do nothing.
	 */
	if (tgt->dumping || dp->b_forw || (bp = dp->av_forw) == NULL) {
		mutex_exit(BST_MUTEX(tgt));
		return;
	}

	/*
	 * If resources have not been allocated yet, allocate them now.
	 * Unlike the 'pre-allocation' calls, specify bst_runout here;
	 * if the allocation fails bst_runout will be called to retry
	 * later.
	 */
	if (!BST_BP_PKT(bp)) {
		bst_make_cmd(tgt, bp, bst_runout);
		if (!BST_BP_PKT(bp)) {
			if (geterror(bp) != 0) {
				/*
				 * scsi_init_pkt failed for other than
				 * out of resources. Mark error and pull
				 * the request off the pending queue
				 * for bst_done.
				 */
				bp->b_resid = bp->b_bcount;
				dp->b_forw = bp;
				dp->av_forw = bp->av_forw;
				mutex_exit(BST_MUTEX(tgt));
				bst_done(tgt, bp);
				return;
			}

			tgt->state = BST_STATE_RWAIT;
			mutex_exit(BST_MUTEX(tgt));
			return;
		} else if (tgt->state == BST_STATE_RWAIT) {
			tgt->state = BST_STATE_NIL;
		}
	}

	dp->b_forw = bp;		/* new active request */
	dp->av_forw = bp->av_forw;	/* take it off the pending queue */

	mutex_exit(BST_MUTEX(tgt));	/* UNLOCK the tgt structure */

	bp->av_forw = 0;

	/*
	 * Clear the resid now. Must do it after bst_make_cmd, else if
	 * that fails and we get held in RWAIT, the next call to disksort
	 * (from another thread) will use our (zero) b_resid as the sort
	 * key, which would be wrong.
	 */
	bp->b_resid = 0;

	/*
	 * Send the SCSI command to the host adapter
	 */
	if (scsi_transport(BST_BP_PKT(bp)) != TRAN_ACCEPT) {
		BST_LOG(tgt->devp, BST_CE_DEBUG1,
		    "Command transport failed\n");
		bioerror(bp, EIO);
		bst_done(tgt, bp);
	} else {
		/*
		 * Command sent OK, try to pre-allocate resources for
		 * the next request, don't care if it fails.
		 * Need to reacquire the mutex to lock the tgt struct
		 */
		BST_LOG(tgt->devp, BST_CE_DEBUG3,
		    "Command Transported\n");
		mutex_enter(BST_MUTEX(tgt));
		if (dp->av_forw && !BST_BP_PKT(dp->av_forw)) {
			bst_make_cmd(tgt, dp->av_forw, NULL_FUNC);
		}
		mutex_exit(BST_MUTEX(tgt));
	}
}


/*
 * 'Out of resources' callback routine. Passed to scsi_init_pkt(); if
 * the allocation fails, this routine is called when resources may be
 * available.
 * Just call bst_start to retry the resource allocation.
 */
/*ARGSUSED*/
static int
bst_runout(caddr_t arg)
{
	int		rval = 1;
	struct bst_soft	*tgt = (struct bst_soft *)arg;

	BST_LOG(0, BST_CE_DEBUG2, "bst_runout\n");

	if (tgt->state == BST_STATE_RWAIT) {
		bst_start(tgt);
		if (tgt->state == BST_STATE_RWAIT) {
			rval = 0;
		}
	}

	return (rval);
}


/*
 * Done with a command.
 * Start the next command and then call biodone() to tell physio or
 * bst_ioctl_cmd that this i/o has completed.
 *
 */
static void
bst_done(struct bst_soft *tgt, struct buf *bp)
{
	struct diskhd		*dp;
	struct scsi_pkt		*pkt;
	struct bst_pkt_private	*prv;
	int			dma_resid;

	BST_LOG(tgt->devp, BST_CE_DEBUG2, "bst_done\n");

	mutex_enter(BST_MUTEX(tgt));	/* LOCK the tgt struct */

	pkt = BST_BP_PKT(bp);
	prv = (struct bst_pkt_private *)pkt->pkt_private;
	dma_resid = prv->dma_resid;

	if (bp != tgt->sbufp && geterror(bp) == 0 && dma_resid != 0) {
		ASSERT(pkt->pkt_resid == 0);
		bp->b_resid -= dma_resid;
		if (bst_next_xfer(tgt, bp, pkt, prv)) {
			mutex_exit(BST_MUTEX(tgt));
			return;
		}

		/*
		 * Error setting up next portion of cmd transfer
		 */
		bp->b_resid += dma_resid;
		bioerror(bp, EIO);
	}

	dp = &tgt->iorq;		/* get the queue pointer */
	if (bp == dp->b_forw) {
		dp->b_forw = NULL;	/* request is no longer active */
	}


	/*
	 * Start the next one before releasing resources on this one
	 * unless a suspend has been requested
	 */
	if (tgt->suspended) {
		cv_broadcast(&tgt->disk_busy_cv);
		mutex_exit(BST_MUTEX(tgt));	/* UNLOCK the tgt struct */
	} else {
		mutex_exit(BST_MUTEX(tgt));	/* UNLOCK the tgt struct */
		bst_start(tgt);
	}

	/*
	 * Free resources unless this was a special command (in which
	 * case bst_ioctl_cmd will do the free).
	 */
	if (bp != tgt->sbufp) {
		scsi_destroy_pkt(BST_BP_PKT(bp));
	}
	biodone(bp);	/* tell interested parties that the i/o is done */
	(void) pm_idle_component(DEVINFO(tgt), 0);
}


static int
bst_next_xfer(
	struct bst_soft		*tgt,
	struct buf		*bp,
	struct scsi_pkt		*pkt,
	struct bst_pkt_private	*prv)
{
	int			secnt;
	daddr_t			blkno;
	int			resid;
	int			overrun;

	ASSERT(mutex_owned(BST_MUTEX(tgt)));
	ASSERT(bp != tgt->sbufp);
	ASSERT(pkt->pkt_resid == 0);

	/*
	 * Calculate next block number and amount to be transferred
	 */
	overrun = bp->b_resid;
	bp->b_bcount -= overrun;

	resid = prv->dma_resid;
	blkno = bp->b_blkno + ((bp->b_bcount - resid) >> DEV_BSHIFT);
	secnt = resid >> DEV_BSHIFT;

	/*
	 * Adjust block number to absolute
	 */
	blkno += tgt->offset[BST_SLICE(bp->b_edev)];
	blkno >>= tgt->blkshf;

	if (tgt->secdiv != DEV_BSHIFT) {
		secnt >>= (tgt->secdiv - DEV_BSHIFT);
	}

	/*
	 * Move pkt to the next portion of the xfer.
	 * We are called in this case with the
	 * disk mutex held, so we cannot sleep.
	 */

	if (scsi_init_pkt(ROUTE(tgt), pkt, bp, 0, 0, 0, 0,
			DDI_DMA_DONTWAIT, NULL) == NULL) {
		bp->b_bcount += overrun;
		return (0);
	}

	bp->b_bcount += overrun;

	/*
	 * Adjust things if there is still more DMA resid
	 */
	prv->dma_resid = pkt->pkt_resid;
	if (pkt->pkt_resid != 0) {
		secnt -= (pkt->pkt_resid >> tgt->secdiv);
	}

	(void) scsi_setup_cdb((union scsi_cdb *)pkt->pkt_cdbp,
	    (bp->b_flags & B_READ) ? SCMD_READ_G1 : SCMD_WRITE_G1,
	    (int)blkno, secnt, 0);

	pkt->pkt_resid = 0;
	mutex_exit(BST_MUTEX(tgt));
	if (scsi_transport(pkt) != TRAN_ACCEPT) {
		mutex_enter(BST_MUTEX(tgt));
		return (0);
	}
	mutex_enter(BST_MUTEX(tgt));

	return (1);
}



/*
 * Allocate resources for a SCSI command - call scsi_init_pkt to get
 * a packet and allocate DVMA resources, and create the SCSI CDB.
 * If scsi_init_pkt fails, give up silently. If bst_runout was specified
 * as func, the allocation will be retried, if NULL_FUNC was called
 * nothing will happen - the caller will not be notified of the failure.
 */
static void
bst_make_cmd(struct bst_soft *tgt, struct buf *bp, int (*func)())
{
	int			com;
	struct scsi_pkt		*pkt;
	struct uscsi_cmd	 *scmd = (struct uscsi_cmd *)bp->b_private;
	struct bst_pkt_private	*prv;
	int			flags, tval;

	/*
	 * Normally, mutexes are used to serialize data structure
	 * access. Here we use it to serialize access to this routine,
	 * which is called from several places in the driver. We could
	 * acquire and drop it in this routine, but since the caller may
	 * need the lock anyway, it's easier just to hold it than to keep
	 * dropping and reacquiring it
	 */
	ASSERT(mutex_owned(BST_MUTEX(tgt)));

	BST_LOG(tgt->devp, BST_CE_DEBUG2,
	    "bst_make_cmd: block %d, count=%d\n", bp->b_blkno, bp->b_bcount);

	flags = 0;
	if ((scsi_options & SCSI_OPTIONS_DR) == 0)
		flags = FLAG_NODISCON;

	if (bp != tgt->sbufp) {
		daddr_t	blkno;
		int	count;

		/*
		 * get starting blk # of slice (partition)
		 */
		blkno = tgt->offset[BST_SLICE(bp->b_edev)] + bp->b_blkno;

		pkt = scsi_init_pkt(ROUTE(tgt), NULL, bp, CDB_GROUP1,
		    (tgt->arq ? sizeof (struct scsi_arq_status) : 1),
		    sizeof (struct bst_pkt_private), PKT_DMA_PARTIAL,
		    func, (caddr_t)tgt);
		if (pkt == (struct scsi_pkt *)NULL) {
			return;
		}
		prv = (struct bst_pkt_private *)pkt->pkt_private;
		prv->bp = bp;
		prv->dma_resid = pkt->pkt_resid;
		prv->cdblen = CDB_GROUP1;

		/*
		 * Figure out how much of this command we are
		 * initially able to transport and turn this
		 * into the number of blocks.
		 */
		count = (bp->b_bcount - pkt->pkt_resid) >> DEV_BSHIFT;

		com = (bp->b_flags & B_READ) ? SCMD_READ_G1 : SCMD_WRITE_G1;

		(void) scsi_setup_cdb((union scsi_cdb *)pkt->pkt_cdbp,
			com, blkno, count, 0);
		BST_DUMP_CDB(tgt, pkt, CDB_GROUP1);
		pkt->pkt_flags = flags;
		tval = bst_io_time;
	} else {
		/*
		 * All special command come through bst_ioctl_cmd, which
		 * uses the uscsi interface. Just need to get the CDB
		 * from scmd and plug it in. Still call scsi_setup_cdb because
		 * it fills in some of the pkt field for us. Its cdb
		 * manipulations will be overwritten by the bcopy.
		 */
		if (scmd->uscsi_flags & USCSI_SILENT)
			flags |= FLAG_SILENT;
		if (scmd->uscsi_flags & USCSI_DIAGNOSE)
			flags |= FLAG_DIAGNOSE;
		if (scmd->uscsi_flags & USCSI_ISOLATE)
			flags |= FLAG_ISOLATE;
		if (scmd->uscsi_flags & USCSI_NODISCON)
			flags |= FLAG_NODISCON;

		pkt = scsi_init_pkt(ROUTE(tgt), (struct scsi_pkt *)NULL,
		    (bp->b_bcount) ? bp : NULL, scmd->uscsi_cdblen,
		    tgt->arq ? sizeof (struct scsi_arq_status) : 1,
		    sizeof (struct bst_pkt_private), 0, func,
		    (caddr_t)tgt);

		prv = (struct bst_pkt_private *)pkt->pkt_private;
		prv->bp = bp;
		prv->dma_resid = pkt->pkt_resid;
		prv->cdblen = scmd->uscsi_cdblen;

		(void) scsi_setup_cdb((union scsi_cdb *)pkt->pkt_cdbp,
			scmd->uscsi_cdb[0], 0, 0, 0);
		pkt->pkt_flags = flags;
		bcopy(scmd->uscsi_cdb,
		    (caddr_t)pkt->pkt_cdbp, scmd->uscsi_cdblen);
		tval = scmd->uscsi_timeout;
		BST_DUMP_CDB(tgt, pkt, scmd->uscsi_cdblen);
	}

	pkt->pkt_comp = bst_callback;
	pkt->pkt_time = tval;
	bp->av_back = (struct buf *)pkt;	/* av_back is BST_BP_PKT */
}


/*
 * -----------------------------------------------------------------------------
 * Interrupt Service Routines
 */

/*
 * Restart a command - the device was either busy or not ready
 */
static void
bst_restart(caddr_t arg)
{
	struct bst_soft *tgt = (struct bst_soft *)arg;
	struct buf	*bp;

	BST_LOG(tgt->devp, BST_CE_DEBUG2, "bst_restart\n");

	bp = tgt->iorq.b_forw;	/* read only, no need to lock */
	if (bp) {
		struct scsi_pkt *pkt = BST_BP_PKT(bp);
		if (!tgt->arq && (pkt->pkt_flags & FLAG_SENSING)) {
			pkt = tgt->rqs;
		}

		if (scsi_transport(pkt) != TRAN_ACCEPT) {
			bp->b_resid = bp->b_bcount;
			bioerror(bp, EIO);
			bst_done(tgt, bp);
		}
	}
}


/*
 * Command completion processing, called by the host adapter driver
 * when it's done with the command.
 */
static void
bst_callback(struct scsi_pkt *pkt)
{
	struct bst_soft		*tgt;
	struct buf		*bp;
	struct bst_pkt_private	*prv;
	int			action;

	prv = (struct bst_pkt_private *)pkt->pkt_private;
	bp = prv->bp;
	tgt = ddi_get_soft_state(bst_state, BST_INST(bp->b_edev));

	BST_LOG(tgt->devp, BST_CE_DEBUG2,
	    "bst_callback: pkt_reason = 0x%x, pkt_flags = 0x%x\n",
	    pkt->pkt_reason, pkt->pkt_flags);

	/*
	 * This mutex locks the tgt struct (as usual!).
	 * We're interested in retry_ct - if we retry, we don't want
	 * the count touched by another thread.
	 */
	mutex_enter(BST_MUTEX(tgt));

	if (pkt->pkt_reason != CMD_CMPLT) {
		/*
		 * The command did not complete. Retry if possible.
		 */
		action = bst_handle_incomplete(tgt, bp);
	} else if (tgt->arq && pkt->pkt_state & STATE_ARQ_DONE) {
		/*
		 * The auto-rqsense happened, and the packet has a
		 * filled-in scsi_arq_status structure, pointed to by
		 * pkt_scbp.
		 */
		action = bst_handle_arq(pkt, tgt, bp);
	} else if (pkt->pkt_flags & FLAG_SENSING) {
		/*
		 * We were running a REQUEST SENSE. Decode the
		 * sense data and decide what to do next.
		 */
		pkt = BST_BP_PKT(bp);	/* get the pkt for the orig command */
		pkt->pkt_flags &= ~FLAG_SENSING;
		action = bst_handle_sense(tgt, bp);
	} else {
		/*
		 * Command completed and we're not getting sense. Check
		 * for errors and decide what to do next.
		 */
		action = bst_check_error(tgt, bp);
	}

	if (tgt->suspended || tgt->dumping) {
		if ((action == QUE_SENSE) || (action == QUE_COMMAND)) {
			/*
			 * return cmd to head of the queue since we don't
			 * want to start more cmds
			 */
			struct diskhd   *dp = &tgt->iorq;

			bp->av_forw = dp->av_forw;
			dp->av_forw = bp;
			action = JUST_RETURN;
		}
	}

	mutex_exit(BST_MUTEX(tgt));

	switch (action) {
	case QUE_SENSE:
		BST_LOG(tgt->devp, BST_CE_DEBUG2, "Getting Sense\n");
		/*
		 * LOCK tgt struct, there's only one sense packet & data for
		 * the unit - don't want to zero sd_sense while someone's
		 * looking at it!
		 */
		mutex_enter(BST_MUTEX(tgt));
		pkt->pkt_flags |= FLAG_SENSING;
		tgt->rqs->pkt_private = (opaque_t)bp;
		bzero((caddr_t)tgt->devp->sd_sense, SENSE_LENGTH);
		mutex_exit(BST_MUTEX(tgt));
		if (scsi_transport(tgt->rqs) == TRAN_ACCEPT) {
			break;
		}
		BST_LOG(tgt->devp, BST_CE_DEBUG1,
		    "Request Sense transport failed\n");
		/*FALLTHROUGH*/
	case COMMAND_DONE_ERROR:
		bp->b_resid = bp->b_bcount;
		bioerror(bp, EIO);
		/*FALLTHROUGH*/
	case COMMAND_DONE:
		BST_LOG(tgt->devp, BST_CE_DEBUG2, "Command Done\n");
		bst_done(tgt, bp);
		break;

	case QUE_COMMAND:
		mutex_enter(BST_MUTEX(tgt));
		if ((int)tgt->retry_ct++ >= bst_retry_count) {
			mutex_exit(BST_MUTEX(tgt));
			BST_LOG(tgt->devp, BST_CE_DEBUG1,
			    "Out of retries, failing command\n");
			bp->b_resid = bp->b_bcount;
			bioerror(bp, EIO);
			bst_done(tgt, bp);
		} else {
			mutex_exit(BST_MUTEX(tgt));
			BST_LOG(tgt->devp, BST_CE_DEBUG2,
			    "Retrying Command\n");
			if (scsi_transport(BST_BP_PKT(bp)) != TRAN_ACCEPT) {
				BST_LOG(tgt->devp, BST_CE_DEBUG1,
				    "Retry transport failed\n");
				bp->b_resid = bp->b_bcount;
				bioerror(bp, EIO);
				bst_done(tgt, bp);
				return;
			}
		}
		break;
	case JUST_RETURN:
		break;
	}

}


/*
 * Incomplete command handling. Figure out what to do based on
 * how far the command did get.
 */
static int
bst_handle_incomplete(struct bst_soft *tgt, struct buf *bp)
{
	int		rval = COMMAND_DONE_ERROR;
	struct scsi_pkt	*pkt = BST_BP_PKT(bp);

	/*
	 * Need the lock, we may be doing retries (retry_ct)
	 * or looking at the sense packet (rqs)
	 */
	ASSERT(mutex_owned(BST_MUTEX(tgt)));

	if (!tgt->arq && (pkt->pkt_flags & FLAG_SENSING)) {
		pkt = tgt->rqs;
	}

	/*
	 * The target may still be running the	command,
	 * so try and reset it, to get it into a known state.
	 * Note: This is forcible, there may be a more polite
	 *	 method for your device.
	 */
	if ((pkt->pkt_statistics &
	    (STAT_BUS_RESET|STAT_DEV_RESET|STAT_ABORTED)) == 0) {
		/*
		 * We can temporarily give up the lock here
		 */
		mutex_exit(BST_MUTEX(tgt));
		BST_LOG(tgt->devp, BST_CE_DEBUG1, "Aborting Command\n");
		if (!(scsi_abort(ROUTE(tgt), (struct scsi_pkt *)NULL))) {
			BST_LOG(tgt->devp, BST_CE_DEBUG1,
			    "Resetting Target\n");
			if (!(scsi_reset(ROUTE(tgt), RESET_TARGET))) {
				BST_LOG(tgt->devp, BST_CE_DEBUG1,
				    "Resetting SCSI Bus\n");
				if (!scsi_reset(ROUTE(tgt), RESET_ALL)) {
					bst_log(tgt->devp, CE_WARN,
					    "SCSI bus reset failed\n");
				}
				mutex_enter(BST_MUTEX(tgt));
				return (COMMAND_DONE_ERROR);
			}
		}
		mutex_enter(BST_MUTEX(tgt));	/* reacquire the lock */
	}

	/*
	 * If we were running a request sense, try it again if possible.
	 * Some devices can handle retries, others will not.
	 */
	if (pkt->pkt_flags & FLAG_SENSING) {
		if ((int)tgt->retry_ct++ < bst_retry_count) {
			rval = QUE_SENSE;
		}
	} else if (bp == tgt->sbufp && (pkt->pkt_flags & FLAG_ISOLATE)) {
		rval = COMMAND_DONE_ERROR;
	} else {
		rval = QUE_COMMAND;	/* retry the command */
	}

	if (pkt->pkt_state == STATE_GOT_BUS && rval == COMMAND_DONE_ERROR) {
		/*
		 * The device is not responding
		 */
		bst_log(tgt->devp, CE_WARN, "Device not responding");
		tgt->state = BST_STATE_NORESP;
	}

	BST_LOG(tgt->devp, BST_CE_DEBUG1, "Cmd incomplete, %s\n",
	    (rval == COMMAND_DONE_ERROR) ? "giving up" : "retrying");

	return (rval);
}


/*
 * Decode sense data
 */
static int
bst_handle_sense(struct bst_soft *tgt, struct buf *bp)
{
	struct diskhd	*dp = &tgt->iorq;
	struct scsi_pkt	*pkt = BST_BP_PKT(bp),
			*rqpkt = tgt->rqs;
	int		rval = COMMAND_DONE_ERROR;
	int		level, amt;

	/*
	 * Need the lock, we may be doing retries (retry_ct)
	 * or looking at the sense packet (rqs) or data
	 * (sd_sense)
	 */
	ASSERT(mutex_owned(BST_MUTEX(tgt)));

	if (SCBP(rqpkt)->sts_busy) {
		/*
		 * Retry the command unless we're out of retries or the
		 * 'fail on error' flag is set
		 */
		if (pkt && !(pkt->pkt_flags & FLAG_DIAGNOSE) &&
		    (int)tgt->retry_ct++ < bst_retry_count) {
			if (dp->b_forw == NULL)
				dp->b_forw = bp;
			tgt->timeout_id = timeout(bst_restart, (caddr_t)tgt,
			    BST_BSY_TIMEOUT);
			rval = JUST_RETURN;
		}
		BST_LOG(tgt->devp, BST_CE_DEBUG1, "Target Busy, %s\n",
		    (rval == JUST_RETURN) ? "restarting" : "giving up");
		return (rval);
	}

	if (SCBP(rqpkt)->sts_chk) {
		BST_LOG(tgt->devp, BST_CE_DEBUG2,
		    "Check Condition on Request Sense!\n");
		return (rval);
	}

	amt = SENSE_LENGTH - rqpkt->pkt_resid;
	if ((rqpkt->pkt_state & STATE_XFERRED_DATA) == 0 || amt == 0) {
		BST_LOG(tgt->devp, BST_CE_DEBUG2, "no sense data\n");
		return (rval);
	}

	/*
	 * Now, check to see whether we got enough sense data to make any
	 * sense out if it (heh-heh).
	 */
	if (amt < SUN_MIN_SENSE_LENGTH) {
		BST_LOG(tgt->devp, BST_CE_DEBUG2,
		    "not enough sense data\n");
		return (rval);
	}

	if (bst_debug > 2) {
		bst_hex_print("bst Sense Data", (caddr_t)tgt->devp->sd_sense,
		    SENSE_LENGTH);
	}

	/*
	 * Decode the sense data
	 * Note: looking only at the sense key here.
	 */
	switch (tgt->devp->sd_sense->es_key) {
	case KEY_NOT_READY:
		/*
		 * If we get a not-ready indication, wait a bit and
		 * try it again, unless this is a special command with
		 * the 'fail on error' (FLAG_DIAGNOSE) option set.
		 */
		if ((bp == tgt->sbufp) &&
		    (pkt->pkt_flags & FLAG_DIAGNOSE)) {
			rval = COMMAND_DONE_ERROR;
			level = SCSI_ERR_FATAL;
		} else if (pkt &&
		    (int)tgt->retry_ct++ < bst_retry_count) {
			if (dp->b_forw == NULL)
				dp->b_forw = bp;
			tgt->timeout_id = timeout(bst_restart, (caddr_t)tgt,
			    BST_BSY_TIMEOUT);
			rval = JUST_RETURN;
			level = SCSI_ERR_RETRYABLE;
		} else {
			rval = COMMAND_DONE_ERROR;
			level = SCSI_ERR_FATAL;
		}
		break;

	case KEY_ABORTED_COMMAND:
	case KEY_UNIT_ATTENTION:
		rval = QUE_COMMAND;
		level = SCSI_ERR_INFO;
		break;
	case KEY_RECOVERABLE_ERROR:
	case KEY_NO_SENSE:
		rval = COMMAND_DONE;
		level = SCSI_ERR_RECOVERED;
		break;
	case KEY_HARDWARE_ERROR:
	case KEY_MEDIUM_ERROR:
	case KEY_MISCOMPARE:
	case KEY_VOLUME_OVERFLOW:
	case KEY_WRITE_PROTECT:
	case KEY_BLANK_CHECK:
	case KEY_ILLEGAL_REQUEST:
	default:
		rval = COMMAND_DONE_ERROR;
		level = SCSI_ERR_FATAL;
		break;
	}

	/*
	 * If this was for a special command, check the options
	 */
	if (bp == tgt->sbufp) {
		if ((rval == QUE_COMMAND) &&
		    (pkt->pkt_flags & FLAG_DIAGNOSE)) {
			rval = COMMAND_DONE_ERROR;
		}
		if (((pkt->pkt_flags & FLAG_SILENT) == 0) || bst_debug) {
			/*
			 * Report the error.
			 * Note: the zero is the error block number.
			 *	You may want to fill it in from the sense data
			 *	for appropriate errors.
			 */
			scsi_errmsg(tgt->devp, pkt, "bst", level,
			    bp->b_blkno, 0, bst_cmds,
			    tgt->devp->sd_sense);
		}
	} else if ((level >= bst_error_reporting) || bst_debug) {
		scsi_errmsg(tgt->devp, pkt, "bst", level,
		    bp->b_blkno, 0, bst_cmds,
		    tgt->devp->sd_sense);
	}

	return (rval);
}


/*
 * Decode auto-rqsense data
 */
static int
bst_handle_arq(struct scsi_pkt *pktp, struct bst_soft *tgt, struct buf *bp)
{
	int		rval = COMMAND_DONE_ERROR;
	struct diskhd	*dp = &tgt->iorq;
	int		level, amt;

	BST_LOG(tgt->devp, BST_CE_DEBUG2, "Auto Request Sense done\n");

	if (ARQP(pktp)->sts_rqpkt_status.sts_chk) {
		BST_LOG(tgt->devp, BST_CE_DEBUG1,
		    "Check Condition on Auto Request Sense!\n");
		return (rval);
	} else if (ARQP(pktp)->sts_rqpkt_reason != CMD_CMPLT) {
		BST_LOG(tgt->devp, BST_CE_DEBUG1,
		    "Auto Request Sense command incomplete\n");
		return (QUE_COMMAND);	/* retry original command */
	}

	amt = SENSE_LENGTH - ARQP(pktp)->sts_rqpkt_resid;
	if ((ARQP(pktp)->sts_rqpkt_state & STATE_XFERRED_DATA) == 0 ||
	    amt == 0) {
		BST_LOG(tgt->devp, BST_CE_DEBUG2,
		    "no auto sense data\n");
		return (rval);
	}

	/*
	 * Stuff the sense data pointer into sd_sense
	 */
	tgt->devp->sd_sense = &(ARQP(pktp)->sts_sensedata);

	/*
	 * Now, check to see whether we got enough sense data to make any
	 * sense out if it (heh-heh).
	 */
	if (amt < SUN_MIN_SENSE_LENGTH) {
		BST_LOG(tgt->devp, BST_CE_DEBUG2,
		    "not enough auto sense data\n");
		return (rval);
	}

	if (bst_debug > 2) {
		bst_hex_print("bst Auto Sense Data",
		    (caddr_t)&(ARQP(pktp)->sts_sensedata), SENSE_LENGTH);
	}

	/*
	 * Decode the sense data
	 * Note: I'm only looking at the sense key here. Most devices
	 *	 have unique additional sense codes & qualifiers, so
	 *	 it's often more useful to look at them instead.
	 *
	 */
	switch (ARQP(pktp)->sts_sensedata.es_key) {
	case KEY_NOT_READY:
		/*
		 * If we get a not-ready indication, wait a bit and
		 * try it again, unless this is a special command with
		 * the 'fail on error' (FLAG_DIAGNOSE) option set.
		 */
		if ((bp == tgt->sbufp) &&
		    (pktp->pkt_flags & FLAG_DIAGNOSE)) {
			rval = COMMAND_DONE_ERROR;
			level = SCSI_ERR_FATAL;
		} else if (tgt->retry_ct++ < bst_retry_count) {
			if (dp->b_forw == NULL)
				dp->b_forw = bp;
			tgt->timeout_id = timeout(bst_restart, (caddr_t)tgt,
			    BST_BSY_TIMEOUT);
			rval = JUST_RETURN;
			level = SCSI_ERR_RETRYABLE;
		} else {
			rval = COMMAND_DONE_ERROR;
			level = SCSI_ERR_FATAL;
		}
		break;

	case KEY_ABORTED_COMMAND:
	case KEY_UNIT_ATTENTION:
		rval = QUE_COMMAND;
		level = SCSI_ERR_INFO;
		break;
	case KEY_RECOVERABLE_ERROR:
	case KEY_NO_SENSE:
		rval = COMMAND_DONE;
		level = SCSI_ERR_RECOVERED;
		break;
	case KEY_HARDWARE_ERROR:
	case KEY_MEDIUM_ERROR:
	case KEY_MISCOMPARE:
	case KEY_VOLUME_OVERFLOW:
	case KEY_WRITE_PROTECT:
	case KEY_BLANK_CHECK:
	case KEY_ILLEGAL_REQUEST:
	default:
		rval = COMMAND_DONE_ERROR;
		level = SCSI_ERR_FATAL;
		break;
	}

	/*
	 * If this was for a special command, check the options
	 */
	if (bp == tgt->sbufp) {
		if ((rval == QUE_COMMAND) &&
		    (pktp->pkt_flags & FLAG_DIAGNOSE)) {
			rval = COMMAND_DONE_ERROR;
		}
		if (((pktp->pkt_flags & FLAG_SILENT) == 0) || bst_debug) {
			/*
			 * Report the error.
			 * Note: the zero is the error block number.
			 *	You may want to fill it in from the sense data
			 *	for appropriate errors.
			 */
			scsi_errmsg(tgt->devp, pktp, "bst", level,
			    bp->b_blkno, 0, bst_cmds,
			    &(ARQP(pktp)->sts_sensedata));
		}
	} else if ((level >= bst_error_reporting) || bst_debug) {
		scsi_errmsg(tgt->devp, pktp, "bst", level,
		    bp->b_blkno, 0, bst_cmds,
		    &(ARQP(pktp)->sts_sensedata));
	}

	return (rval);
}

/*
 * Command completion routine. Check the returned status of the
 * command
 */
static int
bst_check_error(struct bst_soft *tgt, struct buf *bp)
{
	struct diskhd		*dp = &tgt->iorq;
	struct scsi_pkt		*pkt = BST_BP_PKT(bp);
	struct bst_pkt_private	*prv;
	int			action;

	/*
	 * Need the lock, we may be doing retries (retry_ct)
	 * or looking at the sense packet (rqs)
	 */
	ASSERT(mutex_owned(BST_MUTEX(tgt)));

	if (SCBP(pkt)->sts_busy) {
		/*
		 * Target was busy. If we're not out of retries, call
		 * timeout to restart in a bit; otherwise give up and
		 * reset the target. If the fail on error flag is
		 * set, give up immediately.
		 */
		int tval = (pkt->pkt_flags & FLAG_DIAGNOSE) ? 0
							    : BST_BSY_TIMEOUT;

		if (SCBP(pkt)->sts_is) {
			/*
			 * Implicit assumption here is that a device
			 * will only be reserved long enough to
			 * permit a single i/o operation to complete.
			 */
			tval = bst_io_time * drv_usectohz(1000000);
		}

		if ((int)tgt->retry_ct++ < bst_retry_count) {
			if (tval) {
				if (dp->b_forw == NULL)
					dp->b_forw = bp;
				tgt->timeout_id = timeout(bst_restart,
					(caddr_t)tgt, tval);
				action = JUST_RETURN;
				BST_LOG(tgt->devp, BST_CE_DEBUG2,
				    "Target busy, retrying\n");
			} else {
				action = COMMAND_DONE_ERROR;
				BST_LOG(tgt->devp, BST_CE_DEBUG2,
				    "Target busy, no retries\n");
			}
		} else {
			/*
			 * We can temporarily give up the lock here. If
			 * we're pre-empted it's safe for another thread
			 * to get the lock
			 */
			mutex_exit(BST_MUTEX(tgt));
			/*
			 * WARNING: See warning in bst_handle_incomplete
			 */
			BST_LOG(tgt->devp, BST_CE_DEBUG1,
			    "Resetting Target\n");
			if (!scsi_reset(ROUTE(tgt), RESET_TARGET)) {
#ifdef SCSI_BUS_RESET
				BST_LOG(tgt->devp, BST_CE_DEBUG1,
				    "Resetting SCSI Bus\n");
				if (!scsi_reset(ROUTE(tgt), RESET_ALL)) {
					bst_log(tgt->devp, CE_WARN,
					    "SCSI bus reset failed\n");
				}
#else SCSI_BUS_RESET
				bst_log(tgt->devp, CE_WARN,
				    "Reset Target failed\n");
#endif SCSI_BUS_RESET
			}
			action = COMMAND_DONE_ERROR;
			mutex_enter(BST_MUTEX(tgt));	/* reacquire the lock */
		}
	} else if (SCBP(pkt)->sts_chk) {
		if (tgt->arq) {
			bst_log(tgt->devp, CE_WARN,
			    "Check condition with Auto Sensing\n");
			action = COMMAND_DONE_ERROR;
		} else {
			BST_LOG(tgt->devp, BST_CE_DEBUG2,
			    "Check Condition\n");
			action = QUE_SENSE;	/* check condition: get sense */
		}
	} else {
		/*
		 * Extract the 'command' bits out of the SCSI CDB
		 */
		int cmd = (pkt->pkt_cdbp[0]) & 0x1f;

		action = COMMAND_DONE;	/* assume command completed OK */

		/*
		 * pkt_resid will reflect, at this point, a residual
		 * of how many bytes were not transferred; a non-zero
		 * pkt_resid on a read or a write is an error.
		 * Note	I'm doing this because format asks for too much
		 *	data, so there's always a resid on the USCSICMDs
		 *	it sends; if your device uses different commands
		 *	for normal data transfer, test for them here.
		 */
		if (pkt->pkt_resid && (cmd == SCMD_READ || cmd == SCMD_WRITE)) {
			action = COMMAND_DONE_ERROR;
			bp->b_resid += pkt->pkt_resid;
			prv = (struct bst_pkt_private *)pkt->pkt_private;
			bp->b_resid += prv->dma_resid;
			BST_LOG(tgt->devp, BST_CE_DEBUG1,
			    "Residual data (%d bytes)\n", bp->b_resid);
		}
		tgt->retry_ct = 0;
		BST_LOG(tgt->devp, BST_CE_DEBUG2, "Command Complete\n");
	}
	return (action);
}


/*
 * Read and validate the disk's label and geometry
 * Don't need mutexes in this routine because it's is called only from
 * attach and open, both of which are single threaded and can't run
 * concurrently.
 */
static void
bst_checklabel(struct bst_soft *tgt)
{
	struct dk_label	*dkl;
	int		secsize;
	int		label_addr;
	int		secdiv;
	struct buf	*bp;
	int		(*f)() = SLEEP_FUNC;
	int		offset;
	int		i;

	ASSERT(mutex_owned(BST_MUTEX(tgt)));

	/*
	 * take a log base 2 of sector size (sorry)
	 */
	secsize = tgt->secsize;
	for (secdiv = 0; secsize = secsize >> 1; secdiv++)
		;
	tgt->secdiv = secdiv;

	tgt->blkshf = secdiv - DEV_BSHIFT;

	/*
	 * Is the disk a cd-rom?
	 */
	if (INQUIRY(tgt)->inq_dtype == DTYPE_RODIRECT) {
		tgt->iscd = 1;
	}

	/*
	 * Only DIRECT ACCESS devices will have Sun labels.
	 * CD's may have a Sun label, too
	 */
	if (INQUIRY(tgt)->inq_dtype == DTYPE_DIRECT || tgt->iscd) {
		if (bst_read_fdisk(tgt, f))
			return;
	}

	label_addr = (tgt->fdisk_ok == 0) ? 0 :
		tgt->solaris_offset + DK_LABEL_LOC;

	if (INQUIRY(tgt)->inq_dtype == DTYPE_DIRECT || tgt->iscd) {
		bp = scsi_alloc_consistent_buf(ROUTE(tgt),
		    (struct buf *)NULL, tgt->secsize, B_READ, f, NULL);
		if (!bp) {
			scsi_log(DEVINFO(tgt), bst_label, CE_WARN,
			    "no bp for disk label\n");
			return;
		}

		offset = bst_iopb_read_block(tgt, label_addr, bp, f);
		if (offset != -1) {
			dkl = (struct dk_label *)bp->b_un.b_addr + offset;
			bst_uselabel(tgt, dkl);
		}

		scsi_free_consistent_buf(bp);

	} else if (tgt->capacity < 0) {
		return;
	}

	/*
	 * Special exception for disks without a valid label.
	 */
	if (tgt->geom_ok == 0 && tgt->capacity > 0) {
		/*
		 * initialize the geometry and partition table.
		 */
		bzero((caddr_t)&tgt->geom, sizeof (struct dk_geom));
		bzero((caddr_t)&tgt->vtoc, sizeof (struct dk_vtoc));
		bzero((caddr_t)tgt->map,
		    BST_NMAP * (sizeof (struct partition)));
		tgt->asciilabel[0] = '\0';

		if (tgt->iscd) {
			tgt->geom.dkg_ncyl = 1;
			tgt->geom.dkg_acyl = 0;
			tgt->geom.dkg_nhead = 1;
			tgt->geom.dkg_nsect = 0;
			tgt->geom.dkg_intrlv = 1;
			tgt->geom.dkg_rpm = 200;
			tgt->map['a'-'a'].p_start = 0;
			tgt->map['a'-'a'].p_size = tgt->capacity;
			tgt->map['c'-'a'].p_start = 0;
			tgt->map['c'-'a'].p_size = tgt->capacity;
		} else {
			tgt->geom.dkg_pcyl = tgt->capacity >> 5;
			tgt->geom.dkg_ncyl = tgt->geom.dkg_pcyl - 2;
			tgt->geom.dkg_acyl = 2;
			tgt->geom.dkg_nhead = 1;
			tgt->geom.dkg_nsect = 32;
			tgt->geom.dkg_intrlv = 1;
			tgt->geom.dkg_rpm = 3000;
			tgt->map['c'-'a'].p_start = 0;
			tgt->map['c'-'a'].p_size = tgt->capacity;
			bst_inq_fill(INQUIRY(tgt)->inq_vid, 8,
			    tgt->asciilabel);
			scsi_log(DEVINFO(tgt), bst_label, CE_CONT,
			    "<%s %d blocks>\n", tgt->asciilabel,
			    (int)tgt->capacity);
		}
		tgt->geom_ok = 1;
	}

	/*
	 * Set up offset to each fdisk partition/solaris slice
	 */
	if (tgt->geom_ok == 1) {
		for (i = 0; i < NDKMAP; i++) {
			if (tgt->map[i].p_size > 0) {
				tgt->offset[i] = tgt->map[i].p_start +
				    tgt->solaris_offset;
			}
		}

#if defined(_SUNOS_VTOC_16)
		/*
		 * Set up the "whole disk" fdisk partitions
		 */
		tgt->map[FDISK_MAP_ALL].p_start = 0;
		tgt->map[FDISK_MAP_ALL].p_size = tgt->capacity;
		tgt->offset[FDISK_MAP_ALL] = 0;

		/*
		 * If fdisk is valid, set up the remaining
		 * fdisk partitions.
		 */
		if (tgt->fdisk_ok) {
			for (i = 0; i < FD_NUMPART; i++) {
				tgt->map[FDISK_MAP0 + i].p_start =
					tgt->fmap[i].p_start;
				tgt->map[FDISK_MAP0 + i].p_size =
					tgt->fmap[i].p_size;
				tgt->offset[FDISK_MAP0 + i] =
					tgt->fmap[i].p_start;
			}
		}
#endif
	}

#ifdef	DEBUG
	if (bst_debug > 1) {
		cmn_err(CE_CONT, "map:\n");
		cmn_err(CE_CONT, "	  start      size"
		    "	   offset\n");
		for (i = 0; i < BST_NMAP; i++) {
			cmn_err(CE_CONT, "%2d:	%ld  %ld  %d\n",
			    i, tgt->map[i].p_start, tgt->map[i].p_size,
			    tgt->offset[i]);
		}
	}
#endif	DEBUG
}


/*
 * Check for a valid label
 * Marks the geometry of the unit as being valid.
 */

static void
bst_uselabel(struct bst_soft *tgt, struct dk_label *l)
{
	short			*sp, sum, count;
	int			capacity;
	int			i;
#if defined(_SUNOS_VTOC_8)
	int			nblks;
#endif

	ASSERT(mutex_owned(BST_MUTEX(tgt)));

	/*
	 * Check magic number of the label
	 */
	if (l->dkl_magic != DKL_MAGIC) {
		if (tgt->state == BST_STATE_NIL && !REMOVABLE(tgt))
			scsi_log(DEVINFO(tgt), bst_label, CE_WARN,
			    "label invalid - wrong magic number");
		return;
	}

	/*
	 * Check the checksum of the label,
	 */
	sp = (short *)l;
	sum = 0;
	count = sizeof (struct dk_label) / sizeof (short);
	while (count--)	 {
		sum ^= *sp++;
	}

	if (sum) {
		if (tgt->state == BST_STATE_NIL && !REMOVABLE(tgt)) {
			scsi_log(DEVINFO(tgt), bst_label, CE_WARN,
			    "label invalid - label checksum failed");
		}
		return;
	}

	/*
	 * Fill in disk geometry from label.
	 */
	tgt->geom.dkg_ncyl = l->dkl_ncyl;
	tgt->geom.dkg_acyl = l->dkl_acyl;
	tgt->geom.dkg_bcyl = 0;
	tgt->geom.dkg_nhead = l->dkl_nhead;
	tgt->geom.dkg_nsect = l->dkl_nsect;
	tgt->geom.dkg_intrlv = l->dkl_intrlv;
	tgt->geom.dkg_pcyl = l->dkl_pcyl;
	tgt->geom.dkg_rpm = l->dkl_rpm;

#if defined(_SUNOS_VTOC_8)
	tgt->geom.dkg_obs1 = l->dkl_obs1;
	tgt->geom.dkg_obs2 = l->dkl_obs2;
	tgt->geom.dkg_obs3 = l->dkl_obs3;
#endif

	tgt->geom.dkg_read_reinstruct = l->dkl_read_reinstruct;
	tgt->geom.dkg_write_reinstruct = l->dkl_write_reinstruct;

	if (tgt->geom.dkg_pcyl == 0)
		tgt->geom.dkg_pcyl = tgt->geom.dkg_ncyl + tgt->geom.dkg_acyl;

	/*
	 * Fill in partition table.
	 */

#if defined(_SUNOS_VTOC_8)
	nblks = l->dkl_nhead * l->dkl_nsect;
	for (i = 0; i < NDKMAP; i++) {
		tgt->map[i].p_tag = l->dkl_vtoc.v_part[i].p_tag;
		tgt->map[i].p_flag = l->dkl_vtoc.v_part[i].p_flag;
		tgt->map[i].p_start = l->dkl_map[i].dkl_cylno * nblks;
		tgt->map[i].p_size = l->dkl_map[i].dkl_nblk;
	}
#endif

#if defined(_SUNOS_VTOC_16)
	for (i = 0; i < NDKMAP; i++) {
		tgt->map[i].p_tag = l->dkl_vtoc.v_part[i].p_tag;
		tgt->map[i].p_flag = l->dkl_vtoc.v_part[i].p_flag;
		tgt->map[i].p_start = l->dkl_vtoc.v_part[i].p_start;
		tgt->map[i].p_size = l->dkl_vtoc.v_part[i].p_size;
	}
#endif

	/*
	 * Fill in VTOC Structure.
	 */
	bcopy((caddr_t)l->dkl_vtoc.v_bootinfo,
	    (caddr_t)tgt->vtoc.v_bootinfo,
	    sizeof (l->dkl_vtoc.v_bootinfo));

	tgt->vtoc.v_sanity = l->dkl_vtoc.v_sanity;
	tgt->vtoc.v_version = l->dkl_vtoc.v_version;

	bcopy((caddr_t)l->dkl_vtoc.v_volume,
	    (caddr_t)tgt->vtoc.v_volume, LEN_DKL_VVOL);

	tgt->vtoc.v_nparts = l->dkl_vtoc.v_nparts;

	bcopy((caddr_t)l->dkl_vtoc.v_reserved,
	    (caddr_t)tgt->vtoc.v_reserved, sizeof (l->dkl_vtoc.v_reserved));

	bcopy((caddr_t)l->dkl_vtoc.v_part, (caddr_t)tgt->vtoc.v_part,
	    sizeof (l->dkl_vtoc.v_part));

	bcopy((caddr_t)l->dkl_vtoc.v_timestamp,
	    (caddr_t)tgt->vtoc.v_timestamp, sizeof (l->dkl_vtoc.v_timestamp));

	tgt->geom_ok = 1;

	capacity = (tgt->geom.dkg_ncyl + tgt->geom.dkg_acyl) *
		(tgt->geom.dkg_nhead * tgt->geom.dkg_nsect);

	if (tgt->capacity > 0) {
		if (capacity > tgt->capacity &&
		    tgt->state == BST_STATE_NIL) {
			if (tgt->iscd) { /* trust the label */
				tgt->capacity = capacity;
			} else {
				scsi_log(DEVINFO(tgt), bst_label, CE_CONT,
				    "Label says %d blocks,"
				    "Drive says %d blocks\n",
				    capacity, tgt->capacity);
				tgt->geom_ok = 0;
			}
		} else {
			tgt->capacity = capacity;
		}
	} else {
		/*
		 * We have a situation where the target didn't give us a
		 * good 'read capacity' command answer, yet there appears
		 * to be a valid label. In this case, we'll
		 * fake the capacity.
		 */
		tgt->capacity = capacity;
	}

	bcopy((caddr_t)l->dkl_asciilabel,
	    (caddr_t)tgt->asciilabel, LEN_DKL_ASCII);
}



/*
 * Data returned by direct access devices in response
 * to the Read Capacity SCSI-2 command.
 */
struct scsi_read_cap {
	u_char	capacity[4];		/* number of addressable blocks */
	u_char	lbasize[4];		/* logical block size */
};

static void
bst_read_capacity(struct bst_soft *tgt, dev_t dev)
{
	struct uscsi_cmd scmd, *com = &scmd;
	struct scsi_read_cap	*cap;
	char	cmdblk[CDB_GROUP1];
	int	err;
	uint_t	blksiz;
	uint_t	capacity;

	ASSERT(mutex_owned(BST_MUTEX(tgt)));
	mutex_exit(BST_MUTEX(tgt));

	cap = kmem_alloc((size_t)(sizeof (struct scsi_read_cap)), KM_SLEEP);

	com->uscsi_bufaddr = (caddr_t)cap;
	com->uscsi_buflen = sizeof (struct scsi_read_cap);
	bzero(cmdblk, CDB_GROUP1);
	cmdblk[0] = SCMD_READ_CAPACITY;
	com->uscsi_flags = USCSI_SILENT | USCSI_READ;
	com->uscsi_cdb = cmdblk;
	com->uscsi_cdblen = CDB_GROUP1;

	if (err = bst_ioctl_cmd(dev, com, UIO_SYSSPACE, UIO_SYSSPACE)) {
		BST_LOG(SCSIDEV(tgt), BST_CE_DEBUG1,
			"bst_read_capacity failed: 0x%x\n", err);
		kmem_free(cap, (size_t)(sizeof (struct scsi_read_cap)));
		mutex_enter(BST_MUTEX(tgt));
		return;
	}

	mutex_enter(BST_MUTEX(tgt));

	capacity = (cap->capacity[0] << 24 & 0xff000000) |
		    (cap->capacity[1] << 16 & 0xff0000) |
		    (cap->capacity[2] << 8 & 0xff00) |
		    (cap->capacity[3] & 0xff);
	blksiz = (cap->lbasize[0] << 24 & 0xff000000) |
		    (cap->lbasize[1] << 16 & 0xff0000) |
		    (cap->lbasize[2] << 8 & 0xff00) |
		    (cap->lbasize[3] & 0xff);

	tgt->capacity = capacity;
	tgt->secsize = blksiz;

	mutex_exit(BST_MUTEX(tgt));
	(void) scsi_ifsetcap(ROUTE(tgt), "sector-size", blksiz, 1);
	(void) scsi_ifsetcap(ROUTE(tgt), "total-sectors", capacity, 1);
	mutex_enter(BST_MUTEX(tgt));
	kmem_free(cap, (size_t)(sizeof (struct scsi_read_cap)));

	BST_LOG(SCSIDEV(tgt), BST_CE_DEBUG1, "capacity %d, block size %d\n",
	    capacity, blksiz);
}


static void
bst_chgmap(struct bst_soft *tgt, struct dk_map *map)
{
	struct partition	*pp;
	daddr_t			nblks;
	int			i;

	nblks = tgt->geom.dkg_nhead * tgt->geom.dkg_nsect;

	pp = tgt->map;
	for (i = 0; i < NDKMAP; i++, pp++) {
		pp->p_start = map[i].dkl_cylno * nblks;
		pp->p_size  = map[i].dkl_nblk;
	}
}


static void
bst_getmap(struct bst_soft *tgt, struct dk_map *map)
{
	struct partition	*pp;
	daddr_t			nblks;
	int			i;

	nblks = tgt->geom.dkg_nhead * tgt->geom.dkg_nsect;

	pp = tgt->map;
	for (i = 0; i < NDKMAP; i++, pp++) {
		map[i].dkl_cylno = pp->p_start / nblks;
		map[i].dkl_nblk  = pp->p_size;
	}
}



/*
 * See if the device is ready. Send a couple of Test Unit Ready commands
 * to clear Unit Attention conditions. If it's still not ready, send the
 * Start command and try another Test Unit Ready
 * Return 1 is the unit's ready, 0 if not.
 */
static int
bst_ready(struct bst_soft *tgt, dev_t dev)
{
	int retries = 3;

	while (--retries) {
		if (bst_simple(dev, SCMD_TEST_UNIT_READY, 0)) {
			break;
		}
	}

	if (retries > 0) {
		BST_LOG(0, BST_CE_DEBUG2, "bst%d is ready\n", BST_INST(dev));
		mutex_enter(BST_MUTEX(tgt));
		tgt->state = BST_STATE_NIL;
		mutex_exit(BST_MUTEX(tgt));
		return (1);
	}

	/*
	 * Not Ready, try a Start command
	 */
	BST_LOG(0, BST_CE_DEBUG3, "starting bst%d\n", BST_INST(dev));
	if (bst_simple(dev, SCMD_START_STOP, BST_START) == 0) {
		BST_LOG(0, BST_CE_DEBUG1, "unable to start bst%d\n",
		    BST_INST(dev));
		return (0);
	}

	/*
	 * Started OK, try another Test Unit Ready
	 */
	if (bst_simple(dev, SCMD_TEST_UNIT_READY, 0) == 0) {
		BST_LOG(0, BST_CE_DEBUG1, "bst%d not ready\n", BST_INST(dev));
		return (0);
	}

	BST_LOG(0, BST_CE_DEBUG2, "bst%d is ready\n", BST_INST(dev));
	mutex_enter(BST_MUTEX(tgt));
	tgt->state = BST_STATE_NIL;
	mutex_exit(BST_MUTEX(tgt));
	return (1);	/* Unit is ready */
}


/*
 * Send a simple (no DMA) command to the device
 * return 1 if the command succeeds, 0 if it fails.
 */
static int
bst_simple(dev_t dev, int cmd, int flag)
{
	struct uscsi_cmd scmd, *com = &scmd;
	char	cmdblk[CDB_GROUP0];
	int	err;

	com->uscsi_bufaddr = 0;
	com->uscsi_buflen = 0;
	bzero(cmdblk, CDB_GROUP0);
	cmdblk[0] = (char)cmd;
	cmdblk[4] = flag;
	com->uscsi_flags = USCSI_DIAGNOSE|USCSI_SILENT|USCSI_WRITE;
	com->uscsi_cdb = cmdblk;
	com->uscsi_cdblen = CDB_GROUP0;

	if (err = bst_ioctl_cmd(dev, com, UIO_SYSSPACE, UIO_SYSSPACE)) {
		BST_LOG(0, BST_CE_DEBUG2, "bst_simple failed: 0x%x\n", err);
		err = 0;
	} else {
		err = 1;
	}
	return (err);
}


/*
 * Convert the label's struct dk_vtoc (<sys/dklabel.h>) to a user
 * struct vtoc (<sys/vtoc.h>).
 */
static void
bst_dkioc_gvtoc(struct bst_soft *tgt, struct vtoc *vtoc)
{
	int i;
	struct partition	*vp;
	struct partition	*mp;

	ASSERT(mutex_owned(BST_MUTEX(tgt)));

	/*
	 * Return vtoc structure fields in the provided VTOC area,
	 * addressed by *vtoc.
	 */

	bzero((caddr_t)vtoc, sizeof (struct vtoc));

	bcopy((caddr_t)tgt->vtoc.v_bootinfo, (caddr_t)vtoc->v_bootinfo,
	    sizeof (vtoc->v_bootinfo));

	vtoc->v_sanity		= VTOC_SANE;
	vtoc->v_version		= tgt->vtoc.v_version;

	bcopy((caddr_t)tgt->vtoc.v_volume, (caddr_t)vtoc->v_volume,
	    LEN_DKL_VVOL);

	vtoc->v_sectorsz = tgt->secsize;
	vtoc->v_nparts = tgt->vtoc.v_nparts;

	bcopy((caddr_t)tgt->vtoc.v_reserved, (caddr_t)vtoc->v_reserved,
	    sizeof (vtoc->v_reserved));

	/*
	 * Convert partitioning information.
	 */
	mp = tgt->map;
	vp = vtoc->v_part;

	for (i = 0; i < V_NUMPAR; i++, mp++, vp++) {
		vp->p_tag	= mp->p_tag;
		vp->p_flag	= mp->p_flag;
		vp->p_start	= mp->p_start;
		vp->p_size	= mp->p_size;
	}

	bcopy((caddr_t)tgt->vtoc.v_timestamp, (caddr_t)vtoc->timestamp,
	    sizeof (vtoc->timestamp));

	bcopy((caddr_t)tgt->asciilabel, (caddr_t)vtoc->v_asciilabel,
	    LEN_DKL_ASCII);
}


/*
 * Convert a user struct vtoc (<sys/vtoc.h>) to the
 * label's struct dk_vtoc (<sys/dklabel.h>).
 */
static int
bst_dkioc_svtoc(struct vtoc *vtoc, struct bst_soft *tgt)
{
	struct partition	*vp;
	struct partition	*mp;
	int			nblks;
	int			ncyl;
	int			i;

	ASSERT(mutex_owned(BST_MUTEX(tgt)));

	/*
	 * Sanity-check the vtoc
	 */
	if (vtoc->v_sanity != VTOC_SANE || vtoc->v_nparts != V_NUMPAR ||
	    vtoc->v_sectorsz != tgt->secsize) {
		return (EINVAL);
	}

	if ((nblks = tgt->geom.dkg_nsect * tgt->geom.dkg_nhead) == 0) {
		return (EINVAL);
	}

	vp = vtoc->v_part;
	for (i = 0; i < V_NUMPAR; i++, vp++) {
		if ((vp->p_start % nblks) != 0)
			return (EINVAL);
		ncyl = vp->p_start / nblks;
		ncyl += vp->p_size / nblks;
		if ((vp->p_size % nblks) != 0)
			ncyl++;
		if (ncyl > (int)tgt->geom.dkg_pcyl) {
			return (EINVAL);
		}
	}

	/*
	 * Put appropriate vtoc structure fields into state structure
	 */
	mp = tgt->map;
	vp = vtoc->v_part;
	for (i = 0; i < (int)vtoc->v_nparts; i++, mp++, vp++) {
		mp->p_tag  = vp->p_tag;
		mp->p_flag = vp->p_flag;
		mp->p_start = vp->p_start;
		mp->p_size = vp->p_size;
	}

	bcopy((caddr_t)vtoc->v_bootinfo, (caddr_t)tgt->vtoc.v_bootinfo,
	    sizeof (vtoc->v_bootinfo));

	tgt->vtoc.v_sanity = vtoc->v_sanity;
	tgt->vtoc.v_version = vtoc->v_version;

	bcopy((caddr_t)vtoc->v_volume, (caddr_t)tgt->vtoc.v_volume,
	    LEN_DKL_VVOL);

	tgt->vtoc.v_nparts = vtoc->v_nparts;

	bcopy((caddr_t)vtoc->v_reserved, (caddr_t)tgt->vtoc.v_reserved,
	    sizeof (vtoc->v_reserved));

	bcopy((caddr_t)vtoc->timestamp, (caddr_t)tgt->vtoc.v_timestamp,
	    sizeof (vtoc->timestamp));

	bcopy((caddr_t)vtoc->v_asciilabel, (caddr_t)tgt->asciilabel,
	    LEN_DKL_ASCII);

	return (0);
}


/*
 * Disk geometry macros
 *
 *	spc:		sectors per cylinder
 *	chs2bn:		cyl/head/sector to block number
 */
#define	spc(l)		(((l)->dkl_nhead*(l)->dkl_nsect)-(l)->dkl_apc)

#define	chs2bn(l, c, h, s)	\
			((daddr_t)((c)*spc(l)+(h)*(l)->dkl_nsect+(s)))


static int
bst_write_label(dev_t dev)
{
	int			i, status;
	int			sec, blk, head, cyl;
	short			sum, *sp;
	struct bst_soft		*tgt;
	caddr_t			buffer;
	struct dk_label		*dkl;
	u_int			label_addr;
	struct buf		*bp;
	struct buf		*bp2;
	struct partition	*mp;
	int			offset;
#if defined(_SUNOS_VTOC_8)
	int			nblks;
#endif

	tgt = ddi_get_soft_state(bst_state, BST_INST(dev));
	if (tgt == NULL)
		return (ENXIO);

	ASSERT(mutex_owned(BST_MUTEX(tgt)));

	buffer = kmem_zalloc(sizeof (struct dk_label), KM_SLEEP);
	dkl = (struct dk_label *)buffer;
	bzero((caddr_t)dkl, sizeof (struct dk_label));

	bcopy((caddr_t)tgt->asciilabel, (caddr_t)dkl->dkl_asciilabel,
	    LEN_DKL_ASCII);

	dkl->dkl_vtoc.v_sanity = VTOC_SANE;
	dkl->dkl_vtoc.v_version = tgt->vtoc.v_version;
	dkl->dkl_vtoc.v_nparts = tgt->vtoc.v_nparts;

	bcopy((caddr_t)tgt->vtoc.v_volume,
	    (caddr_t)dkl->dkl_vtoc.v_volume, LEN_DKL_VVOL);
	bcopy((caddr_t)tgt->vtoc.v_reserved,
	    (caddr_t)dkl->dkl_vtoc.v_reserved,
	    sizeof (tgt->vtoc.v_reserved));
	bcopy((caddr_t)tgt->vtoc.v_bootinfo,
	    (caddr_t)dkl->dkl_vtoc.v_bootinfo,
	    sizeof (tgt->vtoc.v_bootinfo));
	bcopy((caddr_t)tgt->vtoc.v_timestamp,
	    (caddr_t)dkl->dkl_vtoc.v_timestamp,
	    sizeof (tgt->vtoc.v_timestamp));

	dkl->dkl_rpm	= tgt->geom.dkg_rpm;
	dkl->dkl_pcyl	= tgt->geom.dkg_pcyl;
	dkl->dkl_apc	= tgt->geom.dkg_apc;
	dkl->dkl_intrlv = tgt->geom.dkg_intrlv;
	dkl->dkl_ncyl	= tgt->geom.dkg_ncyl;
	dkl->dkl_acyl	= tgt->geom.dkg_acyl;
	dkl->dkl_nhead	= tgt->geom.dkg_nhead;
	dkl->dkl_nsect	= tgt->geom.dkg_nsect;

#if defined(_SUNOS_VTOC_8)
	dkl->dkl_obs1	= tgt->geom.dkg_obs1;
	dkl->dkl_obs2	= tgt->geom.dkg_obs2;
	dkl->dkl_obs3	= tgt->geom.dkg_obs3;
#endif

	dkl->dkl_magic			= DKL_MAGIC;
	dkl->dkl_write_reinstruct	= tgt->geom.dkg_write_reinstruct;
	dkl->dkl_read_reinstruct	= tgt->geom.dkg_read_reinstruct;

	/*
	 * Note the conversion from starting sector number
	 * to starting cylinder number.
	 */
	mp = tgt->map;
#if defined(_SUNOS_VTOC_8)
	nblks = tgt->geom.dkg_nsect * tgt->geom.dkg_nhead;
#endif

	for (i = 0; i < NDKMAP; i++, mp++) {
		dkl->dkl_vtoc.v_part[i].p_tag  = mp->p_tag;
		dkl->dkl_vtoc.v_part[i].p_flag = mp->p_flag;
#if defined(_SUNOS_VTOC_8)
		dkl->dkl_map[i].dkl_cylno = mp->p_start / nblks;
		dkl->dkl_map[i].dkl_nblk = mp->p_size;
#endif
#if defined(_SUNOS_VTOC_16)
		dkl->dkl_vtoc.v_part[i].p_start = mp->p_start;
		dkl->dkl_vtoc.v_part[i].p_size = mp->p_size;
#endif
	}

	/*
	 * Construct checksum for the new disk label
	 */
	sum = 0;
	sp = (short *)dkl;
	i = sizeof (struct dk_label)/sizeof (short);
	while (i--) {
		sum ^= *sp++;
	}
	dkl->dkl_cksum = sum;

	/*
	 * Get resources to read existing data
	 */
	bp = scsi_alloc_consistent_buf(ROUTE(tgt),
	    (struct buf *)NULL, tgt->secsize, B_READ, SLEEP_FUNC, NULL);
	bp2 = scsi_alloc_consistent_buf(ROUTE(tgt),
	    (struct buf *)NULL, tgt->secsize, B_WRITE, SLEEP_FUNC, NULL);
	ASSERT(bp != NULL && bp2 != NULL);

	/*
	 * Write the primary label at block 0 of the solaris partition.
	 */
	if (tgt->fdisk_ok) {
		label_addr = tgt->solaris_offset + DK_LABEL_LOC;
	} else {
		label_addr = 0;
	}

	offset = bst_iopb_read_block(tgt, label_addr, bp, SLEEP_FUNC);
	if (offset == -1) {
		scsi_free_consistent_buf(bp);
		scsi_free_consistent_buf(bp2);
		kmem_free(buffer, sizeof (struct dk_label));
		return (EIO);
	}

	/*
	 * Write out new label
	 */
	bcopy(bp->b_un.b_addr, bp2->b_un.b_addr, tgt->secsize);
	bcopy((caddr_t)dkl, bp2->b_un.b_addr + offset,
		sizeof (struct dk_label));

	status = bst_iopb_write_block(tgt, label_addr, bp2, SLEEP_FUNC);
	if (status) {
		scsi_free_consistent_buf(bp);
		scsi_free_consistent_buf(bp2);
		kmem_free(buffer, sizeof (struct dk_label));
		return (EIO);
	}

	/*
	 * Calculate where the backup labels go.  They are always on
	 * the last alternate cylinder, but some older drives put them
	 * on head 2 instead of the last head.	They are always on the
	 * first 5 odd sectors of the appropriate track.
	 *
	 * We have no choice at this point, but to believe that the
	 * disk label is valid.	 Use the geometry of the disk
	 * as described in the label.
	 */
	cyl = dkl->dkl_ncyl + dkl->dkl_acyl - 1;
	head = dkl->dkl_nhead-1;

	/*
	 * Write and verify the backup labels.
	 */
	for (sec = 1; sec < 5 * 2 + 1; sec += 2) {
		blk = chs2bn(dkl, cyl, head, sec);
		if (tgt->fdisk_ok)
			blk += tgt->solaris_offset;

		offset = bst_iopb_read_block(tgt, blk, bp, SLEEP_FUNC);
		if (offset == -1) {
			scsi_free_consistent_buf(bp);
			scsi_free_consistent_buf(bp2);
			kmem_free(buffer, sizeof (struct dk_label));
			return (EIO);
		}

		/*
		 * Write out new label
		 */
		bcopy(bp->b_un.b_addr, bp2->b_un.b_addr, tgt->secsize);
		bcopy((caddr_t)dkl, bp2->b_un.b_addr + offset,
			sizeof (struct dk_label));

		status = bst_iopb_write_block(tgt, blk, bp2, SLEEP_FUNC);
		if (status) {
			scsi_free_consistent_buf(bp);
			scsi_free_consistent_buf(bp2);
			kmem_free(buffer, sizeof (struct dk_label));
			return (EIO);
		}
	}

	scsi_free_consistent_buf(bp);
	scsi_free_consistent_buf(bp2);
	kmem_free(buffer, sizeof (struct dk_label));
	return (0);
}


static int
bst_read_fdisk(
	struct bst_soft		*tgt,
	int			(*f)())
{
	struct buf		*bp;
	struct ipart		*fdp;
	struct mboot		*mbp;
	struct ipart		fdisk[FD_NUMPART];
	int			offset;
	int			i;

	/*
	 * Invalidate fdisk info
	 */
	tgt->fdisk_ok = 0;
	tgt->part_type = -1;
	tgt->active_part = 0;
	tgt->solaris_offset = 0;
	tgt->solaris_size = 0;
	bzero((caddr_t)tgt->fmap, sizeof (struct partition) * FD_NUMPART);

	bp = scsi_alloc_consistent_buf(ROUTE(tgt),
	    (struct buf *)NULL, tgt->secsize, B_READ, f, NULL);
	if (!bp) {
		scsi_log(DEVINFO(tgt), bst_label, CE_WARN,
		    "no bp for fdisk\n");
		return (1);
	}

	offset = bst_iopb_read_block(tgt, 0, bp, f);
	mbp = (struct mboot *)bp->b_un.b_addr + offset;

#ifdef	DEBUG
	if (bst_debug > 1) {
		dump("fdisk:\n", (u_char *)bp->b_un.b_addr + BOOTSZ,
			DEV_BSIZE - BOOTSZ);
	}
#endif

	if (offset == -1 || mbp->signature != MBB_MAGIC) {
		scsi_free_consistent_buf(bp);
		return (0);
	}

	/*
	 * The fdisk table does not begin on a 4-byte
	 * boundary within the master boot record, so we
	 * copy it to an aligned structure to avoid
	 * alignment exceptions on some processors.
	 */
	bcopy((caddr_t)&mbp->parts[0], (caddr_t)fdisk, sizeof (fdisk));

#ifdef	DEBUG
	if (bst_debug > 0) {
		fdp = fdisk;
		cmn_err(CE_CONT, "fdisk:\n");
		cmn_err(CE_CONT, "	   relsect    "
			"numsect	 sysid	     bootid\n");
		for (i = 0; i < FD_NUMPART; i++, fdp++) {
			cmn_err(CE_CONT,
				"    %d:  %d	%d	0x%x	   0x%x\n",
				i, (int)fdp->relsect, (int)fdp->numsect,
				(int)fdp->systid, (int)fdp->bootid);
		}
	}
#endif

	/*
	 * Try to find the unix partition
	 */
	fdp = fdisk;
	for (i = 0; i < FD_NUMPART; i++, fdp++) {
		if (fdp->numsect == 0)
			continue;
		tgt->fmap[i].p_start = fdp->relsect;
		tgt->fmap[i].p_size = fdp->numsect;
		if ((fdp->systid == UNIXOS || fdp->systid == SUNIXOS) &&
				fdp->numsect > 0) {
			tgt->fdisk_ok = 1;
			tgt->active_part = i;
			tgt->part_type = fdp->systid;
			tgt->solaris_offset = fdp->relsect;
		}
	}

	scsi_free_consistent_buf(bp);
	return (0);
}



static int
bst_iopb_read_block(
	struct bst_soft			*tgt,
	daddr_t				blkno,
	struct buf			*bp,
	int				(*f)())
{
	register struct scsi_pkt	*pkt;
	register daddr_t		addr;
	int				offset;
	int				rval = -1;

	ASSERT(mutex_owned(BST_MUTEX(tgt)));
	ASSERT(f == NULL_FUNC || f == SLEEP_FUNC);

	addr = blkno >> tgt->blkshf;
	offset = (blkno & ((1 << tgt->blkshf) - 1)) << DEV_BSHIFT;

	pkt = scsi_init_pkt(ROUTE(tgt), (struct scsi_pkt *)NULL,
		bp, CDB_GROUP1, 1, 0, PKT_CONSISTENT, f, NULL);
	if (!pkt) {
		return (rval);
	}

	bzero((caddr_t)bp->b_un.b_addr,  tgt->secsize);

	(void) scsi_setup_cdb((union scsi_cdb *)pkt->pkt_cdbp,
			SCMD_READ_G1, addr, 1, 0);
	pkt->pkt_flags = 0;
	BST_DUMP_CDB(tgt, pkt, CDB_GROUP1);

	/*
	 * Ok, read the block via polling
	 */
	mutex_exit(BST_MUTEX(tgt));
	rval = bst_scsi_poll(tgt, pkt);
	mutex_enter(BST_MUTEX(tgt));

	scsi_destroy_pkt(pkt);
	if (rval == DDI_SUCCESS) {
		return (offset);
	} else {
		return (-1);
	}
}


static int
bst_iopb_write_block(
	struct bst_soft			*tgt,
	daddr_t				blkno,
	struct buf			*bp,
	int				(*f)())
{
	register struct scsi_pkt	*pkt;
	int				rval = -1;

	ASSERT(mutex_owned(BST_MUTEX(tgt)));
	ASSERT(f == NULL_FUNC || f == SLEEP_FUNC);

	blkno >>= tgt->blkshf;


	pkt = scsi_init_pkt(ROUTE(tgt), (struct scsi_pkt *)NULL,
		bp, CDB_GROUP1, 1, 0, PKT_CONSISTENT, f, NULL);
	if (!pkt) {
		return (rval);
	}

	(void) scsi_setup_cdb((union scsi_cdb *)pkt->pkt_cdbp,
		SCMD_WRITE_G1, blkno, 1, 0);
	pkt->pkt_flags = 0;
	BST_DUMP_CDB(tgt, pkt, CDB_GROUP1);

	/*
	 * Ok, write the block via polling
	 */
	mutex_exit(BST_MUTEX(tgt));
	rval = bst_scsi_poll(tgt, pkt);
	mutex_enter(BST_MUTEX(tgt));

	scsi_destroy_pkt(pkt);
	return (rval);
}


static void
bst_clear_cont_alleg(struct bst_soft *tgt, int (*f)())
{
	struct buf	*bp;
	struct scsi_pkt	*pkt;

	ASSERT(mutex_owned(BST_MUTEX(tgt)) == 0);
	ASSERT(f == NULL_FUNC || f == SLEEP_FUNC);

	if ((bp = scsi_alloc_consistent_buf(ROUTE(tgt), NULL,
	    SENSE_LENGTH, B_READ, f, NULL)) == NULL)
		return;

	if ((pkt = scsi_init_pkt(ROUTE(tgt), NULL, bp, CDB_GROUP0,
	    1, 0, PKT_CONSISTENT, f, NULL)) == NULL) {
		scsi_free_consistent_buf(bp);
		return;
	}

	(void) scsi_setup_cdb((union scsi_cdb *)pkt->pkt_cdbp,
		SCMD_REQUEST_SENSE, 0, SENSE_LENGTH, 0);

	BST_DUMP_CDB(tgt, pkt, CDB_GROUP0);

	pkt->pkt_time = bst_io_time;
	pkt->pkt_flags |= FLAG_SENSING;

	(void) scsi_poll(tgt->rqs);

	scsi_destroy_pkt(pkt);
	scsi_free_consistent_buf(bp);
}

static int
bst_scsi_poll(struct bst_soft *tgt, struct scsi_pkt *pkt)
{
	int i, err;
	int rval = DDI_FAILURE;

	BST_LOG(tgt->devp, BST_CE_DEBUG3, "bst_scsi_poll\n");

	for (i = 0; i < 3; i++) {
		err = scsi_poll(pkt);

		BST_LOG(tgt->devp, BST_CE_DEBUG3,
			"bst_scsi_poll: err = %d status = %d\n",
			err, SCBP_C(pkt));

		if (err || SCBP_C(pkt) != STATUS_GOOD ||
		    ((pkt->pkt_state & STATE_XFERRED_DATA) == 0 &&
		    (pkt->pkt_resid != 0))) {
			if (SCBP_C(pkt) ==
			    STATUS_RESERVATION_CONFLICT) {
				break;
			}
			if (i > 2 && tgt->state == BST_STATE_NIL) {
				char buf[32];
				scsi_log(DEVINFO(tgt), bst_label, CE_WARN,
				    "unable to execute cmd 0x%x\n",
				    *(pkt->pkt_cdbp), bst_cmds, buf);
			}
			if (SCBP(pkt)->sts_chk) {
				bst_clear_cont_alleg(tgt, NULL_FUNC);
			}
		} else {
			rval = DDI_SUCCESS;
			break;
		}
	}

	return (rval);
}

/*
 * -----------------------------------------------------------------------------
 *	Error Message Data and Routines
 */

/*
 * Log a message to the console and/or syslog with cmn_err
 */
/*ARGSUSED*/
static void
bst_log(struct scsi_device *devp, int level, const char *fmt, ...)
{
	char name[16];
	char buf[256];
	va_list ap;

	if (devp) {
		(void) sprintf(name, "%s%d", ddi_get_name(devp->sd_dev),
		    ddi_get_instance(devp->sd_dev));
	} else {
		(void) sprintf(name, "bst");
	}

	va_start(ap, fmt);
	(void) vsprintf(buf, fmt, ap);
	va_end(ap);

	switch (level) {
	case CE_CONT:
	case CE_NOTE:
	case CE_WARN:
	case CE_PANIC:
		cmn_err(level, "%s:\t%s", name, buf);
		break;

	case BST_CE_DEBUG4: if (bst_debug < 4) break;
		/*FALLTHROUGH*/
	case BST_CE_DEBUG3: if (bst_debug < 3) break;
		/*FALLTHROUGH*/
	case BST_CE_DEBUG2: if (bst_debug < 2) break;
		/*FALLTHROUGH*/
	case BST_CE_DEBUG1:
	default:
		cmn_err(CE_CONT, "^%s:\t%s", name, buf);
		break;
	}
}


/*
 * Print a buffer readably
 */
static void
bst_hex_print(char *msg, char *cp, int len)
{
	int i = 0, j;
	char	buf[BUFLEN];

	bzero(buf, BUFLEN);
	for (i = 0; i < len; i++) {
		/*
		 * make sure there's room for longest %x (i.e., 8 for
		 * a negative number) plus space (1) plus zero (1)
		 */
		if ((j = strlen(buf)) >= (BUFLEN - 10)) {
			buf[BUFLEN-2] = '>';
			buf[BUFLEN-1] = 0;
			break;		/* cp too long, give up */
		}
		(void) sprintf(&buf[j], "%x ", cp[i]);
	}

	cmn_err(CE_CONT, "^%s: %s\n", msg, buf);
}

/*
 * Print a piece of inquiry data - clean up for non-printable characters
 */
static void
bst_inq_fill(char *p, int l, char *s)
{
	int	i = 0;
	char	c;

	while (i++ < l) {
		if ((c = *p++) < ' ' || c >= 0177) {
			c = '*';
		} else if (i != 1 && c == ' ') {
			break;
		}
		*s++ = c;
	}
	*s++ = 0;
}


static void
bst_dump_cdb(struct bst_soft *tgt, struct scsi_pkt *pkt, int cdblen)
{
	static char	hex[] = "0123456789abcdef";
	char		buf [256];
	u_char		*cdb;
	char		*p;
	int		i;

	p = buf;
	cdb = pkt->pkt_cdbp;
	*p++ = '[';
	for (i = 0; i < cdblen; i++, cdb++) {
		if (i > 0)
			*p++ = ' ';
		*p++ = hex[(*cdb >> 4) & 0x0f];
		*p++ = hex[*cdb & 0x0f];
	}
	*p++ = ']';
	*p++ = '\n';
	*p = 0;
	bst_log(tgt->devp, CE_CONT, buf);
}


/*
 * The system calls bst_power() either directly or as a result of
 * ddi_dev_is_needed() when it determines that a component's
 * current power level needs to be changed.
 * bst_power() might be called from bst_dump() when interrupts
 * have been disabled. therefore, we need to use polled cmds.
 */
static int
bst_power(dev_info_t *devi, int component, int level)
{
	int		instance;
	struct bst_soft	*tgt;
	struct scsi_pkt *pkt;
	int		rval;

	instance = ddi_get_instance(devi);

	if (!(tgt = ddi_get_soft_state(bst_state, instance)) ||
	    (level < 0) || (level > 1) || (component != 0)) {
		return (DDI_FAILURE);
	}

	mutex_enter(BST_MUTEX(tgt));

	BST_LOG(tgt->devp, BST_CE_DEBUG3,
		"bst_power: new level = %d, current power_level = %x\n",
		level, tgt->power_level);

	if (level == tgt->power_level) {
		/* already at this power level */
		mutex_exit(BST_MUTEX(tgt));
		return (DDI_SUCCESS);
	}
	mutex_exit(BST_MUTEX(tgt));

	pkt = scsi_init_pkt(ROUTE(tgt), NULL, NULL, CDB_GROUP0,
	    (tgt->arq ? sizeof (struct scsi_arq_status) : 1),
	    0, 0, NULL_FUNC, NULL);

	if (pkt == (struct scsi_pkt *)NULL) {
		return (DDI_FAILURE);
	}
	(void) scsi_setup_cdb((union scsi_cdb *)pkt->pkt_cdbp,
			SCMD_START_STOP, 0,
			(level ? BST_START : BST_STOP), 0);
	BST_DUMP_CDB(tgt, pkt, CDB_GROUP0);

	rval = bst_scsi_poll(tgt, pkt);

	if (rval == DDI_SUCCESS) {
		mutex_enter(BST_MUTEX(tgt));
		tgt->power_level = level;
		mutex_exit(BST_MUTEX(tgt));
	}

	scsi_destroy_pkt(pkt);

	return (rval);
}


#ifdef	DEBUG
static void
dump(char *hdr, u_char *buf, int nbytes)
{
	static char	hex[] = "0123456789abcdef";
	char		tmpbuf[256];
	char		*dst;
	u_char		*src;
	int		n;
	int		nspaces;
	int		c;


	if (hdr != NULL) {
		cmn_err(CE_CONT, hdr);
	}

	while (nbytes > 0) {
		src = buf;
		dst = tmpbuf;
		n = (nbytes < 16) ? nbytes : 16;
		nspaces = 16 - n + 1;
		while (n-- > 0) {
			*dst++ = hex[(*src >> 4) & 0x0f];
			*dst++ = hex[*src++ & 0x0f];
			*dst++ = ' ';
		}
		while (nspaces-- > 0) {
			*dst++ = ' ';
			*dst++ = ' ';
			*dst++ = ' ';
		}

		src = buf;
		n = (nbytes < 16) ? nbytes : 16;
		while (n-- > 0) {
			c = *src++ & 0xff;
			*dst++ = (char)((c >= ' ' && c <= '~') ? c : '.');
		}
		*dst++ = '\n';
		*dst = 0;

		nbytes -= 16;
		buf += 16;
		cmn_err(CE_CONT, tmpbuf);
	}
}
#endif
