/*
 * Copyright (c) 1996,1977 by Sun Microsystems, Inc.
 * All rights reserved.
 */

#pragma ident "@(#)pcsram.c	1.24     97/09/11 SMI"

/*
 * This file is a product of Sun Microsystems, Inc. and is provided
 * for unrestricted use provided that this legend is included on all
 * tape media and as a part of the software program in whole or part.
 * Users may copy or modify this file without charge, but are not
 * authorized to license or distribute it to anyone else except as
 * part of a product or program developed by the user.
 *
 * THIS FILE IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND
 * INCLUDING THE WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE, OR ARISING FROM A COURSE OF DEALING,
 * USAGE OR TRADE PRACTICE.
 *
 * This file is provided with no support and without any obligation
 * on the part of Sun Microsystems, Inc. to assist in its use,
 * correction, modification or enhancement.
 *
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY THIS
 * FILE OR ANY PART THEREOF.
 *
 * In no event will Sun Microsystems, Inc. be liable for any lost
 * revenue or profits or other special, indirect and consequential
 * damages, even if Sun has been advised of the possibility of
 * such damages.
 *
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 *
 */

/*
 *  Solaris PCMCIA Memory Driver
 *
 *	This is a simple Solaris PCMCIA memory client driver.
 *
 *	The sample memory driver is a part pf the Solaris
 *	PCMCIA Device Driver Developer's Kit (DDK).
 *
 *	This sample driver can be used to support disk-like
 *	I/O access to a standard PCMCIA Non-Volatile Static
 *	RAM (SRAM) with Card Structure Information (CIS).
 *
 *	This is a a block device driver which also provides
 *	a raw character driver interface.
 *
 *  	The sample driver only supports single partition s2.
 *
 *	fdformat(1) is the only utility that can be used for
 *	formatting a SRAM memory card.
 *
 *	format(1M) utility does not support PCMCIA memory card.
 *
 *	Note that all debug messages are printed in the
 *	following format:
 *
 *		pcsram%d	pcsram[instance]
 *		pcsram%d.%d	pcsram[instance].[socket]
 *
 */

#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/uio.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/kmem.h>
#include <sys/conf.h>
#include <sys/cmn_err.h>
#include <sys/ksynch.h>
#include <sys/modctl.h>
#include <sys/stat.h>
#include <sys/dkio.h>
#include <sys/vtoc.h>

/* PCMCIA and DDI header files */
#include <sys/pccard.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>

/* pcsram header file */
#include "pcsram.h"


/*
 * Device Operations (dev_ops) Structure
 */
static int pcsram_getinfo(dev_info_t *, ddi_info_cmd_t,
					void *, void **);
static int pcsram_attach(dev_info_t *, ddi_attach_cmd_t);
static int pcsram_detach(dev_info_t *, ddi_detach_cmd_t);

/*
 * Character/Block Operations (cb_ops) Structure
 */
static int pcsram_open(dev_t *, int, int, cred_t *);
static int pcsram_close(dev_t, int, int, cred_t *);
static int pcsram_read(dev_t, struct uio *, cred_t *);
static int pcsram_write(dev_t, struct uio *, cred_t *);
static int pcsram_strategy(struct buf *);
static void pcsram_start(pcsram_state_t *);
static u_int pcsram_softintr(caddr_t);
static void pcsram_minphys(struct buf *);
static int pcsram_ioctl(dev_t, int, intptr_t, int, cred_t *, int *);
static int pcsram_prop_op(dev_t, dev_info_t *, ddi_prop_op_t,
				int, char *, caddr_t, int *);

/*
 *  Misc functions
 */
static int pcsram_update_mapmempage(pcsram_state_t *, int);

/*
 *  PCMCIA related functions
 */
static int pcsram_event(event_t, int, event_callback_args_t *);
static int pcsram_card_insertion(pcsram_state_t *);
static int pcsram_card_removal(pcsram_state_t *);
static int pcsram_parse_cis(pcsram_state_t *);

#if defined(DEBUG)
static void pcsram_display_card_status(get_status_t *);
#endif


/*
 * Local driver data
 */
static void *pcsram_soft_state_p = NULL;
static char *pcsram_name = PCSRAM_NAME;


/* Character/Block Operations (cb_ops) Structure */
static struct cb_ops pcsram_cb_ops = {
	pcsram_open,		/* open 			*/
	pcsram_close,		/* close 			*/
	pcsram_strategy,	/* strategy 			*/
	nodev,			/* print 			*/
	nodev,			/* dump 			*/
	pcsram_read,		/* read 			*/
	pcsram_write,		/* write 			*/
	pcsram_ioctl,		/* ioctl 			*/
	nodev,			/* devmap 			*/
	nodev,			/* mmap 			*/
	ddi_segmap,		/* segmap 			*/
	nochpoll,		/* poll 			*/
	pcsram_prop_op,		/* prop_op 			*/
	NULL,			/* streamtab			*/
	D_NEW | D_MP,		/* Driver compatibility flag	*/
	CB_REV,			/* cb_rev			*/
	nodev,			/* async I/O read entry point	*/
	nodev 			/* async I/O write entry point	*/
};


/* Device Operations (dev_ops) Structure */
static struct dev_ops pcsram_ops = {
	DEVO_REV,		/* devo_rev	*/
	0,			/* refcnt	*/
	pcsram_getinfo,		/* info		*/
	nulldev,		/* identify	*/
	nulldev,		/* probe 	*/
	pcsram_attach,		/* attach	*/
	pcsram_detach,		/* detach	*/
	nodev,			/* reset	*/
	&pcsram_cb_ops,		/* cb_ops pointer for leaf driver   */
	(struct bus_ops *)NULL  /* bus_ops pointer for nexus driver */
};


/* Module linkage information for the kernel */
extern struct mod_ops mod_driverops;

static struct modldrv md = {
	&mod_driverops,		/* Type of module. This is a driver */
	PCSRAM_DRIVERID,	/* Driver Identifier string */
	&pcsram_ops,		/* Device Operation Structure */
};

static struct modlinkage modlinkage = {
	MODREV_1,
	&md,
	NULL
};


/*
 * Module Initialization functions.
 *
 *	Routine		Called by
 *	_init(9E)	modload(9F)
 *	_info(9E)	modinfo(9F)
 *	_fini(9E)	modunload(9F)
 */
int
_init(void)
{
	int	error;


	PCSRAM_DEBUG((CE_CONT, "pcsram: _init\n"));
	error = ddi_soft_state_init(&pcsram_soft_state_p,
					sizeof (pcsram_state_t),
					1	/* n_items */);
	if (error) {
		PCSRAM_DEBUG((CE_CONT,
			"pcsram: _init ddi_soft_state_init failed\n"));
		return (error);
	}

	error = mod_install(&modlinkage);
	if (error) {
		ddi_soft_state_fini(&pcsram_soft_state_p);
		PCSRAM_DEBUG((CE_CONT,
			"pcsram: _init mod_install failed\n"));
		return (error);
	}

	return (error);
}

int
_info(struct modinfo *modinfop)
{
	return (mod_info(&modlinkage, modinfop));
}

int
_fini(void)
{
	int	error;


	PCSRAM_DEBUG((CE_CONT, "pcsram: _fini\n"));

	error = mod_remove(&modlinkage);
	if (error) {
		PCSRAM_DEBUG((CE_CONT,
			"pcsram: _fini mod_remove failed\n"));
		return (error);
	}

	ddi_soft_state_fini(&pcsram_soft_state_p);

	return (error);
}


/*
 * Autoconfiguration routines
 *	_getinfo(9E)
 *	_attach(9E)
 *	_detach(9E)
 */

/*
 * pcsram_getinfo() - Get device driver information
 *
 *	pcsram_getinfo() uses csx_CS_DDI_Info(9F) as a
 *	mechanism to retrieve the instance information
 *	since PC Card drivers encode the physical socket
 *	number as part of the dev_t of device nodes
 *	instead of the instance number.
 *
 */
/*ARGSUSED*/
static int
pcsram_getinfo(dev_info_t *dip, ddi_info_cmd_t cmd, void *arg,
						void **result)
{
	int		error = DDI_SUCCESS;
	pcsram_state_t	*rs;
	cs_ddi_info_t	cs_ddi_info;


	PCSRAM_DEBUG((CE_CONT, "pcsram: getinfo\n"));

	switch (cmd) {

	    case DDI_INFO_DEVT2DEVINFO:
		cs_ddi_info.Socket = PCSRAM_SOCKET((dev_t)arg);
		cs_ddi_info.driver_name = pcsram_name;
		if (csx_CS_DDI_Info(&cs_ddi_info) != CS_SUCCESS) {
			return (DDI_FAILURE);
		}
		if (!(rs = ddi_get_soft_state(pcsram_soft_state_p,
					cs_ddi_info.instance))) {
			*result = NULL;
		} else {
			*result = rs->dip;
		}
		break;

	    case DDI_INFO_DEVT2INSTANCE:
		cs_ddi_info.Socket = PCSRAM_SOCKET((dev_t)arg);
		cs_ddi_info.driver_name = pcsram_name;
		if (csx_CS_DDI_Info(&cs_ddi_info) != CS_SUCCESS) {
			return (DDI_FAILURE);
		}
		*result = (void *)cs_ddi_info.instance;
		break;

	    default:
		error = DDI_FAILURE;
		break;
	} /* switch */

	return (error);
}


/*
 * pcsram_attach() - Perform device-specific initialization:
 *
 *	Allocating per-instance state structures.
 *	Allocating transfer list header.
 *	Registering with PCMCIA CardServices
 *	Get PCMCIA logical socket number
 *	Add the device driver's soft interrupt handler
 *	Initializing mutex and condition variables.
 *	Using csx_RequestSocketMask(9F) to enable the driver
 *		to start receiving events.
 *
 *	Note that the system resource allocation such as
 *	creating minor node are done in pcsram_card_insertion()
 *	routine which is called when the driver receives an
 *	insertion event.  The resource is de-allocated when
 *	the card is removed and re-allocated when the card
 *	is inserted again.
 *
 */
static int
pcsram_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
	int			instance;
	pcsram_state_t		*rs;
				/* CardServices variables */
	client_reg_t		client_reg;
	sockmask_t		sockmask;
	int			ret;
	map_log_socket_t	map_log_socket;
	get_status_t		get_status;


	instance = ddi_get_instance(dip);

	PCSRAM_DEBUG((CE_CONT, "pcsram%d: attach cmd 0x%x\n",
		instance, cmd));

	/*
	 * make sure we're only being asked to do an attach
	 */
	if (cmd != DDI_ATTACH) {
		PCSRAM_DEBUG((CE_CONT, "pcsram%d: attach failed, "
			"cmd != DDI_ATTACH\n", instance));
		return (DDI_FAILURE);
	}

	if (ddi_soft_state_zalloc(pcsram_soft_state_p,
					instance) != DDI_SUCCESS) {
		PCSRAM_DEBUG((CE_NOTE, "pcsram%d: attach -"
			"could not allocate state structure\n",
			instance));
		return (DDI_FAILURE);
	}

	rs = ddi_get_soft_state(pcsram_soft_state_p, instance);
	if (rs == NULL) {
		PCSRAM_DEBUG((CE_NOTE, "pcsram%d: attach -"
			"could not get state structure\n", instance));
		goto out;
	}


	/* Remember dev_info structure for getinfo  */
	rs->dip = dip;
	rs->instance = instance;
	ddi_set_driver_private(dip, (caddr_t)rs);

	/* clear the per-unit flags field */
	rs->flags = 0;

	/* Initialize the card event */
	rs->card_event = 0;

	rs->blk_open =  0;
	rs->chr_open =  0;
	rs->nlayered =  0;
	rs->busy =  0;

	/*
	 * Initialize the physical I/O max transfer size
	 */
	rs->maxphys = DEFAULT_MAXPHYS;

	/*
	 * Continuing to return EIO if the card is ejected while
	 * it is mounted until the LAST layered close is called
	 */
	rs->ejected_while_mounting = 0;

	/* Allocate hard drive characteristic structure */
	rs->hdrv_chars = (struct hd_char *)
	    kmem_alloc(sizeof (struct hd_char), KM_SLEEP);

	/* allocate transfer list header */
	rs->blist = getrbuf(KM_SLEEP);
	if (rs->blist == NULL) {
		PCSRAM_DEBUG((CE_NOTE, "pcsram%d: attach -"
			"could not allocate transfer list header\n",
			instance));
		goto out;
	}

	/* queue is empty */
	rs->blist->av_forw = NULL;

	/*
	 * Add low level soft interrupt to the system
	 *	It is used in pcsram_start() when ddi_trigger_softintr()
	 *	is called to start pcsram_softintr() for transfering
	 *	any additional data.
	 */
	if (ddi_add_softintr(dip, DDI_SOFTINT_MED, &rs->softint_id,
		&rs->soft_blk_cookie, (ddi_idevice_cookie_t *)NULL,
		pcsram_softintr, (caddr_t)rs) != DDI_SUCCESS) {
			PCSRAM_DEBUG((CE_NOTE,
				"pcsram%d: attach -"
				"could not add soft interrupt\n",
				instance));
			goto out;
	}

	rs->flags |= PCSRAM_SOFTINTROK;

	/*
	 * Register with Card Services
	 * Note that we set CS_EVENT_CARD_REMOVAL_LOWP so that we get
	 *	low priority CS_EVENT_CARD_REMOVAL events as well.
	 */
	client_reg.Attributes = (INFO_MEM_CLIENT |
					INFO_CARD_SHARE |
					INFO_CARD_EXCL);
	client_reg.EventMask = (CS_EVENT_REGISTRATION_COMPLETE |
				CS_EVENT_CARD_INSERTION |
				CS_EVENT_CARD_REMOVAL |
				CS_EVENT_CARD_REMOVAL_LOWP |
				CS_EVENT_CLIENT_INFO);
	client_reg.event_handler = (csfunction_t *)pcsram_event;
	client_reg.event_callback_args.client_data = rs;
	client_reg.Version = CS_VERSION;
	client_reg.dip = dip;
	(void) strcpy(client_reg.driver_name, pcsram_name);
	if ((ret = csx_RegisterClient(&rs->client_handle,
					&client_reg)) != CS_SUCCESS) {
			error2text_t cft;

			cft.item = ret;
			csx_Error2Text(&cft);
			PCSRAM_DEBUG((CE_CONT, "pcsram%d: attach - "
				"RegisterClient failed %s (0x%x)\n",
				instance, cft.text, ret));
			goto out;
	}

	/* Get logical socket number and store in pcsram_state_t */
	if ((ret = csx_MapLogSocket(rs->client_handle,
				&map_log_socket)) != CS_SUCCESS) {
		error2text_t cft;

		cft.item = ret;
		csx_Error2Text(&cft);
		PCSRAM_DEBUG((CE_CONT, "pcsram%d: attach - "
			"MapLogSocket failed %s (0x%x)\n",
			instance, cft.text, ret));
	}

	rs->sn = map_log_socket.PhySocket;

	rs->flags |= PCSRAM_REGCLIENT;

	/* Setup the event handler hi-level mutex */
	mutex_init(&rs->event_hilock, "rs->event_hilock",
			MUTEX_DRIVER, *(client_reg.iblk_cookie));

	/* set up the mutex to protect the pcsram_state_t */
	mutex_init(&rs->mutex, MUTEX_CONTROLER_NAME,
			MUTEX_DRIVER, &rs->soft_blk_cookie);

	/* Init readywait_cv */
	cv_init(&rs->readywait_cv, "rs->readywait_cv",
					CV_DRIVER, (void *)NULL);

	rs->flags |= PCSRAM_DID_INITMUTEX;

	mutex_enter(&rs->event_hilock);

	/*
	 * After the RequestSocketMask call,
	 * we can start receiving events
	 */
	sockmask.EventMask = (CS_EVENT_CARD_INSERTION |
				CS_EVENT_CARD_REMOVAL);

	if ((ret = csx_RequestSocketMask(rs->client_handle,
					&sockmask)) != CS_SUCCESS) {
		error2text_t cft;

		cft.item = ret;
		csx_Error2Text(&cft);
		PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: attach -"
			"RequestSocketMask failed %s (0x%x)\n",
			instance, rs->sn, cft.text, ret));
		mutex_exit(&rs->event_hilock);
		goto out;
	}

	rs->flags |= PCSRAM_REQSOCKMASK;

	/*
	 * If the card is inserted and this attach is triggered
	 *	by an open, that open will wait until
	 *	pcsram_card_insertion() is completed.
	 */
	cv_wait(&rs->readywait_cv, &rs->event_hilock);

	/*
	 * PC Card now must be present and fully functional
	 */
	if (!PCSRAM_CARD_PRESENT(rs)) {
		return (DDI_FAILURE);
	}

	mutex_exit(&rs->event_hilock);

	ddi_report_dev(dip);

	return (DDI_SUCCESS);

out:
	(void) pcsram_detach(dip, DDI_DETACH);
	return (DDI_FAILURE);
}


/*
 * pcsram_detach() - Detach and prepare driver to unload:
 *
 *	Releasing all resouces allocated by pcsram_attach()
 *		or from a card insertion event by calling
 *		pcsram_card_removal() routine.
 *	Release driver socket mask.
 *	Deregister with Card Services to stop getting events.
 *	Free transfer list header.
 *	Unregister the softinterrupt handler
 *	And free the various mutex and condition variables.
 *
 */
static int
pcsram_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	int		instance, ret;
	pcsram_state_t	*rs;


	instance = ddi_get_instance(dip);

	PCSRAM_DEBUG((CE_CONT, "pcsram%d: detach cmd 0x%x\n",
		instance, cmd));

	if (cmd != DDI_DETACH) {
		PCSRAM_DEBUG((CE_CONT, "pcsram%d: detach failed, "
			"cmd != DDI_DETACH\n", instance));
		return (DDI_FAILURE);
	}

	rs = ddi_get_soft_state(pcsram_soft_state_p, instance);
	if (rs == NULL) {
		PCSRAM_DEBUG((CE_NOTE, "pcsram%d.%d: detach -"
			"could not get state structure\n",
			instance, rs->sn));
		return (DDI_FAILURE);
	}

	/*
	 * Call pcsram_card_removal to do any final card cleanup
	 */
	(void) pcsram_card_removal(rs);

	/*
	 * Release driver socket mask
	 */
	if (rs->flags & PCSRAM_REQSOCKMASK) {
		release_socket_mask_t rsm;

		if ((ret = csx_ReleaseSocketMask(rs->client_handle,
					&rsm)) != CS_SUCCESS) {
			error2text_t cft;

			cft.item = ret;
			csx_Error2Text(&cft);
			PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: detach - "
				"ReleaseSocketMask failed %s (0x%x)\n",
				instance, rs->sn, cft.text, ret));
		}
	}

	/*
	 * Deregister with Card Services - we will stop getting
	 *	events at this point.
	 */
	if (rs->flags & PCSRAM_REGCLIENT) {
		if ((ret = csx_DeregisterClient(rs->client_handle))
						!= CS_SUCCESS) {
			error2text_t cft;

			cft.item = ret;
			csx_Error2Text(&cft);

			PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: detach - "
				"DeregisterClient failed %s (0x%x)\n",
				instance, rs->sn, cft.text, ret));
		}
	}

	/* Unregister the softinterrupt handler */
	if (rs->flags & PCSRAM_SOFTINTROK) {
		ddi_remove_softintr(rs->softint_id);
	}

	/* Free transfer list header */
	if (rs->blist) {
		freerbuf(rs->blist);
	}

	/* Allocate hard drive characteristic structure */
	if (rs->hdrv_chars) {
		kmem_free((caddr_t)rs->hdrv_chars,
				sizeof (struct hd_char));
	}

	/* Free the various mutexii */
	if (rs->flags & PCSRAM_DID_INITMUTEX) {
		mutex_destroy(&rs->event_hilock);
		mutex_destroy(&rs->mutex);
		/* Free readywait_cv */
		cv_destroy(&rs->readywait_cv);
	}

	ddi_soft_state_free(pcsram_soft_state_p, instance);

	return (DDI_SUCCESS);
}


/*
 * Controlling Device Access
 *
 *	open(9E)
 *	close(9E)
 *
 */

/*
 * pcsram_open() - open(9E) system call
 */
/*ARGSUSED*/
static int
pcsram_open(dev_t *devp, int flag, int otyp, cred_t *cred)
{
	int		err;
	int		ret;
	pcsram_state_t	*rs;
	get_status_t	get_status;
	cs_ddi_info_t	cs_ddi_info;


	/*
	 * Since PC Card driver encodes physical socket number
	 *	as part of the dev_t of device nodes instead of
	 *	the instance number we have to use csx_CS_DDI_Info(9F)
	 *	to get instance number that is associated with
	 *	a particular PC Card driver and socket number
	 *	for using in ddi_get_soft_state().
	 */
	cs_ddi_info.Socket = PCSRAM_SOCKET((dev_t)*devp);
	cs_ddi_info.driver_name = pcsram_name;
	if (csx_CS_DDI_Info(&cs_ddi_info) != DDI_SUCCESS) {
		PCSRAM_DEBUG((CE_NOTE,
			"pcsram: open, csx_CS_DDI_Info() failed\n"));
		return (ENXIO);
	}

	rs = ddi_get_soft_state(pcsram_soft_state_p,
					cs_ddi_info.instance);
	if (rs == NULL) {
		PCSRAM_DEBUG((CE_NOTE, "pcsram%d: open, "
			"could not get soft state\n",
			cs_ddi_info.instance));
		return (ENXIO);
	}

	PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: open, "
		"flag 0x%x otyp 0x%x\n", cs_ddi_info.instance,
		rs->sn, flag, otyp));

	mutex_enter(&rs->mutex);

	if ((err = csx_GetStatus(rs->client_handle,
					&get_status)) != CS_SUCCESS) {
		error2text_t cft;

		mutex_exit(&rs->mutex);
		cft.item = err;
		csx_Error2Text(&cft);
		PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: open, "
			"GetStatus failed %s (0x%x)\n",
			cs_ddi_info.instance, rs->sn, cft.text, err));
		return (ENXIO);
	}

	/*
	 * Check to see if the card is present.
	 *	If there is no card in the socket,
	 *	then return ENXIO.
	 */
	if (!(get_status.CardState & CS_EVENT_CARD_INSERTION)) {
		PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: open, "
			"Found no card from get_status.CardState\n",
			cs_ddi_info.instance, rs->sn));
		mutex_exit(&rs->mutex);
		return (ENXIO);
	}

	if (get_status.CardState & CS_EVENT_BATTERY_DEAD) {
		if (!rs->battery_dead_posted) {
			PCSRAM_DEBUG((CE_WARN, "pcsram%d.%d: open, "
				"Battery & Data integrity "
				"is not guaranteed\n",
				cs_ddi_info.instance, rs->sn));
			/* Display once on the system console */
			rs->battery_dead_posted++;
		}
		if (flag & FWRITE) {
			PCSRAM_DEBUG((CE_WARN, "pcsram%d.%d: open, "
				"Can not write to card while "
				"the battery is not dead\n",
				cs_ddi_info.instance, rs->sn));
			mutex_exit(&rs->mutex);
			return (ENXIO);
		}
	}

	if (get_status.CardState & CS_EVENT_BATTERY_LOW) {
		if (!rs->battery_low_posted) {
			PCSRAM_DEBUG((CE_WARN, "pcsram%d.%d: open, "
				"Battery should be replaced; "
				"Data is OK\n", cs_ddi_info.instance,
				rs->sn));
			/* Display once on the system console */
			rs->battery_low_posted++;
		}
	}

	/* Next check for read only file system */
	if ((flag & FWRITE) &&
		(get_status.CardState & CS_EVENT_WRITE_PROTECT)) {
			PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: open, "
				"Card is write-protected\n",
				cs_ddi_info.instance, rs->sn));
			mutex_exit(&rs->mutex);
			return (EROFS);
	}


	/*
	 * Only honor FEXCL.  If a regular open or a layered open
	 * is still outstanding on the device, the exclusive open
	 * must fail.
	 */
	if ((flag & FEXCL) && (rs->blk_open || rs->chr_open ||
						rs->nlayered)) {
		mutex_exit(&rs->mutex);
		return (EAGAIN);
	}

	switch (otyp) {

		case OTYP_BLK:
			rs->blk_open = 1;
			break;

		case OTYP_CHR:
			rs->chr_open = 1;
			break;

		case OTYP_LYR:
			rs->nlayered++;
			break;

		default:
			mutex_exit(&rs->mutex);
			return (EINVAL);
	}

	mutex_exit(&rs->mutex);
	return (0);
}



/*
 * pcsram_close() - close(9E) system call
 */
/*ARGSUSED*/
static int
pcsram_close(dev_t devp, int flag, int otyp, cred_t *cred)
{
	int		ret;
	pcsram_state_t	*rs;
	cs_ddi_info_t	cs_ddi_info;


	/*
	 * Since PC Card driver encodes physical socket number
	 *	as part of the dev_t of device nodes instead of
	 *	the instance number we have to use csx_CS_DDI_Info(9F)
	 *	to get instance number that is associated with
	 *	a particular PC Card driver and socket number
	 *	for using in ddi_get_soft_state().
	 */
	cs_ddi_info.Socket = PCSRAM_SOCKET((dev_t)devp);
	cs_ddi_info.driver_name = pcsram_name;
	if (csx_CS_DDI_Info(&cs_ddi_info) != DDI_SUCCESS) {
		PCSRAM_DEBUG((CE_NOTE,
			"pcsram: close, csx_CS_DDI_Info() failed\n"));
		return (ENXIO);
	}

	rs = ddi_get_soft_state(pcsram_soft_state_p,
					cs_ddi_info.instance);
	if (rs == NULL) {
		PCSRAM_DEBUG((CE_NOTE, "pcsram%d: close, "
			"could not get soft state\n",
			cs_ddi_info.instance));
		return (ENXIO);
	}

	PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: close, "
		"flag 0x%x otyp 0x%x\n", cs_ddi_info.instance,
		rs->sn, flag, otyp));

	mutex_enter(&rs->mutex);

	switch (otyp) {

		case OTYP_BLK:
			rs->blk_open = 0;
		break;

		case OTYP_CHR:
			rs->chr_open = 0;
			break;

		case OTYP_LYR:
			rs->nlayered--;
			break;

		default:
			mutex_exit(&rs->mutex);
			return (EINVAL);
	}

	if (rs->blk_open || rs->chr_open || rs->nlayered) {
		/* not done yet */
		mutex_exit(&rs->mutex);
		return (0);
	}

	/*
	 * Continuing to return EIO if the card is ejected
	 * while it is mounted until the last layered close
	 * is called
	 */
	if (rs->nlayered == 0) {
		rs->ejected_while_mounting = 0;
	}

	mutex_exit(&rs->mutex);
	return (0);
}


/*
 * I/O Request Handling routines
 *
 *	read(9E)
 *	write(9E)
 *	strategy(9E)
 *	start(9E)
 *	minphys(9E)
 *
 *	pcsram_softintr()
 *
 */


/*
 * pcsram_read() - read(9E) system call
 */
/*ARGSUSED*/
static int
pcsram_read(dev_t devp, struct uio *uiop, cred_t *credp)
{
	int		ret;
	int		rval;
	int		err;
	pcsram_state_t	*rs;
	get_status_t	get_status;
	cs_ddi_info_t	cs_ddi_info;


	/*
	 * Since PC Card driver encodes physical socket number
	 *	as part of the dev_t of device nodes instead of
	 *	the instance number we have to use csx_CS_DDI_Info(9F)
	 *	to get instance number that is associated with
	 *	a particular PC Card driver and socket number
	 *	for using in ddi_get_soft_state().
	 */
	cs_ddi_info.Socket = PCSRAM_SOCKET((dev_t)devp);
	cs_ddi_info.driver_name = pcsram_name;
	if (csx_CS_DDI_Info(&cs_ddi_info) != DDI_SUCCESS) {
		PCSRAM_DEBUG((CE_NOTE,
			"pcsram: read, csx_CS_DDI_Info() failed\n"));
		return (ENXIO);
	}

	rs = ddi_get_soft_state(pcsram_soft_state_p,
					cs_ddi_info.instance);
	if (rs == NULL) {
		PCSRAM_DEBUG((CE_NOTE, "pcsram%d: read, "
			"could not get soft state\n",
			cs_ddi_info.instance));
		return (ENXIO);
	}

	PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: read\n",
				cs_ddi_info.instance, rs->sn));

	mutex_enter(&rs->mutex);

	/* Get card state status */
	if ((err = csx_GetStatus(rs->client_handle,
					&get_status)) != CS_SUCCESS) {
		error2text_t cft;

		mutex_exit(&rs->mutex);
		cft.item = err;
		csx_Error2Text(&cft);
		PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: read, "
			"GetStatus failed %s (0x%x)\n",
			cs_ddi_info.instance, rs->sn, cft.text, err));
		return (ENXIO);
	}

	/*
	 * Check to see if the card is present.
	 *	If the memory card has been removed or
	 *	was ever removed, return an I/O error (EIO)
	 *	not "No such device or address" (EMXIO).
	 *
	 *	Continuing to return EIO if the card is ejected
	 *	while it is mounted until the LAST layered close
	 *	is called
	 *
	 */
	if (!(get_status.CardState & CS_EVENT_CARD_INSERTION) ||
					rs->ejected_while_mounting) {
		if (!rs->card_eject_posted) {
			/* WARNING - card is ejected */
			rs->card_eject_posted++;
			rs->ejected_while_mounting = 1;
			cmn_err(CE_WARN, "pcsram%d.%d: read, "
				"Card is ejected & "
				"Data integrity is not guaranteed",
				cs_ddi_info.instance, rs->sn);
		}
		mutex_exit(&rs->mutex);
		return (EIO);
	}

	mutex_exit(&rs->mutex);

	if (uiop->uio_offset >= rs->card_size) {
		rval = ENOSPC;
		goto out;
	}

	rval = physio(pcsram_strategy, (struct buf *)0, devp,
				B_READ, pcsram_minphys, uiop);

out:
	return (rval);
}


/*
 * pcsram_write() - write(9E) system call
 */
/*ARGSUSED*/
static int
pcsram_write(dev_t devp, struct uio *uiop, cred_t *credp)
{
	int		rval;
	int		err;
	int		ret;
	pcsram_state_t	*rs;
	get_status_t	get_status;
	cs_ddi_info_t	cs_ddi_info;


	/*
	 * Since PC Card driver encodes physical socket number
	 *	as part of the dev_t of device nodes instead of
	 *	the instance number we have to use csx_CS_DDI_Info(9F)
	 *	to get instance number that is associated with
	 *	a particular PC Card driver and socket number
	 *	for using in ddi_get_soft_state().
	 */
	cs_ddi_info.Socket = PCSRAM_SOCKET((dev_t)devp);
	cs_ddi_info.driver_name = pcsram_name;
	if (csx_CS_DDI_Info(&cs_ddi_info) != DDI_SUCCESS) {
		PCSRAM_DEBUG((CE_NOTE,
			"pcsram: write, csx_CS_DDI_Info() failed\n"));
		return (ENXIO);
	}

	rs = ddi_get_soft_state(pcsram_soft_state_p,
					cs_ddi_info.instance);
	if (rs == NULL) {
		PCSRAM_DEBUG((CE_NOTE, "pcsram%d: write, "
			"could not get soft state\n",
			cs_ddi_info.instance));
		return (ENXIO);
	}

	PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: write\n",
				cs_ddi_info.instance, rs->sn));

	mutex_enter(&rs->mutex);

	/* Get card state status */
	if ((err = csx_GetStatus(rs->client_handle, &get_status))
						!= CS_SUCCESS) {
		error2text_t cft;

		mutex_exit(&rs->mutex);
		cft.item = err;
		csx_Error2Text(&cft);
		PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: write, "
			"GetStatus failed %s (0x%x)\n",
			cs_ddi_info.instance, rs->sn, cft.text, err));
		return (ENXIO);
	}

	/*
	 * Check to see if the card is present.
	 *	If the memory card has been removed or
	 *	was ever removed, return an I/O error (EIO)
	 *	not "No such device or address" (EMXIO).
	 *
	 *	Continuing to return EIO if the card is ejected
	 *	while it is mounted until the LAST layered close
	 *	is called
	 *
	 */
	if (!(get_status.CardState & CS_EVENT_CARD_INSERTION) ||
					rs->ejected_while_mounting) {
		if (!rs->card_eject_posted) {
			/* WARNING - card is ejected */
			rs->card_eject_posted++;
			rs->ejected_while_mounting = 1;
			cmn_err(CE_WARN, "pcsram%d.%d: write, "
				"Card is ejected & "
				"Data integrity is not guaranteed",
				cs_ddi_info.instance, rs->sn);
		}
		mutex_exit(&rs->mutex);
		return (EIO);
	}

	mutex_exit(&rs->mutex);

	if (uiop->uio_offset >= rs->card_size) {
		rval = ENOSPC;
		goto out;
	}

	rval = physio(pcsram_strategy, (struct buf *)0, devp,
				B_WRITE, pcsram_minphys, uiop);

out:
	return (rval);
}


/*
 * pcsram_strategy() - strategy(9E) perform I/O transfer.
 */
static int
pcsram_strategy(struct buf *bp)
{
	int		err;
	int		ret;
	int		offset = bp->b_blkno * DEV_BSIZE;
	pcsram_state_t	*rs;
	get_status_t	get_status;
	cs_ddi_info_t	cs_ddi_info;


	/*
	 * Since PC Card driver encodes physical socket number
	 *	as part of the dev_t of device nodes instead of
	 *	the instance number we have to use csx_CS_DDI_Info(9F)
	 *	to get instance number that is associated with
	 *	a particular PC Card driver and socket number
	 *	for using in ddi_get_soft_state().
	 */
	cs_ddi_info.Socket = PCSRAM_SOCKET((dev_t)bp->b_edev);
	cs_ddi_info.driver_name = pcsram_name;
	if (csx_CS_DDI_Info(&cs_ddi_info) != DDI_SUCCESS) {
		PCSRAM_DEBUG((CE_NOTE, "pcsram: strategy, "
			"csx_CS_DDI_Info() failed\n"));
		bioerror(bp, ENXIO);
		goto out;
	}

	rs = ddi_get_soft_state(pcsram_soft_state_p,
					cs_ddi_info.instance);
	if (rs == NULL) {
		PCSRAM_DEBUG((CE_NOTE, "pcsram%d: strategy, "
			"could not get soft state\n",
			cs_ddi_info.instance));
		bioerror(bp, ENXIO);
		goto out;
	}

	mutex_enter(&rs->mutex);

	/* Get card state status */
	if ((err = csx_GetStatus(rs->client_handle, &get_status))
						!= CS_SUCCESS) {
		error2text_t cft;

		mutex_exit(&rs->mutex);
		cft.item = err;
		csx_Error2Text(&cft);
		PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: strategy, "
			"GetStatus failed %s (0x%x)\n",
			cs_ddi_info.instance, rs->sn, cft.text, err));
		bioerror(bp, ENXIO);
		return (0);
	}

	/*
	 * Check to see if the card is present.
	 *	If the memory card has been removed or
	 *	was ever removed, return an I/O error (EIO)
	 *	not "No such device or address" (EMXIO).
	 *
	 *	Continuing to return EIO if the card is ejected
	 *	while it is mounted until the LAST layered close
	 *	is called
	 *
	 */
	if (!(get_status.CardState & CS_EVENT_CARD_INSERTION) ||
					rs->ejected_while_mounting) {
		if (!rs->card_eject_posted) {
			/* WARNING - card is ejected */
			rs->card_eject_posted++;
			rs->ejected_while_mounting = 1;
			cmn_err(CE_WARN, "pcsram%d.%d: strategy, "
				"Card is ejected & "
				"Data integrity is not guaranteed",
				cs_ddi_info.instance, rs->sn);
		}
		mutex_exit(&rs->mutex);
		bioerror(bp, EIO);
		goto out;
	}

	mutex_exit(&rs->mutex);

	if (offset >= rs->card_size) {
		bioerror(bp, ENOSPC);
		goto out;
	}

	mutex_enter(&rs->mutex);

	rs->busy = 1;

	/*
	 * Maping the user pages into the kernelspace
	 */
	bp_mapin(bp);

	if (rs->blist->av_forw == NULL) {
		/* nothing on queue */
		bp->av_forw = NULL;
		bp->av_back = NULL;
		rs->blist->av_forw = bp;
		rs->blist->av_back = bp;
		/* start it */
		pcsram_start(rs);
	} else {
		/* put on work list */
		bp->av_forw = NULL;
		rs->blist->av_back->av_forw = bp;
		rs->blist->av_back = bp;
		rs->busy = 0;
	}

	mutex_exit(&rs->mutex);
	return (0);

out:
	bp->b_resid = bp->b_bcount;
	biodone(bp);
	return (0);
}


/*
 * pcsram_start() - is called by pcsram_strategy() every time
 *	it queues a request so that an idle device can be
 *	started.  It is also called by the interrupt handler
 *	before it returns from a claimed interrupt so that
 *	a non-empty queue can be serviced.
 */
static void
pcsram_start(pcsram_state_t *rs)
{

	int		offset;
	int		nbytes, origbytes;
	int		next_offset;
	int		remainder_wsize;
	caddr_t		buf_addr;
	struct	buf	*bp;


	bp = rs->blist->av_forw;
	offset = bp->b_blkno * DEV_BSIZE;
	nbytes = min(bp->b_bcount, rs->card_size-offset);
	buf_addr = bp->b_un.b_addr;
	origbytes = nbytes;	/* save the original byte count */

	while (nbytes > 0) {
		int copybytes;

		next_offset = pcsram_update_mapmempage(rs, offset);
		if (next_offset < 0) {
			/* something failed so abort */
			bioerror(bp, EIO);
			break;
		}

		/* get partial block size if not on window boundary */
		remainder_wsize = offset % rs->win_size;
		copybytes = min(nbytes, rs->win_size - remainder_wsize);

		if (bp->b_flags & B_READ) {
			/*
			 * Using CardServices Data Access functions
			 *	to read data from PC Card memory
			 */
			csx_RepGet8(
				/* Card access handle */
				rs->access_handle,
				/* base destination address */
				(uchar_t *)buf_addr,
				/* window offset */
				(uint_t)remainder_wsize,
				/* N bytes transfer */
				copybytes,
				/* flag */
				CS_DEV_AUTOINCR);
		} else {
			/*
			 * Using CardServices Data Access functions
			 *	to write data to PC Card memory
			 */
			csx_RepPut8(
				rs->access_handle,
				(uchar_t *)buf_addr,
				(uint_t)remainder_wsize,
				copybytes,
				CS_DEV_AUTOINCR);
		}

		nbytes -= copybytes;
		buf_addr += copybytes;
		offset += copybytes;

		if (!PCSRAM_CARD_PRESENT(rs)) {
			/*
			 * stop when there is a card removal event
			 */
			bioerror(bp, EIO);
			break;
		}

	}  /* while (nbytes) */

	bp->b_resid = bp->b_bcount - origbytes;

	bp = rs->blist->av_forw;
	rs->blist->av_forw = bp->av_forw;
	bp_mapout(bp);
	biodone(bp);

	ddi_trigger_softintr(rs->softint_id);
}



/*
 * pcsram_softintr() - software interrupt handling routine.
 */
static u_int
pcsram_softintr(caddr_t arg)
{
	register pcsram_state_t *rs = (pcsram_state_t *)arg;
	struct buf	*bp;


	mutex_enter(&rs->mutex);
	if (rs->busy) {
		/* got work to do */
		if (rs->blist->av_forw != NULL) {
			pcsram_start(rs);
		} else {
			rs->busy = 0;
		}
	}  /* rs->busy */
	mutex_exit(&rs->mutex);

	return (DDI_INTR_CLAIMED);
}


/*
 * pcsram_minphys() - minphys(9E)
 *	This is a pointer to a function to be called by
 *	physio(9F) to ensure that the size of the requested
 *	transfer does not exceed a driver-imposed limit.
 */
static void
pcsram_minphys(struct buf *bp)
{
	pcsram_state_t	*rs;
	cs_ddi_info_t	cs_ddi_info;


	/*
	 * Since PC Card driver encodes physical socket number
	 *	as part of the dev_t of device nodes instead of
	 *	the instance number we have to use csx_CS_DDI_Info(9F)
	 *	to get instance number that is associated with
	 *	a particular PC Card driver and socket number
	 *	for using in ddi_get_soft_state().
	 */
	cs_ddi_info.Socket = PCSRAM_SOCKET((dev_t)bp->b_edev);
	cs_ddi_info.driver_name = pcsram_name;
	if (csx_CS_DDI_Info(&cs_ddi_info) != DDI_SUCCESS) {
		PCSRAM_DEBUG((CE_NOTE,
			"pcsram: minphys, csx_CS_DDI_Info() failed\n"));
	}

	rs = ddi_get_soft_state(pcsram_soft_state_p,
					cs_ddi_info.instance);

	if (bp->b_bcount > rs->maxphys) {
		bp->b_bcount = rs->maxphys;
	}
}


/*
 * Miscellaneous I/O Control
 *	ioctl(9E)
 *	prop_op(9E)
 */

/*
 * pcsram_ioctl() - ioctl(9E)
 *	This routine is called when a user thread
 *	issues an ioctl(2) system call on a file
 *	descriptor associated with the device.
 */
/*ARGSUSED*/
static int
pcsram_ioctl(dev_t devp, int cmd, intptr_t arg, int mode, cred_t *credp,
							int *rvalp)
{
	int			err;
	int			ret;
	struct	dk_geom		dkg;	    /* disk geometry */
	struct	vtoc		vtoc;
	struct	dk_cinfo	dkc;	    /* disk controller info */
	pcsram_state_t		*rs;
	enum dkio_state		state;
	get_status_t		get_status;
	cs_ddi_info_t	cs_ddi_info;


	/*
	 * Since PC Card driver encodes physical socket number
	 *	as part of the dev_t of device nodes instead of
	 *	the instance number we have to use csx_CS_DDI_Info(9F)
	 *	to get instance number that is associated with
	 *	a particular PC Card driver and socket number
	 *	for using in ddi_get_soft_state().
	 */
	cs_ddi_info.Socket = PCSRAM_SOCKET((dev_t)devp);
	cs_ddi_info.driver_name = pcsram_name;
	if (csx_CS_DDI_Info(&cs_ddi_info) != DDI_SUCCESS) {
		PCSRAM_DEBUG((CE_NOTE,
			"pcsram: ioctl, csx_CS_DDI_Info() failed\n"));
		return (ENXIO);
	}

	rs = ddi_get_soft_state(pcsram_soft_state_p,
					cs_ddi_info.instance);
	if (rs == NULL) {
		PCSRAM_DEBUG((CE_NOTE, "pcsram%d: ioctl, "
			"could not get soft state\n",
			cs_ddi_info.instance));
		return (ENXIO);
	}

	PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: ioctl, \n"
		"cmd 0x%x arg 0x%x mode 0x%x\n",
		cs_ddi_info.instance, rs->sn, cmd, arg, mode));

	switch (cmd) {

		case DKIOCGGEOM:
			/* return dk_geom structure */
			/* newfs does this first */
			if (rs->got_cardsize_flag) {
				bzero((caddr_t)&dkg,
					sizeof (struct dk_geom));
				dkg.dkg_ncyl  =
					rs->hdrv_chars->drv_ncyl;
				dkg.dkg_nhead =
					rs->hdrv_chars->drv_nhead;
				dkg.dkg_nsect =
					rs->hdrv_chars->drv_secptrack;
				dkg.dkg_pcyl =
					rs->hdrv_chars->drv_ncyl;
				if (ddi_copyout((caddr_t)&dkg,
						(caddr_t)arg,
						sizeof (struct dk_geom),
						mode) != 0) {
					return (EFAULT);
				}
				return (0);
			} else {
				return (EINVAL);
			}

		case DKIOCGVTOC:
			/* return vtoc structure. 1 partition. */
			/* newfs does this second */
			bzero((caddr_t)&vtoc, sizeof (struct vtoc));
			vtoc.v_sanity = VTOC_SANE;
			vtoc.v_version = V_VERSION;
			bcopy("pccard", vtoc.v_volume, 7);
			vtoc.v_sectorsz = DEV_BSIZE;
			vtoc.v_nparts = 1;
			vtoc.v_part[0].p_tag = V_UNASSIGNED;
			vtoc.v_part[0].p_flag = V_UNMNT;
			vtoc.v_part[0].p_start = (daddr_t)0;
			vtoc.v_part[0].p_size =
					rs->card_size / DEV_BSIZE;
			if (ddi_copyout((caddr_t)&vtoc, (caddr_t)arg,
			    sizeof (struct vtoc), mode) != 0) {
				return (EFAULT);
			}
			return (0);

		case DKIOCINFO:
			/* return dk_cinfo structure. */
			/* newfs does this third */
			bzero((caddr_t)&dkc, sizeof (struct dk_cinfo));

			(void) strcpy(dkc.dki_cname, "pcsram");
			dkc.dki_ctype = DKC_PCMCIA_MEM;

			/*
			 * The memory card can be used as a
			 *	pseudo floppy disk.  This is
			 *	required for mounting a PCFS
			 *	file system.
			 */
			dkc.dki_flags = DKI_PCMCIA_PFD;

			(void) strcpy(dkc.dki_dname, "pccard");

			dkc.dki_unit = (u_int)rs->sn;
			dkc.dki_maxtransfer = 1;

			if (ddi_copyout((caddr_t)&dkc, (caddr_t)arg,
			    sizeof (struct dk_cinfo), mode) != 0) {
				return (EFAULT);
			}
			return (0);

		case DKIOCSVTOC:
			/*
			 * Since this sample driver only supports
			 *	a single partition s2 so we do not
			 *	have to support the device labeling.
			 *	If this is a full disk implementation
			 *	then vtoc labeling must be supported.
			 */
			if (ddi_copyin((caddr_t)arg, (caddr_t)&vtoc,
					sizeof (struct vtoc), mode)) {
				return (EFAULT);
			}
			return (0);

		case DKIOCSTATE:
			if (ddi_copyin((caddr_t)arg, (caddr_t)&state,
						sizeof (int), mode)) {
				return (EFAULT);
			}

			/* Get card state status */
			if ((err = csx_GetStatus(rs->client_handle,
					&get_status)) != CS_SUCCESS) {
				error2text_t cft;

				cft.item = err;
				csx_Error2Text(&cft);
				PCSRAM_DEBUG((CE_CONT,
				    "pcsram%d.%d: ioctl, "
				    "GetStatus failed %s (0x%x)\n",
				    cs_ddi_info.instance, rs->sn,
				    cft.text, err));
				return (ENXIO);
			}

			/* Register the card state */
			if ((get_status.CardState &
					CS_EVENT_CARD_INSERTION)) {
				state = DKIO_INSERTED;
			} else {
				state = DKIO_EJECTED;
			}

			if (ddi_copyout((caddr_t)&state, (caddr_t)arg,
						sizeof (int), mode)) {
				return (EFAULT);
			}
			return (0);

		default:
			return (ENOTTY);

	}  /* switch (cmd) */
}


/*
 * pcsram_prop_op() - prop_op(9E) Property Management
 *
 *	All block drivers must support the "nblocks" property,
 *	and all character drivers must support the "size" property.
 *	Since we support hot plugging for many different memory
 *	card size, the nblocks and the size property are changed
 *	dynamically, so these properties should be maintained
 *	by the driver.
 *
 */
static int
pcsram_prop_op(dev_t devp, dev_info_t *dip, ddi_prop_op_t prop_op,
		int flags, char *name, caddr_t valuep, int *lengthp)
{
	int		val;
	int		length;
	int		kmem_flags;
	caddr_t		buffer;
	int		ret;
	pcsram_state_t	*rs;
	cs_ddi_info_t	cs_ddi_info;


	/* pass it on if it is not a "nblocks" and "size" property */
	if (strcmp(name, "nblocks") && strcmp(name, "size")) {
		return (ddi_prop_op(devp, dip, prop_op, flags,
					name, valuep, lengthp));
	}

	/*
	 * Since PC Card driver encodes physical socket number
	 *	as part of the dev_t of device nodes instead of
	 *	the instance number we have to use csx_CS_DDI_Info(9F)
	 *	to get instance number that is associated with
	 *	a particular PC Card driver and socket number
	 *	for using in ddi_get_soft_state().
	 */
	cs_ddi_info.Socket = PCSRAM_SOCKET((dev_t)devp);
	cs_ddi_info.driver_name = pcsram_name;
	if (csx_CS_DDI_Info(&cs_ddi_info) != DDI_SUCCESS) {
		PCSRAM_DEBUG((CE_NOTE,
			"pcsram: prop_op, csx_CS_DDI_Info() failed\n"));
		return (ENXIO);
	}

	rs = ddi_get_soft_state(pcsram_soft_state_p,
					cs_ddi_info.instance);
	if (rs == NULL) {
		PCSRAM_DEBUG((CE_NOTE, "pcsram%d: prop_op, "
			"could not get soft state\n",
			cs_ddi_info.instance));
		return (DDI_PROP_NOT_FOUND);
	}

	if (strcmp(name, "nblocks") == 0) {
		val = rs->card_size / DEV_BSIZE;
	} else {
		/* size */
		val = rs->card_size;
	}

	length = *lengthp;
	*lengthp = sizeof (int);

	if (prop_op == PROP_LEN) {
		return (DDI_PROP_SUCCESS);
	}

	switch (prop_op) {
	case PROP_LEN_AND_VAL_ALLOC:
		kmem_flags = KM_NOSLEEP;
		if (flags & DDI_PROP_CANSLEEP) {
			kmem_flags = KM_SLEEP;
		}
		buffer = (caddr_t)kmem_alloc(sizeof (int), kmem_flags);
		if (buffer == NULL) {
			cmn_err(CE_NOTE,
			    "pcsram%d.%d: no memory for property %s",
			    cs_ddi_info.instance, rs->sn, name);
			return (DDI_PROP_NO_MEMORY);
		}
		/* Set callers buf ptr */
		/* LINTED */
		*(caddr_t *)valuep = (caddr_t)buffer;
		break;

	case PROP_LEN_AND_VAL_BUF:
		if (sizeof (int) > length) {
			return (DDI_PROP_BUF_TOO_SMALL);
		}
		/* get callers buf ptr */
		buffer = (caddr_t)valuep;
		break;
	}

	/* LINTED */
	*((int *)buffer) = val;
	return (DDI_PROP_SUCCESS);
}


/*
 * Miscellaneous functions.
 *
 *	pcsram_update_mapmempage()
 */

/*
 * pcsram_update_mapmempage() - Updating window absolute offset
 *	from the beginning of the PC Card.
 */
static int
pcsram_update_mapmempage(pcsram_state_t *rs, int offset)
{

	int		ret;
	int		i;
	int		err;
	map_mem_page_t	map_mem_page;
	get_status_t	get_status;


	/* Get card state status */
	if ((err = csx_GetStatus(rs->client_handle, &get_status))
						!= CS_SUCCESS) {
		error2text_t cft;

		cft.item = err;
		csx_Error2Text(&cft);
		PCSRAM_DEBUG((CE_CONT,
			"pcsram%d.%d: update_mapmempage, "
			"GetStatus failed %s (0x%x)\n",
			rs->instance, rs->sn, cft.text, err));
		/* Let caller knows that there is some thing wrong */
		return (-1);
	}

	if (!(get_status.CardState & CS_EVENT_CARD_INSERTION)) {
		/* Let caller knows that there is no card */
		PCSRAM_DEBUG((CE_CONT,
			"pcsram%d.%d: update_mapmempage, "
			"Found no card\n", rs->instance, rs->sn));
		return (-1);
	}


	/*
	 * The "Page" field is used internally by CardService;
	 *	PC Card driver must set this field to zero
	 *	before calling this function.
	 *	See csx_MapMemPage(9F) manpage.
	 */
	map_mem_page.Page = 0;

	if (rs->win_size == 0) {
		PCSRAM_DEBUG((CE_CONT,
			"pcsram%d.%d: update_mapmempage, "
			"Found zero rs->win_size %d\n",
			rs->instance, rs->sn, (int)rs->win_size));
		/* To avoid zero division problem */
		return (-1);
	} else {
		/* setup for CardOffset of MapMemPage */
		i = offset / rs->win_size;
	}

	/*
	 * Now map the offset to the card
	 */
	map_mem_page.CardOffset = i * rs->win_size;

	offset -= (i * rs->win_size);

	if ((ret = csx_MapMemPage(rs->window_handle, &map_mem_page))
						!= CS_SUCCESS) {
		error2text_t cft;

		cft.item = ret;
		csx_Error2Text(&cft);
		PCSRAM_DEBUG((CE_CONT,
			"pcsram%d.%d: update_mapmempage, "
			"MapMemPage failed %s (0x%x)\n",
			rs->instance, rs->sn, cft.text, err));
		if ((ret = csx_ReleaseWindow(rs->window_handle))
						!= CS_SUCCESS) {
			error2text_t cft;

			cft.item = ret;
			csx_Error2Text(&cft);
			PCSRAM_DEBUG((CE_CONT,
				"pcsram%d.%d: update_mapmempage, "
				"ReleaseWindow failed %s (0x%x)\n",
				rs->instance, rs->sn, cft.text, err));
		}

		/* Let caller knows that there is some thing wrong */
		return (-1);
	}

	return (offset);
}


/*
 * CardServices related functions.
 *
 *	pcsram_event()
 *	pcsram_card_insertion()
 *	pcsram_card_removal()
 *	pcsram_parse_cis()
 */

/*
 * pcsram_event - This is client event callback handler
 *	entry point which is passed in the event_handler
 *	field of csx_RegisterClient() function.
 */
static int
pcsram_event(event_t event, int priority, event_callback_args_t *eca)
{
	int		retcode = CS_UNSUPPORTED_EVENT;
	pcsram_state_t	*rs = eca->client_data;
	client_info_t	*ci = &eca->client_info;


	/*
	 * If the priority argument indicates CS_EVENT_PRI_HIGH,
	 *	the callback handler is being called at the nexus
	 *	event interrupt priority which is above lock-level.
	 *	You can not use a mutex defined at a lower priority
	 *	for the high priority event.  At the same time,
	 *	you do not want to use the high priority mutex
	 *	when it is not necessary since it blocks out other
	 *	interrupts while held.
	 */
	if (priority & CS_EVENT_PRI_HIGH) {
		/*
		 * Use rs->event_hilock mutex for a high priority
		 *	event to modify the card event state.
		 */
		mutex_enter(&rs->event_hilock);
	} else {
		/*
		 * Use rs->mutex for a low priority event
		 *	to modify the driver state structure.
		 */
		mutex_enter(&rs->mutex);
	}

	{
		event2text_t	event2text;

		event2text.event = event;
		csx_Event2Text(&event2text);
		PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: event, \n"
			"\tevent %s (0x%x) priority 0x%x\n",
			rs->instance, rs->sn, event2text.text,
			event, priority));
	}

	/*
	 * Find out which event we got and do the appropriate thing
	 */
	switch (event) {

		/*
		 * When the driver register with CardServices
		 *	through csx_RegisterClient() function,
		 *	CardServices saves the client information
		 *	and immediately returns to the client.
		 *	CardServices then performs the registration
		 *	in the background.  When the registration
		 *	processing is completed, CardServices notifies
		 *	the request client through pcsram_event with
		 *	with CS_EVENT_REGISTRATION_COMPLETE event.
		 */
		case CS_EVENT_REGISTRATION_COMPLETE:
			PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: event, "
				"Registration is completed\n",
				rs->instance, rs->sn));
			retcode = CS_SUCCESS;
			break;

		/*
		 * This event indicates a PC Card has been inserted
		 *	in a socket or a client has just registered
		 *	for insertion events.
		 * It is available as a low priority event.
		 */
		case CS_EVENT_CARD_INSERTION:
			if (priority & CS_EVENT_PRI_LOW) {
				retcode = pcsram_card_insertion(rs);
			}
			break;

		/*
		 * This event indicates the battery on a PC Card
		 *	is weak and should be replaced.  Data
		 *	integrity on the PC Card is still assured.
		 */
		case CS_EVENT_BATTERY_LOW:
			/*
			 * The battery LOW condition is also
			 *	checked in pcsam_open() routine.
			 */
			cmn_err(CE_WARN, "pcsram%d.%d: event, "
				"Battery is LOW and "
				"should be replaced.\n",
				rs->instance, rs->sn);
			retcode = CS_SUCCESS;
			break;

		/*
		 * This event indicates the battery on a PC Card
		 *	is no longer serviceable and data
		 *	may be lost.
		 */
		case CS_EVENT_BATTERY_DEAD:
			/*
			 * The battery LOW condition is also
			 *	checked in pcsam_open() routine.
			 */
			cmn_err(CE_WARN, "pcsram%d.%d: event, "
				"Battery is DEAD, data may be lost.\n",
				rs->instance, rs->sn);
			retcode = CS_SUCCESS;
			break;

		/*
		 * This event indicates the write protect status
		 *	of the PC Card in the socket has changed.
		 *
		 * NOTE:  Not all socket hardware is capable of
		 *	reporting a change in a PC Card's write
		 *	protect status.  For this reason, a client
		 *	should not rely on a CS_EVENT_WRITE_PROTECT
		 *	notification as the sole method of
		 *	determining a PC Card's write protect status.
		 *
		 */
		case CS_EVENT_WRITE_PROTECT:
			/*
			 * The write protect condition is also
			 *	checked in pcsam_open() routine.
			 */
			if (eca->info) {
				cmn_err(CE_CONT, "pcsram%d.%d: event, "
					"Write protect status of PC "
					" Card has changed.\n",
					rs->instance, rs->sn);
			}
			break;

		/*
		 * This event indicates a PC Card has been removed
		 *	from a socket.
		 * It is available as both a low priority
		 *	and high priority event.  As a high priority
		 *	event the client is expected to set the
		 *	internal state the the card has been removed,
		 *	and that the event at low priority will be
		 *	used to release any resources that is allocated
		 *	by the client.
		 */
		case CS_EVENT_CARD_REMOVAL_LOWP:
		case CS_EVENT_CARD_REMOVAL:
			if (priority & CS_EVENT_PRI_HIGH) {
				retcode = CS_SUCCESS;
				rs->card_event &= ~PCSRAM_CARD_IS_READY;
			} else {
				retcode = pcsram_card_removal(rs);
				rs->card_event &= ~PCSRAM_CARD_IS_READY;
			}
			break;

		/*
		 * This event requests that the client return its
		 *	client information data such as the driver
		 *	revision level and date.
		 */
		case CS_EVENT_CLIENT_INFO:
			if (GET_CLIENT_INFO_SUBSVC(ci->Attributes) ==
					CS_CLIENT_INFO_SUBSVC_CS) {
				ci->Attributes |= CS_CLIENT_INFO_VALID;
				ci->Revision = PCSRAM_REV_LEVEL;
				ci->CSLevel = CS_VERSION;
				ci->RevDate = PCSRAM_REV_DATE;
				strcpy(ci->ClientName,
					PCSRAM_CLIENT_DESCRIPTION);
				strcpy(ci->VendorName,
					PCSRAM_VENDOR_DESCRIPTION);
				/*
				 * Note that PC Card driver does not
				 *	need to initialize DriverName
				 *	field since CardService will
				 *	copy the name of PC card driver
				 *	into this space after the PC
				 *	Card driver has sucessfully
				 *	processed this event.
				 */
				retcode = CS_SUCCESS;
			} /* CS_CLIENT_INFO_SUBSVC_CS */
			break;

		default:
			retcode = CS_UNSUPPORTED_EVENT;
			break;

	} /* switch(event) */

	if (priority & CS_EVENT_PRI_HIGH) {
		mutex_exit(&rs->event_hilock);
	} else {
		mutex_exit(&rs->mutex);
	}

	return (retcode);
}


/*
 * pcsram_card_insertion() - Allocating the system resources.
 *	It is called from pcsram_event().
 *	Request a memory window to be assigned to
 *		Common Memory space of PC Card.
 *	Start to parse CIS to determine the card size and speed.
 *	Creating the minor devices.
 *	Note that the resource is de-allocated when
 *		the card is removed and re-allocated
 *		when the card is inserted again.
 */
static int
pcsram_card_insertion(pcsram_state_t *rs)
{
	int		ret;
	int		rval = CS_SUCCESS;
	sockevent_t	se;
	win_req_t	win_req;
	convert_speed_t	convert_speed;
	map_mem_page_t	map_mem_page;
	get_status_t	get_status;


	PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: card_insertion\n",
					rs->instance, rs->sn));

	/* Reset battery DEAD/LOW status posted flag */
	rs->battery_dead_posted = 0;
	rs->battery_low_posted = 0;

	rs->got_cardsize_flag = 0;

	rs->card_eject_posted = 0;

	/* Get card state status */
	if ((ret = csx_GetStatus(rs->client_handle, &get_status))
						!= CS_SUCCESS) {
		error2text_t cft;

		cft.item = ret;
		csx_Error2Text(&cft);
		PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: card_insertion, "
			"GetStatus failed %s (0x%x)\n",
			rs->instance, rs->sn, cft.text, ret));
		cv_signal(&rs->readywait_cv);
		return (ret);
	}

#if defined(DEBUG)
	cmn_err(CE_CONT, "pcsram%d.%d: card_insertion, "
		"GetStatus returns:\n", rs->instance, rs->sn);
	pcsram_display_card_status(&get_status);
#endif

	/* Make sure that there is a card in the socket */
	if (!(get_status.CardState & CS_EVENT_CARD_INSERTION)) {
		cv_signal(&rs->readywait_cv);
		return (CS_NO_CARD);
	}

	/*
	 * Set up the client event mask to give us WP and battery
	 *	events as well as what other events we have already
	 *	registered for.
	 * Note that since we set the global event mask in the call
	 *	to RegisterClient in pcsram_attach, we don't have to
	 *	duplicate those events in this event mask.
	 */
	se.Attributes = CONF_EVENT_MASK_CLIENT;
	if ((ret = csx_GetEventMask(rs->client_handle, &se))
						!= CS_SUCCESS) {
		error2text_t cft;

		cft.item = ret;
		csx_Error2Text(&cft);
		PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: card_insertion, "
			"GetEventMask failed %s (0x%x)\n",
			rs->instance, rs->sn, cft.text, ret));
		cv_signal(&rs->readywait_cv);
		return (ret);
	}

	se.EventMask |= (CS_EVENT_BATTERY_LOW |
				CS_EVENT_BATTERY_DEAD |
				CS_EVENT_WRITE_PROTECT);

	if ((ret = csx_SetEventMask(rs->client_handle, &se))
						!= CS_SUCCESS) {
		error2text_t cft;

		cft.item = ret;
		csx_Error2Text(&cft);
		PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: card_insertion, "
			"SetEventMask failed %s (0x%x)\n",
			rs->instance, rs->sn, cft.text, ret));
		cv_signal(&rs->readywait_cv);
		return (ret);
	}

	/*
	 * If we had a window, make sure to release
	 * before we request for the new window
	 */
	if (rs->flags & PCSRAM_HAS_WINDOW) {
		rs->flags &= ~PCSRAM_HAS_WINDOW;

		if ((ret = csx_ReleaseWindow(rs->window_handle))
						!= CS_SUCCESS) {
			error2text_t cft;

			cft.item = ret;
			csx_Error2Text(&cft);
			PCSRAM_DEBUG((CE_CONT,
				"pcsram%d.%d: card_insertion, "
				"ReleaseWindow failed %s (0x%x)\n",
				rs->instance, rs->sn, cft.text, ret));
		}
	}

	/*
	 * Try to get a memory window to Common Memory space
	 */
	win_req.Attributes = (WIN_MEMORY_TYPE_CM |
				WIN_DATA_WIDTH_16 |
				WIN_ENABLE |
				WIN_ACC_NEVER_SWAP |
				WIN_ACC_STRICT_ORDER);
	win_req.Base.base = 0;	/* let CS find us a base address */
	win_req.Size = 0;	/* let CS return the smallest size */
				/* window it finds */

	convert_speed.Attributes = CONVERT_NS_TO_DEVSPEED;
	convert_speed.nS = 250;
	csx_ConvertSpeed(&convert_speed);

	win_req.win_params.AccessSpeed = convert_speed.devspeed;

	if ((ret = csx_RequestWindow(rs->client_handle,
			&rs->window_handle, &win_req)) != CS_SUCCESS) {
		error2text_t cft;

		cft.item = ret;
		csx_Error2Text(&cft);
		PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: card_insertion, "
			"RequestWindow failed %s (0x%x)\n",
			rs->instance, rs->sn, cft.text, ret));
		cv_signal(&rs->readywait_cv);
		return (ret);
	}

	rs->flags |= PCSRAM_HAS_WINDOW;

	/*
	 * Now map the offset to the start of the card
	 */
	map_mem_page.CardOffset = 0;
	map_mem_page.Page = 0;

	if ((ret = csx_MapMemPage(rs->window_handle,
				&map_mem_page)) != CS_SUCCESS) {
		error2text_t cft;

		cft.item = ret;
		csx_Error2Text(&cft);
		rs->flags &= ~PCSRAM_HAS_WINDOW;
		PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: card_insertion, "
			"MapMemPage failed %s (0x%x)\n",
			rs->instance, rs->sn, cft.text, ret));

		if ((ret = csx_ReleaseWindow(rs->window_handle))
						!= CS_SUCCESS) {
			error2text_t cft;

			cft.item = ret;
			csx_Error2Text(&cft);
			PCSRAM_DEBUG((CE_CONT,
				"pcsram%d.%d: card_insertion, "
				"ReleaseWindow failed %s (0x%x)\n",
				rs->instance, rs->sn, cft.text, ret));
		}
		cv_signal(&rs->readywait_cv);
		return (ret);
	}

	/* Store rs->access_handle */
	rs->access_handle = win_req.Base.handle;
	rs->win_size = win_req.Size;

	/* Read CIS information for the card size */
	if (pcsram_parse_cis(rs) <= 0) {
		rval = CS_BAD_SOCKET;
	} else {
		/* Get cylinder information */
		*rs->hdrv_chars = hdtypes;
		rs->hdrv_chars->drv_ncyl =
			PCSRAM_GET_NCYL(rs->card_size,
					rs->hdrv_chars->drv_nhead,
					rs->hdrv_chars->drv_sec_size,
					rs->hdrv_chars->drv_secptrack);
		/*
		 * Create the device nodes.
		 *	This is an example that assumes
		 *	that we want to create two devices
		 *	for this instance.
		 */
		{
			int			n;
			char			devname[80];
			char			*dname;
			devnode_desc_t		*dnd;
			make_device_node_t	make_device_node;


			make_device_node.Action = CREATE_DEVICE_NODE;
			/*
			 * This is for both raw and block device
			 *	of a single slice s2
			 */
			make_device_node.NumDevNodes = 2;

			make_device_node.devnode_desc =
			    kmem_zalloc(sizeof (struct devnode_desc) *
			    make_device_node.NumDevNodes, KM_SLEEP);

			dname = devname;

			/*
			 * Create only "c" partition for now since
			 *	the memory card can be used as
			 *	pseudo-floppy disks.
			 */

			for (n = 0; n < (make_device_node.NumDevNodes);
								n++) {

				dnd = &make_device_node.devnode_desc[n];

				dnd->name = dname;
				dnd->minor_num = PCSRAM_SETMINOR(rs->sn,
							(n+4)/2);

				if (n&1) {
					dnd->spec_type = S_IFCHR;
					sprintf(dname, "%c,raw",
						(((n+4)/2)+'a'));
				} else {
					dnd->spec_type = S_IFBLK;
					sprintf(dname, "%c",
						(((n+4)/2)+'a'));
				}

				dnd->node_type = DDI_NT_BLOCK;

				dname = &dname[strlen(dname)+1];
			} /* for */

			if ((ret = csx_MakeDeviceNode(rs->client_handle,
				&make_device_node)) != CS_SUCCESS) {

				error2text_t cft;

				cft.item = ret;
				csx_Error2Text(&cft);
				PCSRAM_DEBUG((CE_CONT,
				    "pcsram%d.%d: card_insertion, "
				    "MakeDeviceNode failed %s (0x%x)\n",
				    rs->instance, rs->sn,
				    cft.text, ret));
			}

			/*
			 * We do not need this structure anymore
			 *	since we've created the devices.
			 *	If we need to keep track of the
			 *	devices that we've created for some
			 *	reason, then we want to keep this
			 *	structure and the make_device_node_t
			 *	structure around in a per-instance
			 *	data area.
			 */
			kmem_free(make_device_node.devnode_desc,
					sizeof (struct devnode_desc) *
					make_device_node.NumDevNodes);

			make_device_node.devnode_desc = NULL;

			rs->flags |= PCSRAM_MAKEDEVICENODE;

		} /* create device nodes */

		rs->got_cardsize_flag = 1;

	}  /* if (pcsram_parse_cis) */

	rs->card_event |= PCSRAM_CARD_IS_READY;

	/*
	 * Wake up readywait_cv in pcsram_open()
	 */
	cv_signal(&rs->readywait_cv);

	return (rval);
}



/*
 * pcsram_card_removal() - Releasing the system resources.
 *      It is called from either pcsram_detach() or pcsram_event().
 *      Setting up the client event mask so that we will not
 *              receive the card ready events.
 *      Removing the minor devices.
 *
 *	Note that the resource is de-allocated when
 *		the card is removed and re-allocated
 *		when the card is inserted again.
 */
static int
pcsram_card_removal(pcsram_state_t *rs)
{
	int		ret;
	get_status_t	get_status;
	sockevent_t	se;


	PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: card_removal\n",
					rs->instance, rs->sn));

	/* Reset battery DEAD/LOW status posted flag */
	rs->battery_dead_posted = 0;
	rs->battery_low_posted = 0;

	/* Misc. flags */
	rs->busy =  0;
	rs->card_eject_posted = 0;

	/*
	 * Remove all the device nodes.  We don't have to explictly
	 *	specify the names if we want Card Services to remove
	 *	all of the devices.
	 * Note that when you call MakeDeviceNode with the Action
	 *	argument set to REMOVAL_ALL_DEVICE_NODES, the
	 *	NumDevNodes must be zero.
	 */
	if (rs->flags & PCSRAM_MAKEDEVICENODE) {
		remove_device_node_t	remove_device_node;

		remove_device_node.Action = REMOVAL_ALL_DEVICE_NODES;
		remove_device_node.NumDevNodes = 0;

		if ((ret = csx_RemoveDeviceNode(rs->client_handle,
			&remove_device_node)) != CS_SUCCESS) {

			error2text_t cft;

			cft.item = ret;
			csx_Error2Text(&cft);
			PCSRAM_DEBUG((CE_CONT,
				"pcsram%d.%d: card_removal, "
				"RemoveDeviceNode failed %s (0x%x)\n",
				rs->instance, rs->sn, cft.text, ret));
		}
	} /* remove device nodes */

	/* Get card state status */
	if ((ret = csx_GetStatus(rs->client_handle, &get_status))
						!= CS_SUCCESS) {
		error2text_t cft;

		cft.item = ret;
		csx_Error2Text(&cft);
		PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: card_removal, "
			"GetStatus failed %s (0x%x)\n",
			rs->instance, rs->sn, cft.text, ret));
		return (ret);
	}

#if defined(DEBUG)
	cmn_err(CE_CONT, "pcsram%d.%d: card_removal, "
		"GetStatus returns:\n", rs->instance, rs->sn);
	pcsram_display_card_status(&get_status);
#endif

	/*
	 * Release the window if we allocated one
	 */
	if (rs->flags & PCSRAM_HAS_WINDOW) {
		rs->flags &= ~PCSRAM_HAS_WINDOW;

		if ((ret = csx_ReleaseWindow(rs->window_handle))
						!= CS_SUCCESS) {
			error2text_t cft;

			cft.item = ret;
			csx_Error2Text(&cft);
			PCSRAM_DEBUG((CE_CONT,
				"pcsram%d.%d: card_removal, "
				"ReleaseWindow failed %s (0x%x)\n",
				rs->instance, rs->sn, cft.text, ret));
			return (ret);
		}
	}

	/*
	 * Set up the client event mask to clear WP and battery
	 *	events as well as what other events we have already
	 *	registered for.
	 * Note that since we set the global event mask in the call
	 *	to RegisterClient in pcsram_attach, we don't have to
	 *	duplicate those events in this event mask.
	 */
	se.Attributes = CONF_EVENT_MASK_CLIENT;
	if ((ret = csx_GetEventMask(rs->client_handle, &se))
						!= CS_SUCCESS) {
		error2text_t cft;

		cft.item = ret;
		csx_Error2Text(&cft);
		PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: card_removal, "
			"GetEventMask failed %s (0x%x)\n",
			rs->instance, rs->sn, cft.text, ret));
		return (ret);
	}

	se.EventMask &= ~(CS_EVENT_BATTERY_LOW |
				CS_EVENT_BATTERY_DEAD |
				CS_EVENT_WRITE_PROTECT);

	if ((ret = csx_SetEventMask(rs->client_handle, &se))
						!= CS_SUCCESS) {
		error2text_t cft;

		cft.item = ret;
		csx_Error2Text(&cft);
		PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: card_removal, "
			"SetEventMask failed %s (0x%x)\n",
			rs->instance, rs->sn, cft.text, ret));
		return (ret);
	}

	return (CS_SUCCESS);
}


/*
 * pcsram_parse_cis() - gets Card Structure Information (CIS) to
 *	configure the card by using csx_GetFirstTuple(9F) and
 *	csx_GetNextTuple(9F) to find the desired tuple.
 *	And using csx_Parse_CISTPL_DEVICE(9F) call to
 *	process the tuple.
 * This routine is called from pcsram_card_insertion().
 *
 * Return Codes
 *
 *	0	for bad socket or can not find a CIS or
 *		a desired tuple.
 *	1	successful
 *
 */
static int
pcsram_parse_cis(pcsram_state_t *rs)
{
	int		ret;
	int		found;
	cisinfo_t	cisinfo;
	cisdata_t	device_tuple;
	tuple_t		tuple;


	PCSRAM_DEBUG((CE_CONT, "pcsram%d.%d: parse_cis\n",
					rs->instance, rs->sn));

	/* Check for a CIS on this card */
	if ((ret = csx_ValidateCIS(rs->client_handle, &cisinfo))
						!= CS_SUCCESS) {
		error2text_t cft;

		cft.item = ret;
		csx_Error2Text(&cft);

		if (ret != CS_NO_CIS) {
			PCSRAM_DEBUG((CE_CONT,
				"pcsram%d.%d: parse_cis, "
				"ValidateCIS failed %s (0x%x)\n",
				rs->instance, rs->sn, cft.text, ret));
			return (0);
		} else {
			/* No CIS on card - Unsupported SRAM card */
			cmn_err(CE_CONT, "?pcsram%d.%d: "
				"Found unsupported memory "
				"card with no CIS\n",
				rs->instance, rs->sn);
			return (0);
		} /* CS_NO_CIS */

	} else {

		/* There is a CIS - Read this card size information */
		device_tuple = CISTPL_DEVICE;

		/*
		 * Search for the first device tuple
		 *	if it's not found, there's no point
		 *	in searching for anything else.
		 */
		tuple.Attributes = 0;
		tuple.DesiredTuple = device_tuple;
		if ((ret = csx_GetFirstTuple(rs->client_handle, &tuple))
						!= CS_SUCCESS) {
			return (0);
		}

		/*
		 * Got the device tuple, now parse it.
		 */
		found = 0;
		do {
			int			i;
			cistpl_device_t		cistpl_device;

			bzero((caddr_t)&cistpl_device,
				sizeof (struct cistpl_device_t));

			if (csx_Parse_CISTPL_DEVICE(rs->client_handle,
					&tuple, &cistpl_device)
						!= CS_SUCCESS) {
				PCSRAM_DEBUG((CE_CONT,
					"pcsram%d.%d: parse_cis, "
					"Can not find CISTPL_DEVICE\n",
					rs->instance, rs->sn));
				continue;
			}

			/* should see at least one CISTPL_DEVICE */
			if (!cistpl_device.num_devices) {
				PCSRAM_DEBUG((CE_CONT,
					"pcsram%d.%d: parse_cis, "
					"Found CISTPL_DEVICE tuple, "
					"but num_devices is zero\n",
					rs->instance, rs->sn));
				continue;
			}

			found = 0;
			for (i = 0; i < cistpl_device.num_devices;
								i++) {
			    cistpl_device_node_t *cistpl_device_node;

			    cistpl_device_node =
					&cistpl_device.devnode[i];

			    if (cistpl_device_node->type ==
					CISTPL_DEVICE_DTYPE_SRAM) {

				/* Support this device */
				rs->card_size =
				    cistpl_device_node->size_in_bytes;
				found = 1;
				break;
			    }

			} /* for (cistpl_device_node->num_devices) */

			if (found) break;

		} while ((ret = csx_GetNextTuple(rs->client_handle,
						&tuple)) == CS_SUCCESS);

		if (found == 0) {
			cmn_err(CE_CONT, "?pcsram%d.%d: "
				"Found unsupported "
				"memory card at socket %d\n",
				rs->instance, rs->sn);
			return (0);
		}

	}  /* if (csx_ValidateCIS) */

	cmn_err(CE_CONT, "?pcsram%d.%d: (CIS) socket %d card size %d\n",
				rs->instance, rs->sn, rs->card_size);
	return (1);
}

#if defined(DEBUG)
/*
 * pcsram_display_card_status()
 *	Wrapper for CardServices GetStatus function
 */
static void
pcsram_display_card_status(get_status_t *gs)
{
	event2text_t	event2text;


	event2text.event = gs->CardState;
	csx_Event2Text(&event2text);
	cmn_err(CE_CONT, "\tCardState [%s]\n", event2text.text);

	event2text.event = gs->SocketState;
	csx_Event2Text(&event2text);
	cmn_err(CE_CONT, "\tSocketState [%s]\n", event2text.text);
}
#endif
