/*
 * Copyright (c) 1992-1997, by Sun Microsytems, Inc.
 * All rights reserved.
 */

/*
 * Sample programmed I/O (pio) driver for Solaris
 */

/*
 * This file is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify this file without charge, but are not authorized to
 * license or distribute it to anyone else except as part of a product
 * or program developed by the user.
 *
 * THIS FILE IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 *
 * This file is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 *
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY THIS FILE
 * OR ANY PART THEREOF.
 *
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even
 * if Sun has been advised of the possibility of such damages.
 *
 * Sun Microsystems, Inc.
 * 901 San Antonio Rd
 * Palo Alto, California 94303
 */
#pragma ident	"@(#)pio.c	1.7	97/09/25 Sun Microsystems Inc."

/*
 * This driver is for a fictional device that accepts one character
 * at a time and generates an interrupt when it is ready for another.
 * This device implements two register sets, the first an 8-bit CSR
 * register and and the second an 8-bit data register.
 *
 * This driver is not intended to be efficient or bulletproof; it is
 * intended to show how to write a simple DDI/DKI-compliant driver.
 */

#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>

#include "pio_io.h"
#include "pio_reg.h"

static	void *state_head;	/* opaque handle top of state structs */

/*
 * These are the entry points into our driver that are called when the
 * driver is loaded, during a system call, or in response to an interrupt.
 */
static	int	pio_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg,
		    void **result);
static	int	pio_probe(dev_info_t *dip);
static	int	pio_attach(dev_info_t *dip, ddi_attach_cmd_t cmd);
static	int	pio_open(dev_t *dev, int openflags, int otyp, cred_t *credp);
static	int	pio_close(dev_t dev, int openflags, int otyp, cred_t *credp);
static	int	pio_write(dev_t dev, struct uio *uiop, cred_t *credp);
static	int	pio_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
		    cred_t *credp, int *rvalp);
static	u_int	pio_intr(caddr_t instance);
static	int	pio_detach(dev_info_t *dip, ddi_detach_cmd_t cmd);

/*
 * When our driver is loaded or unloaded, the system calls our _init or
 * _fini routine with a modlinkage structure.  The modlinkage structure
 * contains:
 *
 *	modlinkage->
 *		modldrv->
 *			dev_ops->
 * 				cb_ops
 *
 * cb_ops contains the normal driver entry points and is roughly equivalent
 * to the cdevsw & bdevsw structures in previous releases.
 *
 * dev_ops contains, in addition to the pointer to cb_ops, the routines
 * that support loading and unloading our driver.
 */
static struct cb_ops	pio_cb_ops = {
	pio_open,
	pio_close,
	nodev,			/* not a block driver	*/
	nodev,			/* no print routine	*/
	nodev,			/* no dump routine	*/
	nodev,			/* no read routine; write-only device */
	pio_write,
	pio_ioctl,
	nodev,			/* no devmap routine	*/
	nodev,			/* no mmap routine	*/
	nodev,			/* no segmap routine	*/
	nochpoll,		/* no chpoll routine	*/
	ddi_prop_op,
	0,			/* not a STREAMS driver	*/
	D_NEW | D_MP,		/* safe for multi-thread/multi-processor */
};

static struct dev_ops pio_ops = {
	DEVO_REV,		/* devo_rev */
	0,			/* devo_refcnt */
	pio_getinfo,		/* devo_getinfo */
	nulldev,		/* devo_identify */
	pio_probe,		/* devo_probe */
	pio_attach,		/* devo_attach */
	pio_detach,		/* devo_detach */
	nodev,			/* devo_reset */
	&pio_cb_ops,		/* devo_cb_ops */
	(struct bus_ops *)0,	/* devo_bus_ops */
	NULL,			/* devo_power */
};

extern	struct	mod_ops mod_driverops;

static	struct modldrv modldrv = {
	&mod_driverops,
	"SUNW sample PIO driver 1.7",
	&pio_ops,
};

static	struct modlinkage modlinkage = {
	MODREV_1,		/* MODREV_1 indicated by manual */
	(void *)&modldrv,
	NULL,			/* termination of list of linkage structures */
};

/*
 * _init, _info, and _fini support loading and unloading the driver.
 */
int
_init(void)
{
	int	error;

	if ((error = ddi_soft_state_init(&state_head, sizeof (Pio), 1)) != 0)
		return (error);

	if ((error = mod_install(&modlinkage)) != 0)
		ddi_soft_state_fini(&state_head);

	return (error);
}

int
_info(struct modinfo *modinfop)
{
	return (mod_info(&modlinkage, modinfop));
}

int
_fini(void)
{
	int status;

	if ((status = mod_remove(&modlinkage)) != 0)
		return (status);

	ddi_soft_state_fini(&state_head);

	return (status);
}


/*ARGSUSED*/
static int
pio_probe(dev_info_t *dip)
{
	/*
	 * We have no way to check if the pio device actually exists,
	 * so return DONTCARE, which will cause the system to proceed
	 * as if the device *does* exist.
	 */
	return (DDI_PROBE_DONTCARE);
}

/*
 * Attach an instance of the driver.  We take all the knowledge we
 * have about our board and check it against what has been filled in for
 * us from our FCode or from our driver.conf(4) file.
 */
static int
pio_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
	int			instance;
	Pio			*pio_p;
	u_int			result;
	int			error;
	ddi_device_acc_attr_t	dev_acc_attr;

	switch (cmd) {
	case DDI_ATTACH:
		break;

	default:
		return (DDI_FAILURE);
	}

	instance = ddi_get_instance(dip);

	if (ddi_soft_state_zalloc(state_head, instance) != 0)
		return (DDI_FAILURE);

	pio_p = (Pio *) ddi_get_soft_state(state_head, instance);
	ddi_set_driver_private(dip, (caddr_t)pio_p);
	pio_p->dip = dip;

	/*
	 * we only use one interrupt level, level 0
	 *
	 * We make no attempt to handle hi-level interrupts.
	 */
	if (ddi_intr_hilevel(dip, 0)) {
		ddi_soft_state_free(state_head, instance);
		return (DDI_FAILURE);
	}

	/*
	 * Before adding the interrupt, get the interrupt block
	 * cookie associated with the interrupt specification, to
	 * be used to initialize the mutex used by the interrupt
	 * handler.
	 */
	if (ddi_get_iblock_cookie(dip, 0,
	    &pio_p->iblock_cookie) != DDI_SUCCESS) {
		ddi_soft_state_free(state_head, instance);
		return (DDI_FAILURE);
	}

	mutex_init(&pio_p->mutex, "pio mutex",
		MUTEX_DRIVER, (void *)pio_p->iblock_cookie);

	/*
	 * Now that the mutex is initialized, we can add
	 * the interrupt itself.
	 */
	if (ddi_add_intr(dip, 0, &pio_p->iblock_cookie,
	    (ddi_idevice_cookie_t *)NULL, pio_intr,
	    (caddr_t)instance) != DDI_SUCCESS) {
		mutex_destroy(&pio_p->mutex);
		ddi_soft_state_free(state_head, instance);
		return (DDI_FAILURE);
	}

	/*
	 * we have one CSR register and one data register
	 * make sure nreg==2 and that the registers are the correct sizes
	 */
	error = ddi_dev_nregs(dip, (int *)&result);
	if (error == DDI_FAILURE || result != 2) {
		mutex_destroy(&pio_p->mutex);
		ddi_remove_intr(pio_p->dip, 0, pio_p->iblock_cookie);
		ddi_soft_state_free(state_head, instance);
		return (DDI_FAILURE);
	}

	if (ddi_dev_regsize(dip, 0, (off_t *)&result) != DDI_SUCCESS) {
		return (DDI_FAILURE);
	}

	if (result != sizeof (Pio_csr)) {
		mutex_destroy(&pio_p->mutex);
		ddi_remove_intr(pio_p->dip, 0, pio_p->iblock_cookie);
		ddi_soft_state_free(state_head, instance);
		return (DDI_FAILURE);
	}

	if (ddi_dev_regsize(dip, 1, (off_t *)&result) != DDI_SUCCESS) {
		return (DDI_FAILURE);
	}

	if (result != sizeof (u_char)) {
		mutex_destroy(&pio_p->mutex);
		ddi_remove_intr(pio_p->dip, 0, pio_p->iblock_cookie);
		ddi_soft_state_free(state_head, instance);
		return (DDI_FAILURE);
	}

	/*
	 * the registers look OK, so let's map them in so we can use them
	 */

	/*
	 * Initialize the device access attributes for the register
	 * mapping
	 */
	dev_acc_attr.devacc_attr_version = DDI_DEVICE_ATTR_V0;
	dev_acc_attr.devacc_attr_endian_flags = DDI_STRUCTURE_LE_ACC;
	dev_acc_attr.devacc_attr_dataorder = DDI_STRICTORDER_ACC;

	/*
	 * Map in the csr register
	 */
	if (ddi_regs_map_setup(dip, 0, (caddr_t *)&(pio_p->csr), 0,
	    sizeof (Pio_csr), &dev_acc_attr, &pio_p->csr_handle) !=
	    DDI_SUCCESS) {
		mutex_destroy(&pio_p->mutex);
		ddi_remove_intr(pio_p->dip, 0, pio_p->iblock_cookie);
		ddi_soft_state_free(state_head, instance);
		return (DDI_FAILURE);
	}

	/*
	 * Map in the data register
	 */
	if (ddi_regs_map_setup(dip, 1, (caddr_t *)&(pio_p->data), 0,
	    sizeof (u_char), &dev_acc_attr, &pio_p->data_handle) !=
	    DDI_SUCCESS) {
		mutex_destroy(&pio_p->mutex);
		ddi_remove_intr(pio_p->dip, 0, pio_p->iblock_cookie);
		ddi_soft_state_free(state_head, instance);
		ddi_regs_map_free(&pio_p->csr_handle);
		return (DDI_FAILURE);
	}

	/*
	 * ddi_create_minor_node creates an entry in an internal kernel
	 * table; the actual entry in the file system is created by
	 * drvconfig(1) when you run add_drv(1);
	 */
	if (ddi_create_minor_node(dip, ddi_get_name(dip), S_IFCHR, instance,
	    DDI_PSEUDO, 0) == DDI_FAILURE) {
		mutex_destroy(&pio_p->mutex);
		ddi_remove_intr(pio_p->dip, 0, pio_p->iblock_cookie);
		ddi_soft_state_free(state_head, instance);
		ddi_regs_map_free(&pio_p->csr_handle);
		ddi_regs_map_free(&pio_p->data_handle);
		return (DDI_FAILURE);
	}

	/*
	 * reset device (including disabling interrupts)
	 */
	ddi_put8(pio_p->csr_handle, pio_p->csr, PIO_RESET);

	ddi_report_dev(dip);

	return (DDI_SUCCESS);
}

/*
 * This is a pretty generic getinfo routine as describe in the manual.
 */
/*ARGSUSED*/
static int
pio_getinfo(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg, void **result)
{
	int 	error;
	Pio	*pio_p;

	switch (infocmd) {
	case DDI_INFO_DEVT2DEVINFO:
		pio_p = (Pio *) ddi_get_soft_state(state_head,
			    getminor((dev_t)arg));
		if (pio_p == NULL) {
			*result = NULL;
			error = DDI_FAILURE;
		} else {
			mutex_enter(&pio_p->mutex);
			*result = pio_p->dip;
			mutex_exit(&pio_p->mutex);
			error = DDI_SUCCESS;
		}
		break;
	case DDI_INFO_DEVT2INSTANCE:
		*result = (void *)getminor((dev_t)arg);
		error = DDI_SUCCESS;
		break;
	default:
		*result = NULL;
		error = DDI_FAILURE;
	}

	return (error);
}


/*
 * When our driver is unloaded, pio_detach cleans up and frees the
 * resources we allocated in pio_attach.
 */
static int
pio_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	Pio		*pio_p;
	int		instance;

	switch (cmd) {
	case DDI_DETACH:
		break;

	default:
		return (DDI_FAILURE);
	}

	instance = ddi_get_instance(dip);
	pio_p = (Pio *) ddi_get_soft_state(state_head, instance);

	/*
	 * turn off the device (including interrupts)
	 */
	ddi_put8(pio_p->csr_handle, pio_p->csr, PIO_RESET);

	/*
	 * Remove the minor node created in attach
	 */
	ddi_remove_minor_node(dip, NULL);

	ddi_regs_map_free(&pio_p->csr_handle);
	ddi_regs_map_free(&pio_p->data_handle);

	ddi_remove_intr(pio_p->dip, 0, pio_p->iblock_cookie);
	mutex_destroy(&pio_p->mutex);
	ddi_soft_state_free(state_head, instance);

	return (DDI_SUCCESS);
}

/*
 * pio_open is called in response to the open(2) system call
 */
/*ARGSUSED*/
static	int
pio_open(dev_t *dev, int openflags, int otyp, cred_t *credp)
{
	int	retval = 0;
	Pio	*pio_p;

	pio_p = (Pio *) ddi_get_soft_state(state_head, getminor(*dev));

	/*
	 * Verify instance structure
	 */
	if (pio_p == NULL)
		return (ENXIO);

	/*
	 * Verify we are being opened as a character device
	 */
	if (otyp != OTYP_CHR)
		return (EINVAL);

	/*
	 * lock control status register and exclusive open flag
	 */
	mutex_enter(&pio_p->mutex);

	/*
	 * Only allow a single open
	 */
	if (pio_p->pio_state == PIO_OPEN) {
		retval = EBUSY;
	} else {
		/*
		 * mark device as open
		 */
		pio_p->pio_state |= PIO_OPEN;

		/*
		 * start device except for interrupts
		 */
		ddi_put8(pio_p->csr_handle, pio_p->csr, PIO_ENABLE);
	}

	mutex_exit(&pio_p->mutex);

	return (retval);
}

/*
 * pio_close is called after the last process that has the device open
 * calls close(2)
 *
 * since we enforce exclusive open, pio_close will always be called when
 * the user process calls close(2)
 */
/*ARGSUSED*/
static	int
pio_close(dev_t dev, int openflags, int otyp, cred_t *credp)
{
	Pio	*pio_p;

	pio_p = (Pio *) ddi_get_soft_state(state_head, getminor(dev));

	/*
	 * lock control status register and exclusive open flag
	 */
	mutex_enter(&pio_p->mutex);

	/*
	 * let the next person open the device
	 */
	pio_p->pio_state &= ~PIO_OPEN;

	/*
	 * shut down the device
	 */
	ddi_put8(pio_p->csr_handle, pio_p->csr, PIO_RESET);

	mutex_exit(&pio_p->mutex);

	return (0);
}

/*ARGSUSED*/
static	int
pio_write(dev_t dev, struct uio *uiop, cred_t *credp)
{
	int	retval;
	int	error = OK;
	Pio	*pio_p;

	pio_p = (Pio *) ddi_get_soft_state(state_head, getminor(dev));

	/*
	 * lock control status register and data register
	 */
	mutex_enter(&pio_p->mutex);

	/*
	 * enable interrupts
	 */
	ddi_put8(pio_p->csr_handle, pio_p->csr,
	    (ddi_get8(pio_p->csr_handle, pio_p->csr) | PIO_INTR_ENABLE));

	while (uiop->uio_resid > 0) {
		/*
		 * this device issues an IDLE interrupt when it is ready
		 * to accept a character; the interrupt can be cleared
		 * by setting PIO_INTR_CLEAR.  The interrupt is reasserted
		 * after the next character is written or the next time
		 * PIO_INTR_ENABLE is toggled on.
		 *
		 * we wait for an interrupt between each character; in
		 * real life very few devices would behave this way, but
		 * this is a simple way to illustrate interrupt handling
		 */

		/*
		 * wait for interrupt (see pio_intr)
		 *
		 * Note that we currently hold the MUTEX.  cv_wait will
		 * temporarily drop the MUTEX so that pio_intr can
		 * acquire it; pio_intr drops the MUTEX after calling
		 * cv_signal, and cv_wait re-acquires it for us before
		 * it returns.
		 *
		 * Note to 4.x developers: this use of cv_wait/cv_signal
		 * within a MUTEX is similar to the use of sleep/wakeup
		 * within an spl.  There are, however, important differences.
		 * MUTEXes lock data structures; they do not block processor
		 * interrupts.  Also, the interrupt routine can block waiting
		 * for a MUTEX.
		 */
		cv_wait(&pio_p->cv, &pio_p->mutex);

		/*
		 * get a character from the user's write request
		 */
		retval = uwritec(uiop);
		if (retval == -1) {
			/*
			 * error: bail out
			 */
			error = retval;
			break;
		} else {
			/*
			 * send the character to the device
			 */
			ddi_put8(pio_p->data_handle, pio_p->data,
			    (u_char)retval);
		}
	}

	/*
	 * disable interrupts
	 */
	ddi_put8(pio_p->csr_handle, pio_p->csr,
	    (ddi_get8(pio_p->csr_handle, pio_p->csr) & ~PIO_INTR_ENABLE));

	mutex_exit(&pio_p->mutex);

	return (error);
}

/*
 * This is a sample ioctl routine to show the use of ddi_copyin(9F)
 * and ddi_copyout(9F).
 *
 * Note to 4.x developers: in previous releases, up to 128 bytes of data
 * were copied for you into the area pointed to by "arg".  This is no
 * longer true.  You must always use ddi_copyin(9F) or ddi_copyout(9F);
 * you can no longer simply dereference "arg".
 */

/*ARGSUSED*/
static	int
pio_ioctl(dev_t dev, int cmd, intptr_t arg, int flags,
	cred_t *credp, int *rvalp)
{
	Pio		*pio_p;
	int		retval = 0;
	uint8_t		tmp;

	pio_p = (Pio *) ddi_get_soft_state(state_head, getminor(dev));

	switch (cmd) {
	case PIO_GET_CSR:
		mutex_enter(&pio_p->mutex);
		tmp = ddi_get8(pio_p->csr_handle, pio_p->csr);
		if (ddi_copyout(&tmp, (void *)arg, sizeof (Pio_csr),
		    flags) != 0) {
			retval = EFAULT;
		}
		mutex_exit(&pio_p->mutex);
		break;

	case PIO_SET_CSR:
		mutex_enter(&pio_p->mutex);
		if (ddi_copyin((void *)arg, &tmp, sizeof (Pio_csr),
		    flags) != 0) {
			retval = EFAULT;
		} else {
			ddi_put8(pio_p->csr_handle, pio_p->csr, tmp);
		}
		mutex_exit(&pio_p->mutex);
		break;

	default:
		retval = ENOTTY;
		break;
	}

	return (retval);
}

/*
 * Whenever the system sees an interrupt at the level we specified with
 * add_intr in pio_attach, we get a chance to determine if it is our
 * interrupt or somebody else's.
 *
 * This routine will get called for each instance of each driver that uses
 * that level interrupt until somebody returns DDI_INTR_CLAIMED.
 *
 * Note to 4.x developers: in 4.x your xxpoll() routine had to check each
 * of your instances (units) to determine if any were interrupting.  This
 * is no longer necessary as the system now calls your xxintr() routine
 * for each attached instance of your device.
 */
static u_int
pio_intr(caddr_t instance)
{
	Pio		*pio_p;
	u_int  		int_serviced = DDI_INTR_UNCLAIMED;

	pio_p = (Pio *) ddi_get_soft_state(state_head, (u_int)instance);

	/*
	 * lock the control status register
	 */
	mutex_enter(&pio_p->mutex);

	if (ddi_get8(pio_p->csr_handle, pio_p->csr) & PIO_IDLE_INTR) {
		/*
		 * we are interrupting
		 */
		int_serviced = DDI_INTR_CLAIMED;

		/*
		 * clear the interrupt
		 */
		ddi_put8(pio_p->csr_handle, pio_p->csr,
		    (ddi_get8(pio_p->csr_handle,
			pio_p->csr) | PIO_INTR_CLEAR));

		/*
		 * "wake up" the top half
		*/
		cv_signal(&pio_p->cv);
	}

	mutex_exit(&pio_p->mutex);

	return (int_serviced);
}
