/*
 * Copyright (c) 1993,1997, Sun Microsystems, Inc.
 * All rights reserved.
 */

#ident	"@(#)pvip.c	1.19	97/04/07 SMI"

/*
 * P9000 theory of operation:
 *
 * Most P9000 operations are done by mapping the P9000 components into
 * user process memory.  User processes that share mappings
 * must cooperate among themselves to prevent damaging
 * the state of the P9000.  User processes may also acquire private
 * mappings (MAP_PRIVATE flag to mmap(2)), in which case the P9000
 * driver will preserve device state for each mapping.
 *
 * Mappings to the registers may be MAP_PRIVATE, in which case
 * the driver keeps a per-context copy of the drawing and control registers in
 * local memory.  Only one context at a time may have valid mappings.
 * If a process tries to access the registers through an invalid
 * mapping, the driver in invoked to swap register state and validate
 * and validate the mappings.
 *
 * Finally, processes have the option of mapping the "vertical retrace
 * page".  This is a page in shared memory containing a 32-bit integer
 * that is incremented each time a vertical retrace interrupt occurs.  It
 * is used so that programs may synchronize themselves with vertical
 * retrace.
 */

#define	ALWAYS_NO_INTERRUPTS 0

/*
 * Weitek P9000 color frame buffer driver
 */

#include <sys/debug.h>
#include <sys/types.h>
#include <sys/buf.h>
#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/map.h>
#include <sys/cred.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/kmem.h>
#include <sys/cmn_err.h>
#include <sys/kmem.h>
#include <sys/conf.h>
#include <sys/ddi.h>
#include <sys/devops.h>
#include <sys/sunddi.h>
#include <sys/modctl.h>

#include <sys/visual_io.h>
#ifdef _DDICT
#define	NBBY 8
#endif

#include <sys/fbio.h>
#include "pvipreg.h"
#include "bt485reg.h"

#include "viperreg.h"
#include "viperio.h"

/* configuration options */

#ifndef lint
char _depends_on[] = "misc/seg_mapdev";
#endif /* lint */

#define	PVIPDEBUG	0
#if PVIPDEBUG >= 1
#define	ASSERT2(e) \
{\
	if (!(e)) \
		cmn_err(CE_NOTE, \
		"pvip: assertion failed \"%s\", line %d\n", #e, __LINE__); }
#else
#define	ASSERT2(e)		/* nothing */
#endif

#if PVIPDEBUG >= 2
int	pvip_debug = 1;

#define	DEBUGF(level, args) \
	{ if (pvip_debug == (level)) cmn_err args; }
#define	DUMP_SEGS(level, s, c) \
	{ if (pvip_debug == (level)) dump_segs(s, c); }
#else
#define	DEBUGF(level, args)	/* nothing */
#define	DUMP_SEGS(level, s, c)	/* nothing */
#endif

/* Data access requirements. */
#ifdef _LITTLE_ENDIAN
static struct ddi_device_acc_attr endian_attr = {
	DDI_DEVICE_ATTR_V0,
	DDI_STRUCTURE_LE_ACC,
	DDI_STRICTORDER_ACC
};
#else /* _LITTLE_ENDIAN */
static struct ddi_device_acc_attr endian_attr = {
	DDI_DEVICE_ATTR_V0,
	DDI_STRUCTURE_BE_ACC,
	DDI_STRICTORDER_ACC
};
#endif /* _LITTLE_ENDIAN */

static struct ddi_device_acc_attr nosw_attr = {
	DDI_DEVICE_ATTR_V0,
	DDI_NEVERSWAP_ACC,
	DDI_STRICTORDER_ACC
};

#define	CDELAY(c, n)	\
{ \
	register int N = n; \
	while (--N > 0) { \
		if (c) \
			break; \
		drv_usecwait(1); \
	} \
}

#define	getprop(devi, name, def)	\
		ddi_getprop(DDI_DEV_T_ANY, (devi), \
		DDI_PROP_DONTPASS, (name), (def))

/* config info */

static int	pvip_open(dev_t *, int, int, cred_t *);
static int	pvip_close(dev_t, int, int, cred_t *);
static int	pvip_ioctl(dev_t, int, int, int, cred_t *, int *);
static int	pvip_mmap(dev_t, off_t, int);
static int	pvip_segmap(dev_t, off_t, struct as *, caddr_t *, off_t,
				u_int, u_int, u_int, cred_t *);

static struct cb_ops pvip_cb_ops = {
	pvip_open,		/* open */
	pvip_close,		/* close */
	nodev,			/* strategy */
	nodev,			/* print */
	nodev,			/* dump */
	nodev,			/* read */
	nodev,			/* write */
	pvip_ioctl,		/* ioctl */
	nodev,			/* devmap */
	pvip_mmap,		/* mmap */
	pvip_segmap,		/* segmap */
	nochpoll,		/* poll */
	ddi_prop_op,		/* cb_prop_op */
	0,			/* streamtab */
	D_NEW | D_MP		/* Driver compatibility flag */
};

static int	pvip_info(dev_info_t *, ddi_info_cmd_t, void *, void **);
static int	pvip_attach(dev_info_t *, ddi_attach_cmd_t);
static int	pvip_detach(dev_info_t *, ddi_detach_cmd_t);
static int	pvip_probe(dev_info_t *);
static int	pvip_power(dev_info_t *, int, int);

static struct dev_ops pvip_ops = {
	DEVO_REV,		/* devo_rev, */
	0,			/* refcnt */
	pvip_info,		/* info */
	nulldev,		/* identify */
	pvip_probe,		/* probe */
	pvip_attach,		/* attach */
	pvip_detach,		/* detach */
	nodev,			/* reset */
	&pvip_cb_ops,		/* driver operations */
	(struct bus_ops *) 0,	/* bus operations */
	pvip_power,		/* power operations */
};

static struct pvip_map_pvt {
	struct pvip_softc *softc;
	ddi_mapdev_handle_t handle;	/* handle of mapdev object	*/
	u_int		offset;		/* starting offset of this map	*/
	u_int		len;		/* length of this map		*/
	struct pvip_cntxt *context;	/* associated context		*/
	struct pvip_map_pvt *next;	/* List of associated pvt's	*/
};

#define	PVIP_CMAP_ENTRIES	256
#define	PVIP_CURSOR_ENTRIES	2

/*
 * Per-context info:
 *	many registers in the tec and fbc do
 *	not need to be saved/restored.
 */
struct pvip_cntxt {
	struct pvip_cntxt *link;
	struct pvip_map_pvt *pvip_ctx_pvt;	/* List of associated pvt's */
						/* for this context */
	int32_t		pid;			/* owner of mapping */
	int32_t		pvip_ctx_flag;
	uint32_t	pvip_ctx_interrupt;		/* 0x000z0008 rw */
	uint32_t	pvip_ctx_interrupt_en;		/* 0x000z000c rw */
	uint32_t	pvip_ctx_cindex;		/* 0x000s018c rw */
	int32_t		pvip_ctx_w_off_xy;		/* 0x000s0190 rw */
	uint32_t	pvip_ctx_fground;		/* 0x000s0200 rw */
	uint32_t	pvip_ctx_bground;		/* 0x000s0204 rw */
	uint32_t	pvip_ctx_pmask;			/* 0x000s0208 rw */
	uint32_t	pvip_ctx_draw_mode;		/* 0x000s020c rw */
	int32_t		pvip_ctx_pat_originx;		/* 0x000s0210 rw */
	int32_t		pvip_ctx_pat_originy;		/* 0x000s0214 rw */
	uint32_t	pvip_ctx_raster;		/* 0x000s0218 rw */
	uint32_t	pvip_ctx_pixel8_reg;		/* 0x000s021c rw */
	int32_t		pvip_ctx_w_min;			/* 0x000s0220 rw */
	int32_t		pvip_ctx_w_max;			/* 0x000s0224 rw */
	uint32_t	pvip_ctx_pattern[PVIP_NUM_PAT];	/* 0x000s0280 rw */
	int32_t		pvip_ctx_x0;			/* 0x000s1008 rw */
	int32_t		pvip_ctx_y0;			/* 0x000s1010 rw */
	int32_t		pvip_ctx_x1;			/* 0x000s1048 rw */
	int32_t		pvip_ctx_y1;			/* 0x000s1050 rw */
	int32_t		pvip_ctx_x2;			/* 0x000s1088 rw */
	int32_t		pvip_ctx_y2;			/* 0x000s1090 rw */
	int32_t		pvip_ctx_x3;			/* 0x000s10c8 rw */
	int32_t		pvip_ctx_y3;			/* 0x000s10d0 rw */
};

/* per-unit data */
struct pvip_softc {
	int		w;
	int		h;
	int		size;		/* total size of frame buffer */
	kmutex_t	interlock;	/* interrupt locking */
	pvip_t	pvip;		/* pvip hardware mapping structure */
	ddi_acc_handle_t	pvip_dr_handle[8];	/* handle to regs */
	ddi_acc_handle_t	pvip_cr_handle[8];	/* handle to regs */
	ddi_acc_handle_t	pvip_fb_handle;		/* handle to regs */
	ddi_acc_handle_t	pvip_bt_handle;		/* handle to regs */
	ddi_acc_handle_t	pvip_vga_handle;	/* handle to regs */
	ddi_acc_handle_t	pvip_conf_handle;	/* handle to regs */
	ddi_acc_handle_t	pvip_compat_handle;
	uchar_t *pvip_compat_page;
	u_char *ndvram;		/* for non display VRAM while suspended */
	int	ndvramsz;	/* size of non display VRAM */

	struct softcur {
	int16_t		enable;		/* cursor enable */
	int16_t		pad1;
	struct fbcurpos	pos;	/* cursor position */
	struct fbcurpos	hot;	/* cursor hot spot */
	struct fbcurpos	size;	/* cursor bitmap size */
	uint32_t	image[32];	/* cursor image bitmap */
	uint32_t	mask[32];	/* cursor mask bitmap */
	} cur;

	u_char	omap[PVIP_CURSOR_ENTRIES][3];
	u_char	cmap[PVIP_CMAP_ENTRIES][3];	/* shadow color map */
	uint16_t omap_update;		/* overlay colormap update flag */

	uint16_t	cmap_index;		/* colormap update index */
	uint16_t	cmap_count;		/* colormap update count */

#define	PVIP_VRTIOCTL	1		/* FBIOVERTICAL in effect */
#define	PVIP_VRTCTR	2		/* OWGX vertical retrace counter */

	int		*vrtpage;		/* pointer to VRT page */
	int		*vrtalloc;		/* pointer to VRT allocation */
	int		vrtmaps;		/* number of VRT page maps */
	int		vrtflag;		/* vrt interrupt flag */
	struct pvip_cntxt *curctx;	/* context switching */
	struct pvip_cntxt shared_ctx;	/* shared context */
	struct pvip_cntxt *pvt_ctx;	/* list of non-shared contexts */
	u_int		chiprev;		/* fbc chip revision # */
	u_int		pu_config;		/* PO config */
	int		emulation;		/* emulation type */
	dev_info_t	*devi;			/* back pointer */
	int		have_intr;		/* do interrupts work */
	ddi_iblock_cookie_t	iblock_cookie;	/* block interrupts */
	kmutex_t	mutex;			/* mutex locking */
	kcondvar_t	vrtsleep;		/* for sleeping */
	int		suspended;		/* true if suspended */
	viper_port_t	viperport;
	viper_init_t	viperinit;
	int		viper_id;
	int		bus_type;
	struct	vis_identifier pvip_ident;
};

static int	pvip_map_access(ddi_mapdev_handle_t, void *, off_t);
static void	pvip_map_free(ddi_mapdev_handle_t, void *);
static int	pvip_map_dup(ddi_mapdev_handle_t, void *,
				ddi_mapdev_handle_t, void **);

static struct ddi_mapdev_ctl pvip_map_ops = {
	MAPDEV_REV,		/* mapdev_ops version number	*/
	pvip_map_access,	/* mapdev access routine	*/
	pvip_map_free,		/* mapdev free routine		*/
	pvip_map_dup		/* mapdev dup routine		*/
};

static void	*pvip_softc_head;
static u_int	pvip_pagesize;

#if ALWAYS_NO_INTERRUPTS
static u_int	pvip_pageoffset;
#endif

/* default structure for FBIOGATTR ioctl */
static struct fbgattr	pvip_attr;

/*
 * handy macros
 */

#define	getsoftc(unit) \
	((struct pvip_softc *)ddi_get_soft_state(pvip_softc_head, (unit)))

#define	btob(n)			ctob(btoc(n))

#define	PVIP_WAIT		500000	/* .5 seconds */

#define	pvip_int_enable(soft) \
	ddi_putl((soft)->pvip_cr_handle[0], &(soft)->pvip.pvip_interrupt_en, \
	PVIP_INTEN_MEN_CTRL | PVIP_INTEN_MEN | \
	PVIP_INTEN_VBLANKED_EN_CTRL | PVIP_INTEN_VBLANKED_EN);

#define	pvip_int_disable_intr(soft) \
	ddi_putl((soft)->pvip_cr_handle[0], &(soft)->pvip.pvip_interrupt_en, \
	PVIP_INTEN_VBLANKED_EN_CTRL);

#define	pvip_int_clear(soft) \
	ddi_putl((soft)->pvip_cr_handle[0], &(soft)->pvip.pvip_interrupt, \
	PVIP_INT_VBLANKED_CTRL);

/* check if color map update is pending */
#define	pvip_update_pending(softc) \
	((softc)->cmap_count || (softc)->omap_update)

/*
 * forward references
 */
static u_int	pvip_intr(caddr_t);
static void	pvip_reset_cmap(u_char *, register u_int);
static void	pvip_update_cmap(struct pvip_softc *, u_int, u_int);
static void	pvip_update_omap(struct pvip_softc *);
static void	pvip_cmap_copyout(u_char *, u_char *, u_int);
static void	pvip_cmap_copyin(u_char *, u_char *, u_int);
static void	pvip_set_video(struct pvip_softc *, int);
static int	pvip_cntxsave(struct pvip_softc *, struct pvip_cntxt *);
static int	pvip_cntxrestore(struct pvip_softc *, struct pvip_cntxt *);
static struct	pvip_cntxt *pvip_ctx_map_insert(struct pvip_softc *, int);
static int	pvip_getpid(void);

static int	viper_ioctl(dev_t, int, int, int);
static void	viper_setcurpos(struct pvip_softc *);
static void	viper_setcurshape(struct pvip_softc *);
static void	viper_setcurcolor(struct pvip_softc *);
static void	viper_setpalet(struct pvip_softc *);
static int	viper_enter_graph_mode(struct pvip_softc *);
static int	viper_enter_text_mode(struct pvip_softc *);
static int	viper_unlock_regs(struct pvip_softc *);
static int	viper_lock_regs(struct pvip_softc *);
static void	viper_write_ic2061a(struct pvip_softc *, uint32_t);

/* Loadable Driver stuff */

extern struct mod_ops	mod_driverops;

static struct modldrv	pvip_modldrv = {
	&mod_driverops,		/* Type of module.This one is a driver */
	"weitek p9000 driver v95/01/21",	/* Name of the module. */
	&pvip_ops,			/* driver ops */
};

static struct modlinkage	pvip_modlinkage = {
	MODREV_1,
	(void *) &pvip_modldrv,
	NULL,
};

static offset_t p9000_control_bases[PVIP_NUM_CONTROL_REG] = {
	P9000_CONTROL_BASE0, P9000_CONTROL_BASE1, P9000_CONTROL_BASE2,
	P9000_CONTROL_BASE3, P9000_CONTROL_BASE4, P9000_CONTROL_BASE5,
	P9000_CONTROL_BASE6, P9000_CONTROL_BASE7
};

static offset_t p9000_drawing_bases[PVIP_NUM_DRAWING_REG] = {
	P9000_DRAWING_BASE0, P9000_DRAWING_BASE1, P9000_DRAWING_BASE2,
	P9000_DRAWING_BASE3, P9000_DRAWING_BASE4, P9000_DRAWING_BASE5,
	P9000_DRAWING_BASE6, P9000_DRAWING_BASE7
};

/* DDX compatibility arrays */
static offset_t pvip_control_bases[PVIP_NUM_CONTROL_REG] = {
	PVIP_C_CONTROL_BASE0, PVIP_C_CONTROL_BASE1, PVIP_C_CONTROL_BASE2,
	PVIP_C_CONTROL_BASE3, PVIP_C_CONTROL_BASE4, PVIP_C_CONTROL_BASE5,
	PVIP_C_CONTROL_BASE6, PVIP_C_CONTROL_BASE7
};

static offset_t pvip_drawing_bases[PVIP_NUM_DRAWING_REG] = {
	PVIP_C_DRAWING_BASE0, PVIP_C_DRAWING_BASE1, PVIP_C_DRAWING_BASE2,
	PVIP_C_DRAWING_BASE3, PVIP_C_DRAWING_BASE4, PVIP_C_DRAWING_BASE5,
	PVIP_C_DRAWING_BASE6, PVIP_C_DRAWING_BASE7
};

static unsigned char pvip_bitflip[256] = {
	0x00, 0x80, 0x40, 0xc0, 0x20, 0xa0, 0x60, 0xe0,
	0x10, 0x90, 0x50, 0xd0, 0x30, 0xb0, 0x70, 0xf0,
	0x08, 0x88, 0x48, 0xc8, 0x28, 0xa8, 0x68, 0xe8,
	0x18, 0x98, 0x58, 0xd8, 0x38, 0xb8, 0x78, 0xf8,
	0x04, 0x84, 0x44, 0xc4, 0x24, 0xa4, 0x64, 0xe4,
	0x14, 0x94, 0x54, 0xd4, 0x34, 0xb4, 0x74, 0xf4,
	0x0c, 0x8c, 0x4c, 0xcc, 0x2c, 0xac, 0x6c, 0xec,
	0x1c, 0x9c, 0x5c, 0xdc, 0x3c, 0xbc, 0x7c, 0xfc,
	0x02, 0x82, 0x42, 0xc2, 0x22, 0xa2, 0x62, 0xe2,
	0x12, 0x92, 0x52, 0xd2, 0x32, 0xb2, 0x72, 0xf2,
	0x0a, 0x8a, 0x4a, 0xca, 0x2a, 0xaa, 0x6a, 0xea,
	0x1a, 0x9a, 0x5a, 0xda, 0x3a, 0xba, 0x7a, 0xfa,
	0x06, 0x86, 0x46, 0xc6, 0x26, 0xa6, 0x66, 0xe6,
	0x16, 0x96, 0x56, 0xd6, 0x36, 0xb6, 0x76, 0xf6,
	0x0e, 0x8e, 0x4e, 0xce, 0x2e, 0xae, 0x6e, 0xee,
	0x1e, 0x9e, 0x5e, 0xde, 0x3e, 0xbe, 0x7e, 0xfe,
	0x01, 0x81, 0x41, 0xc1, 0x21, 0xa1, 0x61, 0xe1,
	0x11, 0x91, 0x51, 0xd1, 0x31, 0xb1, 0x71, 0xf1,
	0x09, 0x89, 0x49, 0xc9, 0x29, 0xa9, 0x69, 0xe9,
	0x19, 0x99, 0x59, 0xd9, 0x39, 0xb9, 0x79, 0xf9,
	0x05, 0x85, 0x45, 0xc5, 0x25, 0xa5, 0x65, 0xe5,
	0x15, 0x95, 0x55, 0xd5, 0x35, 0xb5, 0x75, 0xf5,
	0x0d, 0x8d, 0x4d, 0xcd, 0x2d, 0xad, 0x6d, 0xed,
	0x1d, 0x9d, 0x5d, 0xdd, 0x3d, 0xbd, 0x7d, 0xfd,
	0x03, 0x83, 0x43, 0xc3, 0x23, 0xa3, 0x63, 0xe3,
	0x13, 0x93, 0x53, 0xd3, 0x33, 0xb3, 0x73, 0xf3,
	0x0b, 0x8b, 0x4b, 0xcb, 0x2b, 0xab, 0x6b, 0xeb,
	0x1b, 0x9b, 0x5b, 0xdb, 0x3b, 0xbb, 0x7b, 0xfb,
	0x07, 0x87, 0x47, 0xc7, 0x27, 0xa7, 0x67, 0xe7,
	0x17, 0x97, 0x57, 0xd7, 0x37, 0xb7, 0x77, 0xf7,
	0x0f, 0x8f, 0x4f, 0xcf, 0x2f, 0xaf, 0x6f, 0xef,
	0x1f, 0x9f, 0x5f, 0xdf, 0x3f, 0xbf, 0x7f, 0xff,
};


int
_init(void)
{
	register int	e;

	DEBUGF(1, (CE_CONT, "pvip: compiled %s, %s\n", __TIME__, __DATE__));

	if ((e = ddi_soft_state_init(&pvip_softc_head,
		sizeof (struct pvip_softc), 1)) != 0) {
		DEBUGF(1, (CE_CONT, "done\n"));
		return (e);
	}

	e = mod_install(&pvip_modlinkage);

	if (e) {
		ddi_soft_state_fini(&pvip_softc_head);
		DEBUGF(1, (CE_CONT, "done\n"));
	}

	DEBUGF(1, (CE_CONT, "pvip: _init done rtn=%d\n", e));

	return (e);
}

int
_fini(void)
{
	register int	e;

	if ((e = mod_remove(&pvip_modlinkage)) != 0)
		return (e);

	ddi_soft_state_fini(&pvip_softc_head);

	return (0);
}

int
_info(struct modinfo *modinfop)
{
	return (mod_info(&pvip_modlinkage, modinfop));
}

static int
pvip_probe(register dev_info_t *devi)
{
	DEBUGF(1, (CE_CONT, "pvip_probe (%s) unit=%d\n",
	ddi_get_name(devi), ddi_get_instance(devi)));

	if (ddi_dev_is_sid(devi) == DDI_SUCCESS) {
		DEBUGF(2, (CE_CONT, "pvip_probe dev is sid\n"));
		return (DDI_PROBE_DONTCARE);
	}

	return (DDI_FAILURE);
}


static int
pvip_attach(register dev_info_t *devi,
	register ddi_attach_cmd_t cmd)
{
	register struct pvip_softc	*softc;
	register int unit = ddi_get_instance(devi);
	pvipp_t	volatile pvipreg;
	char		name[16];
	uint32_t	timestamp[2];
	int		power[2];
	int		proplen, i;

	DEBUGF(1, (CE_CONT, "pvip_attach unit=%d cmd=%d\n", unit, (int) cmd));

	switch (cmd) {
	case DDI_ATTACH:
		break;

	case DDI_RESUME:
		if (!(softc = (struct pvip_softc *)
				ddi_get_driver_private(devi)))
			return (DDI_FAILURE);
		if (!softc->suspended)
			return (DDI_SUCCESS);
		mutex_enter(&softc->mutex);

		if (softc->curctx) {
			u_int displaysize = softc->w * softc->h;

			pvipreg = &softc->pvip;

			/* Restore non display RAM */
			ddi_rep_putb(softc->pvip_fb_handle, softc->ndvram,
				((u_char *)
				&pvipreg->pvip_frame_bufferp[displaysize]),
				softc->ndvramsz, DDI_DEV_AUTOINCR);

			kmem_free(softc->ndvram, softc->ndvramsz);

			/* Restore other frame buffer state */
			(void) pvip_cntxrestore(softc, softc->curctx);

			viper_setcurpos(softc);
			viper_setcurshape(softc);
			pvip_update_cmap(softc, (u_int)0, PVIP_CMAP_ENTRIES);
			pvip_update_omap(softc);

			pvip_int_enable(softc);	/* Schedule the update */
		}

		softc->suspended = 0;
		mutex_exit(&softc->mutex);
		/* Restore brightness level */
		if (ddi_getlongprop_buf(DDI_DEV_T_ANY, devi, 0,
				"pm_norm_pwr", (caddr_t)power, &proplen)
				== DDI_PROP_SUCCESS)

			if (pvip_power(devi, 1, power[1]) != DDI_SUCCESS)
				return (DDI_FAILURE);

		return (DDI_SUCCESS);

	default:
		return (DDI_FAILURE);
	}

	DEBUGF(1, (CE_CONT, "pvip_attach unit=%d\n", unit));

	pvip_pagesize = ddi_ptob(devi, 1);

#if ALWAYS_NO_INTERRUPTS
	pvip_pageoffset = pvip_pagesize - 1;
#endif

	/* Allocate softc struct */
	if (ddi_soft_state_zalloc(pvip_softc_head, unit))
		return (DDI_FAILURE);

	softc = getsoftc(unit);

	/* link it in */
	softc->devi = devi;

	DEBUGF(1, (CE_CONT, "pvip_attach devi=0x%x unit=%d\n", devi, unit));

	ddi_set_driver_private(devi, (caddr_t) softc);

	pvipreg = &softc->pvip;

	(void) strcpy((caddr_t)softc->pvip_ident.name, "SUNWp9000");

	/* P9000 on the PCI */

	/* map in the configuration registers */
	if (pci_config_setup(devi, &softc->pvip_conf_handle)) {
		DEBUGF(2, (CE_CONT,
			"pvip_attach: ddi_conf_regs (pvip) FAILED\n"));
		return (DDI_FAILURE);
	}

	/* Map in the device registers. */
	for (i = 0; i < PVIP_NUM_CONTROL_REG; i++) {
		if (ddi_regs_map_setup(devi, 1,
				(caddr_t *)&pvipreg->pvip_control_regp[i],
				p9000_control_bases[i], P9000_CONTROL_SIZE,
				&endian_attr, &softc->pvip_cr_handle[i])) {
			DEBUGF(2, (CE_CONT,
				"pvip_attach: ddi_cr_regs (p9000) FAILED\n"));
			return (DDI_FAILURE);
		}
	}

	for (i = 0; i < PVIP_NUM_DRAWING_REG; i++) {
		if (ddi_regs_map_setup(devi, 1,
				(caddr_t *)&pvipreg->pvip_drawing_regp[i],
				p9000_drawing_bases[i], P9000_DRAWING_SIZE,
				&endian_attr, &softc->pvip_dr_handle[i])) {
			DEBUGF(2, (CE_CONT,
				"pvip_attach: ddi_dr_regs (p9000) FAILED\n"));
			return (DDI_FAILURE);
		}
	}

	/* map in the framebuffer */
	if (ddi_regs_map_setup(devi, 1,
			(caddr_t *)&pvipreg->pvip_frame_bufferp,
			P9000_FRAME_BUFFER_BASE, P9000_FRAME_BUFFER_SIZE,
			&nosw_attr, &softc->pvip_fb_handle)) {
		DEBUGF(2, (CE_CONT,
			"pvip_attach: ddi_map_regs (pvip) FAILED\n"));
		return (DDI_FAILURE);
	}

	/* DDX compatibility mappings */
	softc->pvip_compat_handle = softc->pvip_fb_handle;
	softc->pvip_compat_page = pvipreg->pvip_frame_bufferp;

	/* map in the extended registers */
	if (ddi_regs_map_setup(devi, 2,
			(caddr_t *)&pvipreg->pvip_bt485_basep,
			0, 0, &endian_attr, &softc->pvip_bt_handle)) {
		DEBUGF(2, (CE_CONT,
			"pvip_attach: ddi_map_regs (pvip) FAILED\n"));
		return (DDI_FAILURE);
	}

	/* map in the "hard-decode" addresses for VGA */
	if (ddi_regs_map_setup(devi, 5,
			(caddr_t *)&pvipreg->pvip_vga_basep, 0, 0,
			&endian_attr, &softc->pvip_vga_handle)) {
		DEBUGF(2, (CE_CONT,
			"pvip_attach: ddi_map_regs (pvip) FAILED\n"));
		return (DDI_FAILURE);
	}

	/* export addresses */
#define	bt485_comreg3		bt485_stat_reg
#define	bt485_ram_write		pvip_ramdac0
	softc->viperport.vp_bt485_ram_write =
		(u_char *)&pvipreg->pvip_ramdac0;

#define	bt485_palet_data	pvip_ramdac1
	softc->viperport.vp_bt485_palet_data =
		(u_char *)&pvipreg->pvip_ramdac1;

#define	bt485_pixel_mask	pvip_ramdac2
	softc->viperport.vp_bt485_pixel_mask =
		(u_char *)&pvipreg->pvip_ramdac2;

#define	bt485_ram_read	pvip_ramdac3
	softc->viperport.vp_bt485_ram_read =
		(u_char *)&pvipreg->pvip_ramdac3;

#define	bt485_color_write	pvip_ramdac4
	softc->viperport.vp_bt485_color_write =
		(u_char *)&pvipreg->pvip_ramdac4;

#define	bt485_color_data	pvip_ramdac5
	softc->viperport.vp_bt485_color_data =
		(u_char *)&pvipreg->pvip_ramdac5;

#define	bt485_comreg0	pvip_ramdac6
	softc->viperport.vp_bt485_comreg0 =
		(u_char *)&pvipreg->pvip_ramdac6;

#define	bt485_color_read	pvip_ramdac7
	softc->viperport.vp_bt485_color_read =
		(u_char *)&pvipreg->pvip_ramdac7;

#define	bt485_comreg1	pvip_ramdac8
	softc->viperport.vp_bt485_comreg1 =
		(u_char *)&pvipreg->pvip_ramdac8;

#define	bt485_comreg2	pvip_ramdac9
	softc->viperport.vp_bt485_comreg2 =
		(u_char *)&pvipreg->pvip_ramdac9;

#define	bt485_stat_reg	pvip_ramdac10
	softc->viperport.vp_bt485_stat_reg =
		(u_char *)&pvipreg->pvip_ramdac10;

#define	bt485_cursor_data	pvip_ramdac11
	softc->viperport.vp_bt485_cursor_data =
		(u_char *)&pvipreg->pvip_ramdac11;

#define	bt485_cursor_x_low	pvip_ramdac12
	softc->viperport.vp_bt485_cursor_x_low =
		(u_char *)&pvipreg->pvip_ramdac12;

#define	bt485_cursor_x_high	pvip_ramdac13
	softc->viperport.vp_bt485_cursor_x_high =
		(u_char *)&pvipreg->pvip_ramdac13;

#define	bt485_cursor_y_low	pvip_ramdac14
	softc->viperport.vp_bt485_cursor_y_low =
		(u_char *)&pvipreg->pvip_ramdac14;

#define	bt485_cursor_y_high	pvip_ramdac15
	softc->viperport.vp_bt485_cursor_y_high =
		(u_char *)&pvipreg->pvip_ramdac15;

	/* "hard-decode" registers in the VGA space */
	softc->viperport.vp_miscout =
		(u_char *)pvipreg->pvip_vga_basep + (VIPER_MISCOUT & 0xf);
	softc->viperport.vp_miscin =
		(u_char *)pvipreg->pvip_vga_basep + (VIPER_MISCIN & 0xf);
	softc->viperport.vp_seq_index_port =
		(u_char *)pvipreg->pvip_vga_basep +
		(VIPER_SEQ_INDEX_PORT & 0xf);
	softc->viperport.vp_seq_data_port =
		(u_char *)pvipreg->pvip_vga_basep +
		(VIPER_SEQ_DATA_PORT & 0xf);

	softc->vrtpage = NULL;
	softc->vrtalloc = NULL;
	softc->vrtmaps = 0;
	softc->vrtflag = 0;

#if ALWAYS_NO_INTERRUPTS
	mutex_init(&softc->interlock, "pvip_interlock", MUTEX_DRIVER,
		softc->iblock_cookie);
	mutex_init(&softc->mutex, "pvip", MUTEX_DRIVER, softc->iblock_cookie);
	cv_init(&softc->vrtsleep, "pvip", CV_DRIVER, softc->iblock_cookie);

	ddi_remove_intr(devi, 0, softc->iblock_cookie);

	if (ddi_add_intr(devi, 0, &softc->iblock_cookie, 0,
			pvip_intr, (caddr_t) softc) != DDI_SUCCESS) {
		DEBUGF(2, (CE_CONT,
			"pvip_attach%d add_intr failed\n", unit));

		(void) pvip_detach(devi, DDI_DETACH);
		return (DDI_FAILURE);
	}

	softc->have_intr = 1;

#else /* ALWAYS_NO_INTERRUPTS */

	softc->have_intr = 0;
	mutex_init(&softc->interlock, "pvip_interlock", MUTEX_DRIVER,
		softc->iblock_cookie);
	mutex_init(&softc->mutex, "pvip", MUTEX_DRIVER, softc->iblock_cookie);
	cv_init(&softc->vrtsleep, "pvip", CV_DRIVER, softc->iblock_cookie);

#endif /* ALWAYS_NO_INTERRUPTS */

	/*
	* Initialize hardware colormap and software colormap images. It might
	* make sense to read the hardware colormap here.
	*/
	pvip_reset_cmap(softc->cmap[0], PVIP_CMAP_ENTRIES);
	pvip_reset_cmap(softc->omap[0], PVIP_CURSOR_ENTRIES);
	pvip_update_cmap(softc, (u_int) 0, PVIP_CMAP_ENTRIES);
	pvip_update_omap(softc);

	DEBUGF(2, (CE_CONT,
		"pvip_attach%d just before create_minor node\n", unit));

	(void) sprintf(name, "pvip_%d", unit);

	if (ddi_create_minor_node(devi, name, S_IFCHR,
			unit, DDI_NT_DISPLAY, 0) == DDI_FAILURE) {
		ddi_remove_minor_node(devi, NULL);

		DEBUGF(2, (CE_CONT,
			"pvip_attach%d create_minor node failed\n", unit));

		return (DDI_FAILURE);
	}

	ddi_report_dev(devi);

	/* Initialize the shared context for this unit */
	softc->shared_ctx.pvip_ctx_pvt = NULL;
	softc->pvt_ctx = NULL;

	/* Create power management properties */
	timestamp[0] = 0;
	(void) drv_getparm(TIME, timestamp+1);
	(void) ddi_prop_create(DDI_DEV_T_NONE, devi, DDI_PROP_CANSLEEP,
		"pm_timestamp", (caddr_t)timestamp,
		sizeof (timestamp));
	power[0] = 1;
	power[1] = 255;
	(void) ddi_prop_create(DDI_DEV_T_NONE, devi, DDI_PROP_CANSLEEP,
		"pm_norm_pwr", (caddr_t)power, sizeof (power));

	return (DDI_SUCCESS);
}


static int
pvip_detach(register dev_info_t *devi,
	register ddi_detach_cmd_t cmd)
{
	register int unit = ddi_get_instance(devi);
	register struct pvip_softc *softc = getsoftc(unit);
	volatile pvipp_t pvipreg = &softc->pvip;
	int i;

	DEBUGF(1, (CE_CONT, "pvip_detach softc=%x, devi=0x%x\n", softc, devi));

	switch (cmd) {
	case DDI_DETACH:
		break;

	case DDI_SUSPEND:
		if (softc == NULL)
			return (DDI_FAILURE);

		if (softc->suspended)
			return (DDI_FAILURE);

		mutex_enter(&softc->mutex);

		if (softc->curctx) {
			register u_int displaysize =
				softc->w * softc->h;

			/*
			 * Save non display RAM - displayable memory
			 */
			softc->ndvramsz = (2 * 1024 * 1024) - displaysize;
			if ((softc->ndvram = kmem_alloc(softc->ndvramsz,
					KM_NOSLEEP)) == NULL) {
				mutex_exit(&softc->mutex);
				return (DDI_FAILURE);
			}

			ddi_rep_getb(softc->pvip_fb_handle, softc->ndvram,
			(u_char *) &pvipreg->pvip_frame_bufferp[displaysize],
			softc->ndvramsz, DDI_DEV_AUTOINCR);

			/* Save other frame buffer state */
			(void) pvip_cntxsave(softc, softc->curctx);
		}

		softc->suspended = 1;

		mutex_exit(&softc->mutex);

		return (DDI_SUCCESS);

	default:
		return (DDI_FAILURE);
	}

	if (softc->have_intr)
		ddi_remove_intr(devi, 0, softc->iblock_cookie);

	/*
	* Unmap the device registers.
	*/
	for (i = 0; i < PVIP_NUM_CONTROL_REG; i++)
		ddi_regs_map_free(&softc->pvip_cr_handle[i]);

	for (i = 0; i < PVIP_NUM_DRAWING_REG; i++)
		ddi_regs_map_free(&softc->pvip_dr_handle[i]);

	/* unmap the framebuffer */
	ddi_regs_map_free(&softc->pvip_fb_handle);

	/* unmap the extended registers */
	ddi_regs_map_free(&softc->pvip_bt_handle);

	/* unmap the "hard-decode" addresses for VGA */
	ddi_regs_map_free(&softc->pvip_vga_handle);

	if (softc->vrtalloc != NULL)
		kmem_free(softc->vrtalloc, pvip_pagesize * 2);

	mutex_destroy(&softc->mutex);

	cv_destroy(&softc->vrtsleep);

	ASSERT2(softc->curctx == NULL);

	ddi_soft_state_free(pvip_softc_head, unit);

	return (0);
}


static int
pvip_power(dev_info_t *dip, int cmpt, int level)
{
	int		power[2];
	register struct pvip_softc	*softc;

	if (cmpt != 1 || 0 > level || level > 255 || !(softc =
			(struct pvip_softc *) ddi_get_driver_private(dip)))
		return (DDI_FAILURE);

	if (level) {
		pvip_set_video(softc, FBVIDEO_ON);
		power[0] = 1;
		power[1] = level;
		(void) ddi_prop_modify(DDI_DEV_T_NONE, dip, DDI_PROP_CANSLEEP,
			"pm_norm_pwr", (caddr_t)power, sizeof (power));
	} else {
		pvip_set_video(softc, FBVIDEO_OFF);
	}

	(void) ddi_power(dip, cmpt, level);

	return (DDI_SUCCESS);
}


/* ARGSUSED */
static int
pvip_info(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg, void **result)
{
	register int	instance;
	register int	error = DDI_SUCCESS;
	register struct pvip_softc	*softc;

	instance = getminor((dev_t)arg);
	switch (infocmd) {
	case DDI_INFO_DEVT2DEVINFO:
		softc = getsoftc(instance);
		if (softc == NULL) {
			error = DDI_FAILURE;
		} else {
			*result = (void *)softc->devi;
			error = DDI_SUCCESS;
		}
		break;
	case DDI_INFO_DEVT2INSTANCE:
		*result = (void *)instance;
		error = DDI_SUCCESS;
		break;
	default:
		error = DDI_FAILURE;
	}

	return (error);
}

/*ARGSUSED*/
static
pvip_open(register dev_t *devp, register int flag,
	register int otyp, register cred_t *cred)
{
	register int unit = getminor(*devp);
	struct pvip_softc	*softc = getsoftc(unit);
	int	error = 0;

	DEBUGF(2, (CE_CONT, "pvip_open(%d)\n", unit));

	/*
	* is this gorp necessary?
	*/
	if (otyp != OTYP_CHR) {
		error = EINVAL;
	} else if (softc == NULL) {
		error = ENXIO;
	}
	return (error);
}


/*ARGSUSED*/
static
pvip_close(register dev_t dev, register int flag,
	register int otyp, register cred_t *cred)
{
	register int unit = getminor(dev);
	struct pvip_softc	*softc = getsoftc(unit);
	int	error = 0;

	DEBUGF(2, (CE_CONT, "pvip_close(%d, %d, %d)\n", unit, flag, otyp));

	if (otyp != OTYP_CHR) {
		error = EINVAL;
	} else if (softc == NULL) {
		error = ENXIO;
	} else {
		mutex_enter(&softc->mutex);
		softc->cur.enable = 0;
		softc->curctx = NULL;
		mutex_exit(&softc->mutex);
	}

	return (error);
}

/*ARGSUSED*/
static
pvip_mmap(register dev_t dev, register off_t off,
	register int prot)
{
	register struct pvip_softc *softc = getsoftc(getminor(dev));
	volatile pvipp_t pvipreg = &softc->pvip;
	register int	diff;
	register caddr_t	vaddr;
	register int	rval, i;

	DEBUGF(off ? 5 : 1, (CE_CONT, "pvip_mmap(%d, 0x%x)\n",
		getminor(dev), (u_int) off));

	if ((diff = off - PVIP_CONTROL_BASE) >= 0 &&
			diff < PVIP_CONTROL_SIZE) {
		for (i = 0; i < PVIP_NUM_CONTROL_REG; i++) {
			if ((diff = off - pvip_control_bases[i]) >= 0 &&
					diff < PVIP_C_CONTROL_SIZE) {
				vaddr = (caddr_t)pvipreg->pvip_control_regp[i]
					+ diff;
				break;
			}
		}

		if (i == PVIP_NUM_CONTROL_REG)
			vaddr = (caddr_t)softc->pvip_compat_page;

	} else if ((diff = off - PVIP_DRAWING_BASE) >= 0 &&
			diff < PVIP_DRAWING_SIZE) {
		for (i = 0; i < PVIP_NUM_DRAWING_REG; i++) {
			if ((diff = off - pvip_drawing_bases[i]) >= 0 &&
					diff < PVIP_C_DRAWING_SIZE) {
				vaddr = (caddr_t)pvipreg->pvip_drawing_regp[i]
					+ diff;
				break;
			}
		}

		if (i == PVIP_NUM_DRAWING_REG)
			vaddr = (caddr_t)softc->pvip_compat_page;

	} else if ((diff = off - PVIP_FRAME_BUFFER_BASE) >= 0 &&
			diff < PVIP_FRAME_BUFFER_SIZE)
		vaddr = (caddr_t)pvipreg->pvip_frame_bufferp + diff;

#if ALWAYS_NO_INTERRUPTS

	else if ((diff = off - PVIP_VRT_VADDR) >= 0 && diff < PVIP_VRT_SIZE)
		vaddr = softc->vrtpage ?
			(caddr_t) softc->vrtpage + diff : (caddr_t) - 1;

#endif /* ALWAYS_NO_INTERRUPTS */

	else
		vaddr = (caddr_t) - 1;

	if (vaddr != (caddr_t) - 1)
		rval = hat_getkpfnum(vaddr);
	else
		rval = -1;

	DEBUGF(5, (CE_CONT, "pvip_mmap returning 0x%x\n", rval));

	return (rval);
}


/*ARGSUSED*/
static
pvip_ioctl(dev_t dev, int cmd, int data, int mode, cred_t *cred, int *rval)
{
	register struct pvip_softc *softc = getsoftc(getminor(dev));
	volatile pvipp_t pvipreg = &softc->pvip;
	register int cursor_cmap;
	int	 i;

	u_char	*iobuf_cmap_red;
	u_char	*iobuf_cmap_green;
	u_char	*iobuf_cmap_blue;
	u_char	*stack_cmap;

	struct fbcmap	*cmap, cma;
	struct fbcursor	cp;

	u_int	index;
	u_int	count;
	u_char	*map;
	u_int	entries;

	DEBUGF(3, (CE_CONT, "pvip_ioctl(%d, 0x%x)\n", getminor(dev), cmd));

	/* default to updating normal colormap */
	cursor_cmap = 0;

	switch (cmd) {
	case VIS_GETIDENTIFIER:
		if (ddi_copyout((caddr_t)&softc->pvip_ident,
				(caddr_t)data,
				sizeof (struct vis_identifier),
				mode))
			return (EFAULT);
		break;

	case FBIOPUTCMAP:
	case FBIOGETCMAP: {

cmap_ioctl:

		if (ddi_copyin((caddr_t)data,
			(caddr_t)&cma, sizeof (struct fbcmap), mode))
			return (EFAULT);

		cmap = &cma;
		index = (u_int) cmap->index;
		count = (u_int) cmap->count;

		if (count == 0) {
			return (0);
		}

		if (cursor_cmap == 0) {
			map = softc->cmap[0];
			entries = PVIP_CMAP_ENTRIES;
		} else {
			map = softc->omap[0];
			entries = PVIP_CURSOR_ENTRIES;
		}

		if (index >= entries || index + count > entries) {
			return (EINVAL);
		}

		/*
		* Allocate memory for color map RGB entries.
		*/
		stack_cmap = kmem_alloc((PVIP_CMAP_ENTRIES * 3), KM_SLEEP);

		iobuf_cmap_red = stack_cmap;
		iobuf_cmap_green = stack_cmap + PVIP_CMAP_ENTRIES;
		iobuf_cmap_blue = stack_cmap + (PVIP_CMAP_ENTRIES * 2);

		if (cmd == FBIOPUTCMAP) {
			int error;

			DEBUGF(3, (CE_CONT, "FBIOPUTCMAP\n"));
			if (error = ddi_copyin((caddr_t)cmap->red,
					(caddr_t)iobuf_cmap_red,
					count, mode)) {
				kmem_free(stack_cmap, (PVIP_CMAP_ENTRIES * 3));
				return (error);
			}

			if (error = ddi_copyin((caddr_t)cmap->green,
					(caddr_t)iobuf_cmap_green,
					count, mode)) {
				kmem_free(stack_cmap, (PVIP_CMAP_ENTRIES * 3));
				return (error);
			}

			if (error = ddi_copyin((caddr_t)cmap->blue,
					(caddr_t)iobuf_cmap_blue,
					count, mode)) {
				kmem_free(stack_cmap, (PVIP_CMAP_ENTRIES * 3));
				return (error);
			}

			mutex_enter(&softc->mutex);
			map += index * 3;
			if (pvip_update_pending(softc)) {
				mutex_enter(&(softc)->interlock);
				pvip_int_disable_intr(softc);
				mutex_exit(&(softc)->interlock);
			}

			/*
			* Copy color map entries from stack to the color map
			* table in the softc area.
			*/

			pvip_cmap_copyin(map++, iobuf_cmap_red, count);
			pvip_cmap_copyin(map++, iobuf_cmap_green, count);
			pvip_cmap_copyin(map, iobuf_cmap_blue, count);

			/* cursor colormap update */
			if (entries < PVIP_CMAP_ENTRIES)
				pvip_update_omap(softc);
			else
				pvip_update_cmap(softc, index, count);

			pvip_int_enable(softc);

			mutex_exit(&softc->mutex);

		} else {
			/* FBIOGETCMAP */
			DEBUGF(3, (CE_CONT, "FBIOGETCMAP\n"));

			mutex_enter(&softc->mutex);
			map += index * 3;

		/*
		* Copy color map entries from soft area to
		* local storage and prepare for a copyout
		*/

		pvip_cmap_copyout(iobuf_cmap_red, map++, count);
		pvip_cmap_copyout(iobuf_cmap_green, map++, count);
		pvip_cmap_copyout(iobuf_cmap_blue, map, count);

		mutex_exit(&softc->mutex);

		if (ddi_copyout((caddr_t)iobuf_cmap_red,
				(caddr_t)cmap->red,
				count,
				mode)) {
			kmem_free(stack_cmap, (PVIP_CMAP_ENTRIES * 3));
			return (EFAULT);
		}
		if (ddi_copyout((caddr_t)iobuf_cmap_green,
				(caddr_t)cmap->green,
				count,
				mode)) {
			kmem_free(stack_cmap, (PVIP_CMAP_ENTRIES * 3));
			return (EFAULT);
		}
		if (ddi_copyout((caddr_t)iobuf_cmap_blue,
				(caddr_t)cmap->blue,
				count,
				mode)) {
			kmem_free(stack_cmap, (PVIP_CMAP_ENTRIES * 3));
			return (EFAULT);
		}
		}
		kmem_free(stack_cmap, (PVIP_CMAP_ENTRIES * 3));
		if (!softc->have_intr) {
		(void) pvip_intr((caddr_t) softc);
		}
	}
	break;


	case FBIOSATTR: {
		struct fbsattr	attr;

		if (ddi_copyin((caddr_t)data,
			(caddr_t)&attr,
			sizeof (attr),
			mode))
		return (EFAULT);
		DEBUGF(3, (CE_CONT, "FBIOSATTR, type=%d\n", attr.emu_type));
		if (attr.emu_type != -1)
		return (EINVAL);
		/* ignore device-dependent stuff */
	}
	break;


	case FBIOGATTR: {
		struct fbgattr	attr;

		DEBUGF(3, (CE_CONT,
			"FBIOGATTR, emu_type=%d\n", softc->emulation));

		bcopy((caddr_t)&pvip_attr, (caddr_t)&attr, sizeof (attr));

		mutex_enter(&softc->mutex);
		attr.fbtype.fb_type = softc->emulation;
		attr.fbtype.fb_width = softc->w;
		attr.fbtype.fb_height = softc->h;
		attr.fbtype.fb_size = softc->size;
		attr.sattr.emu_type = softc->emulation;
		mutex_exit(&softc->mutex);

		if (ddi_copyout((caddr_t)&attr,
				(caddr_t)data,
				sizeof (struct fbgattr),
				mode))
		return (EFAULT);
	}
	break;

	case FBIOGTYPE: {
		struct fbtype	fb;

		mutex_enter(&softc->mutex);
		bcopy((caddr_t)&pvip_attr.fbtype, (caddr_t)&fb,
			sizeof (struct fbtype));

		DEBUGF(3, (CE_CONT, "FBIOGTYPE\n"));

		fb.fb_width = softc->w;
		fb.fb_height = softc->h;
		fb.fb_size = softc->size;
		mutex_exit(&softc->mutex);

		if (ddi_copyout((caddr_t)&fb,
			(caddr_t)data, sizeof (struct fbtype), mode))

		return (EFAULT);
	}
	break;

	case FBIOSVIDEO:
		DEBUGF(3, (CE_CONT, "FBIOSVIDEO\n"));
		if (ddi_copyin((caddr_t)data,
			(caddr_t)&i, sizeof (int), mode))
			return (EFAULT);

		mutex_enter(&softc->mutex);
		pvip_set_video(softc, i);
		mutex_exit(&softc->mutex);
	break;

	case FBIOGVIDEO:
		DEBUGF(3, (CE_CONT, "FBIOGVIDEO\n"));
		mutex_enter(&softc->mutex);
		i = (ddi_getb(softc->pvip_bt_handle,
			&pvipreg->bt485_comreg0) &
			BT485_POWER_DOWN_ENABLE) ? FBVIDEO_OFF : FBVIDEO_ON;
		mutex_exit(&softc->mutex);

		if (ddi_copyout((caddr_t)&i,
			(caddr_t)data, sizeof (int), mode))
		return (EFAULT);
	break;

#if ALWAYS_NO_INTERRUPTS
	/* vertical retrace interrupt */

	case FBIOVERTICAL:

		if (softc->have_intr) {
			mutex_enter(&softc->mutex);
			softc->vrtflag |= PVIP_VRTIOCTL;
			pvip_int_enable(softc);
			cv_wait(&softc->vrtsleep, &softc->mutex);
			mutex_exit(&softc->mutex);
		} else
			PVIP_WAIT_VSYNC(pvipreg);
		return (0);

	case FBIOVRTOFFSET:
		i = PVIP_VRT_VADDR;

		if (ddi_copyout((caddr_t)&i,
				(caddr_t)data, sizeof (int), mode))
			return (EFAULT);
		return (0);

#endif /* ALWAYS_NO_INTERRUPTS */

	/* HW cursor control */
	case FBIOSCURSOR: {
		int		set;
		int		cbytes;
		ulong_t	stack_image[32];
		ulong_t	stack_mask[32];

		if (ddi_copyin((caddr_t)data,
				(caddr_t)&cp, sizeof (struct fbcursor), mode))
			return (EFAULT);

		set = cp.set;
		softc->cur.size = cp.size;

		/* compute cursor bitmap bytes */
		cbytes = softc->cur.size.y * sizeof (softc->cur.image[0]);
		if (set & FB_CUR_SETSHAPE) {

			if ((u_int) cp.size.x > 32 || (u_int) cp.size.y > 32) {
				return (EINVAL);
			}

			/* copy cursor image into softc */
			if (cp.image != NULL) {
				if (ddi_copyin((caddr_t)cp.image,
						(caddr_t)&stack_image,
						cbytes, mode))
					return (EFAULT);
			}

			if (cp.mask != NULL) {
				if (ddi_copyin((caddr_t)cp.mask,
						(caddr_t)&stack_mask,
						cbytes, mode))
					return (EFAULT);
			}
		}

		mutex_enter(&softc->mutex);
		if (set & FB_CUR_SETCUR)
			softc->cur.enable = cp.enable;

		if (set & FB_CUR_SETPOS)
			softc->cur.pos = cp.pos;

		if (set & FB_CUR_SETHOT)
			softc->cur.hot = cp.hot;

		/* update hardware */
		viper_setcurpos(softc);

		if (set & FB_CUR_SETSHAPE) {

			if (cp.image != NULL) {
				bzero((caddr_t)softc->cur.image,
					sizeof (softc->cur.image));
				bcopy((caddr_t)&stack_image,
					(caddr_t)softc->cur.image, cbytes);
			}

			if (cp.mask != NULL) {
				bzero((caddr_t)softc->cur.mask,
					sizeof (softc->cur.mask));
				bcopy((caddr_t)&stack_mask,
					(caddr_t)softc->cur.mask, cbytes);
			}

			/* load into hardware */
			viper_setcurshape(softc);
		}

		mutex_exit(&softc->mutex);

		/* load colormap */
		if (set & FB_CUR_SETCMAP) {
			cursor_cmap = 1;

			cmd = FBIOPUTCMAP;
			data = (int) &cp.cmap;
			goto cmap_ioctl;
		}
		break;
	}

	case FBIOGCURSOR: {
		int		cbytes;
		ulong_t	stack_image[32];
		ulong_t	stack_mask[32];

		if (ddi_copyin((caddr_t)data,
				(caddr_t)&cp, sizeof (struct fbcursor), mode))
			return (EFAULT);

		mutex_enter(&softc->mutex);
		cp.set = 0;
		cp.enable = softc->cur.enable;
		cp.pos = softc->cur.pos;
		cp.hot = softc->cur.hot;
		cp.size = softc->cur.size;
		cp.cmap.index = 0;
		cp.cmap.count = PVIP_CURSOR_ENTRIES;

		/* compute cursor bitmap bytes */
		cbytes = softc->cur.size.y * sizeof (softc->cur.image[0]);

		bcopy((caddr_t)softc->cur.image,
			(caddr_t)&stack_image, cbytes);
		bcopy((caddr_t)softc->cur.mask, (caddr_t)&stack_mask, cbytes);

		mutex_exit(&softc->mutex);

		if (ddi_copyout((caddr_t)&cp,
				(caddr_t)data, sizeof (struct fbcursor), mode))
			return (EFAULT);

		/* if image pointer is non-null copy both bitmaps */
		if (cp.image != NULL) {
			if (ddi_copyout((caddr_t)&stack_image,
					(caddr_t)cp.image, cbytes, mode))
				return (EFAULT);

			if (ddi_copyout((caddr_t)&stack_mask,
				(caddr_t)cp.mask, cbytes, mode))
				return (EFAULT);
		}

		/* if red pointer is non-null copy colormap */
		if (cp.cmap.red) {
			cursor_cmap = 1;
			cmd = FBIOGETCMAP;
			data = (int) &cp.cmap;
			goto cmap_ioctl;
		}
	}
	break;

	case FBIOSCURPOS: {

		struct fbcurpos	stack_curpos; /* cursor position */

		if (ddi_copyin((caddr_t)data,
				(caddr_t)&stack_curpos,
				sizeof (struct fbcurpos), mode))
			return (EFAULT);

		mutex_enter(&softc->mutex);

		bcopy((caddr_t)&stack_curpos, (caddr_t)&softc->cur.pos,
			sizeof (struct fbcurpos));
		viper_setcurpos(softc);

		mutex_exit(&softc->mutex);
	}
	break;

	case FBIOGCURPOS: {

		struct fbcurpos	stack_curpos; /* cursor position */

		mutex_enter(&softc->mutex);
		bcopy((caddr_t)&softc->cur.pos,
			(caddr_t)&stack_curpos, sizeof (struct fbcurpos));
		mutex_exit(&softc->mutex);

		if (ddi_copyout((caddr_t)&stack_curpos,
				(caddr_t)data,
				sizeof (struct fbcurpos), mode))
			return (EFAULT);
	}
	break;

	case FBIOGCURMAX: {
		static struct fbcurpos curmax = {32, 32};

		if (ddi_copyout((caddr_t)&curmax,
				(caddr_t)data, sizeof (struct fbcurpos),
				mode))
			return (EFAULT);
	}
	break;

#if PVIPDEBUG >= 3
	case 253:
		break;

	case 255:
		pvip_debug = (int) data;
		if (pvip_debug == -1)
			pvip_debug = PVIPDEBUG;
		cmn_err(CE_CONT, "pvip_debug is now %d\n", pvip_debug);
		break;

#endif /* PVIPDEBUG */

	default:
		return (viper_ioctl(dev, cmd, data, mode));
	}				/* switch(cmd) */

	return (0);
}


static u_int
pvip_intr(caddr_t arg)
{
	register struct pvip_softc *softc = (struct pvip_softc *) arg;

	DEBUGF(7, (CE_CONT,
		"pvip_intr: softc=%x, vrtflag=%x\n", softc, softc->vrtflag));

	mutex_enter(&softc->mutex);
	mutex_enter(&softc->interlock);

	if (!(pvip_update_pending(softc) || (softc)->vrtflag)) {

		/* TODO catch stray interrupts? */
		pvip_int_disable_intr(softc);
		pvip_int_clear(softc);
		mutex_exit(&softc->interlock);
		mutex_exit(&softc->mutex);

		return (DDI_INTR_CLAIMED);
	}

#if ALWAYS_NO_INTERRUPTS

	if (softc->vrtflag & PVIP_VRTCTR) {
		if (softc->vrtmaps == 0) {
			softc->vrtflag &= ~PVIP_VRTCTR;
		}
	else
		*softc->vrtpage += 1;
	}

	if (softc->vrtflag & PVIP_VRTIOCTL) {
		softc->vrtflag &= ~PVIP_VRTIOCTL;
		cv_broadcast(&softc->vrtsleep);
	}

#endif /* ALWAYS_NO_INTERRUPTS */

	if (pvip_update_pending(softc)) {

		/* load cursor color map */
		if (softc->omap_update) {
			viper_setcurcolor(softc);
			softc->omap_update = 0;
		}

		/* load main color map */
		if (softc->cmap_count) {
			viper_setpalet(softc);
			softc->cmap_count = 0;
		}
	}

	pvip_int_disable_intr(softc);
	pvip_int_clear(softc);

	if (softc->vrtflag)
		pvip_int_enable(softc);

#ifdef	COMMENT

	if (!softc->vrtflag)
		pvip_int_disable_intr(softc);

#endif	/* COMMENT */

	mutex_exit(&softc->interlock);
	mutex_exit(&softc->mutex);

	return (DDI_INTR_CLAIMED);
}

/*
 * Initialize a colormap: background = white, all others = black
 */
static void
pvip_reset_cmap(register u_char *cmap, register u_int entries)
{
	bzero((char *) cmap, entries * 3);
	cmap[0] = 255;
	cmap[1] = 255;
	cmap[2] = 255;
}

/*
 * Compute color map update parameters: starting index and count.
 * If count is already nonzero, adjust values as necessary.
 * Zero count argument indicates cursor color map update desired.
 */
static void
pvip_update_cmap(register struct pvip_softc *softc, register u_int index,
	register u_int count)
{
	register u_int	high, low;

	if (count == 0) {
		softc->omap_update = 1;
		return;
	}

	high = softc->cmap_count;

	if (high != 0) {
		high += (low = softc->cmap_index);

		if (index < low)
			softc->cmap_index = low = index;

		if (index + count > high)
			high = index + count;

		softc->cmap_count = high - low;
	} else {
		softc->cmap_index = index;
		softc->cmap_count = count;
	}
}


static void
pvip_update_omap(register struct pvip_softc *softc)
{
	softc->omap_update = 1;
}

/*
 * Copy colormap entries between red, green, or blue array to
 * interspersed rgb array.
 */
static void
pvip_cmap_copyin(register u_char *rgb, register u_char *buf,
	register u_int count)
{
	while (count != 0) {
		*rgb = *buf;
		rgb += 3;
		buf++;
		count--;
	}
}

/*
 * Copy colormap entries between interspersed rgb array to
 * red, green, or blue array.
 */
static void
pvip_cmap_copyout(register u_char *buf, register u_char *rgb,
	register u_int count)
{
	while (count != 0) {
		*buf = *rgb;
		rgb += 3;
		buf++;
		count--;
	}
}


/*
 * P9000 Context Management
 *
 * This virtualizes the P9000 register file by associating a register
 * save area with each mapping of the device and using the ddi_mapdev
 * mechanism to allow us to detect mapping changes.
 * Only one of these contexts is valid at any time; an access to one
 * of the invalid contexts saves the current P9000 context, invalidates
 * the current valid mapping, restores the former register contents
 * appropriate to the faulting mapping, and then validates it.
 *
 * This implements a graphical context switch that is transparent to
 * the user.
 *
 * the REGS contain the interesting context registers.
 *
 * Per-mapping info:
 *	Some, but not all, mappings are part of a context.
 *	Any mapping that is MAP_PRIVATE to the REGS will be
 *	part of a unique context.MAP_SHARED mappings are part
 *	of the shared context and all such programs must arbitrate among
 *	themselves to keep from stepping on each other's register settings.
 *	Mappings to the framebuffer may or may not be part of a context,
 *	depending on exact hardware type.
 */

#define	PVIP_MAP_SHARED 0x02	/* shared context */
#define	PVIP_MAP_LOCK	0x04	/* lock page */
#if ALWAYS_NO_INTERRUPTS
#define	PVIP_MAP_VRT	0x08	/* vrt page */
#endif
#define	PVIP_MAP_REGS	0x10	/* mapping includes registers */
#define	PVIP_MAP_FB	0x40	/* mapping includes framebuffer */

#define	PVIP_MAP_CTX	(PVIP_MAP_REGS | PVIP_MAP_FB) /* needs context */

static struct pvip_map_pvt *
pvip_pvt_alloc(register struct pvip_cntxt *ctx, register u_int off,
	register u_int len, register struct pvip_softc *softc)
{
	register struct pvip_map_pvt	*pvt;

	/*
	 * create the private data portion of the mapdev object
	 */
	pvt = (struct pvip_map_pvt *)kmem_zalloc(sizeof (struct pvip_map_pvt),
		KM_SLEEP);

	pvt->offset = off;
	pvt->len = len;
	pvt->context = ctx;
	pvt->softc = softc;

	/*
	 * Link this pvt into the list of associated pvt's for this
	 * context.
	 */
	pvt->next = ctx->pvip_ctx_pvt;
	ctx->pvip_ctx_pvt = pvt;

	return (pvt);
}

/*
 * This routine is called through the cb_ops table to handle
 * pvip mmap requests.
 */
/*ARGSUSED*/
static int
pvip_segmap(register dev_t dev, register off_t off,
	register struct as *as, register caddr_t *addrp,
	register off_t len, register u_int prot,
	register u_int maxprot, register u_int flags,
	register cred_t *cred)
{
	register struct pvip_softc *softc = getsoftc(getminor(dev));
	register struct pvip_cntxt *ctx = (struct pvip_cntxt *)NULL;
	register struct pvip_cntxt *shared_ctx = &softc->shared_ctx;
	register struct pvip_map_pvt	*pvt;
	register u_int	maptype = 0;
	register int	error;
	register int	i;

	DEBUGF(3, (CE_CONT, "segmap: off=%x, len=%x\n", off, len));

	mutex_enter(&softc->mutex);

	/*
	 * Validate and categorize the map request.Valid mmaps to
	 * the pvip driver are to device hardware or to the vertical
	 * retrace counter page
	 *
	 * VRT page mapping?  If so, be sure to count VRT events
	 */
#if ALWAYS_NO_INTERRUPTS
	if (off == PVIP_VRT_VADDR) {
		if (!softc->have_intr)
			return (EIO);

		if (len != pvip_pagesize) {
			mutex_exit(&softc->mutex);

			DEBUGF(3, (CE_CONT,
				"rejecting: off=vrt and len=%x\n", len));

			return (EINVAL);
		}

		maptype = PVIP_MAP_VRT;

		if (softc->vrtmaps++ == 0) {
			if (softc->vrtpage == NULL)
				softc->vrtalloc = (int *)
					kmem_alloc(pvip_pagesize * 2,
					KM_SLEEP);
			softc->vrtpage = (int *)
					(((ulong_t)softc->vrtalloc +
					pvip_pagesize) & ~pvip_pageoffset);
			*softc->vrtpage = 0;
			softc->vrtflag |= PVIP_VRTCTR;
			pvip_int_enable(softc);
		}
	} else {

#else /* ALWAYS_NO_INTERRUPTS */

	{

#endif /* ALWAYS_NO_INTERRUPTS */

		/*
		 * Check to insure that the entire range is legal and we are
		 * not trying to map more than defined by the device.
		 */
		for (i = 0; i < len; i += pvip_pagesize) {
			if (pvip_mmap(dev, (off_t)off + i,
					(int) maxprot) == -1) {
				mutex_exit(&softc->mutex);
				DEBUGF(3, (CE_CONT,
					"rejecting mmap returns -1, off=%x\n",
					off + i));
				return (ENXIO);
			}
		}

		/* classify it */
		if ((PVIP_REGISTER_BASE <= off) &&
				(off + (len - 1)) <
				(PVIP_REGISTER_BASE + PVIP_REGISTER_SIZE)) {
			maptype |= PVIP_MAP_CTX;
		}
	}


	/*
	 * Is this mapping for part of the pvip context?
	 * If so, splice it into the softc context list.
	 */
	if (maptype & PVIP_MAP_CTX) {

		/* Is this a shared mapping.  If so use the shared_ctx. */
		if (flags & MAP_SHARED) {
			ctx = shared_ctx;
			/* XXX: move to attach */
			ctx->pvip_ctx_flag = PVIP_MAP_CTX;
		} else {
			ctx = pvip_ctx_map_insert(softc, maptype);
			ctx->pvip_ctx_flag |= maptype;
		}

		pvt = pvip_pvt_alloc(ctx, 0, (u_int) len, softc);

		/* create the mapdev object */
		error = ddi_mapdev(dev, off, as, addrp, len, prot, maxprot,
			flags, cred, &pvip_map_ops, &pvt->handle,
			(void *) pvt);

		if (error != 0) {
			cmn_err(CE_WARN, "ddi_mapdev failure\n");
		}

		error = ddi_mapdev_set_device_acc_attr(
				pvt->handle,
				0,
				len,
				&endian_attr,
				1);

		if (error != 0) {
			cmn_err(CE_WARN, "ddi_mapdev_set_device failure\n");
		}

	} else {

		/*
		 * Yet another kludge: some programs may decide to map random
		 * parts of the device MAP_PRIVATE, even those that don't form
		 * part of the context. This has the effect of potentially
		 * asking ddi_segmap(aka spec_segmap) to create a MAP_PRIVATE
		 * mapping, which it will not do. We turn all gratuitous
		 * private mappings into shared ones here.
		 *
		 * XXX do we need this check?
		 */
		if (flags & MAP_PRIVATE)
			flags = (flags & ~MAP_PRIVATE) | MAP_SHARED;

		error = ddi_segmap_setup(dev,
				off,
				as,
				addrp,
				len,
				prot,
				maxprot,
				flags,
				cred,
				&endian_attr,
				1);

	}

	mutex_exit(&softc->mutex);

	return (error);
}

/*
 * An access has been made to a context other than the current one
 */
/*ARGSUSED*/
static int
pvip_map_access(register ddi_mapdev_handle_t handle,
	register void *pvt,
	register off_t offset)
{
	register struct pvip_map_pvt *p = (struct pvip_map_pvt *)pvt;
	register struct pvip_map_pvt	*pvts;
	register struct pvip_softc *softc = p->softc;
	volatile pvipp_t	pvipreg = &softc->pvip;
	register ddi_acc_handle_t cr_handle = softc->pvip_cr_handle[0];
	register int	err = 0;

	ASSERT(pvt);

	mutex_enter(&softc->mutex);

	/*
	 * Do we need to switch contexts?
	 */
	if (softc->curctx != p->context) {

		/*
		 * If there's a current context, save it
		 */
		if (softc->curctx != (struct pvip_cntxt *)NULL) {
			/*
			 * Set mapdev for current context and all associated
			 * handles to intercept references to their addresses
			 */
			ASSERT(softc->curctx->pvip_ctx_pvt);
			for (pvts = softc->curctx->pvip_ctx_pvt; pvts != NULL;
				pvts = pvts->next) {

				err = ddi_mapdev_intercept(pvts->handle,
					pvts->offset, pvts->len);

				if (err) {
					mutex_exit(&softc->mutex);
					return (err);
				}
			}

			if (pvip_cntxsave(softc, softc->curctx) == 0) {
				DEBUGF(1, (CE_CONT,
					"pvip: context save failed\n"));

				/*
				 * At this point we have no current context.
				 */
				softc->curctx = NULL;

				mutex_exit(&softc->mutex);
				return (-1);
			}
		}

		/*
		 * Idle the chips
		 */
		CDELAY(!(ddi_getl(cr_handle, &pvipreg->pvip_status) &
			PVIP_STATUS_QUAD_OR_BUSY), PVIP_WAIT);

		if (ddi_getl(cr_handle, &pvipreg->pvip_status) &
			PVIP_STATUS_QUAD_OR_BUSY) {

			DEBUGF(1, (CE_CONT, "pvip: idle_pvip: status = %x\n",
			ddi_getl(cr_handle, &pvipreg->pvip_status)));

			/*
			 * At this point we have no current context.
			 */
			softc->curctx = NULL;

			mutex_exit(&softc->mutex);
			return (-1);
		}

		DEBUGF(4, (CE_CONT, "loading context %x\n", p->context));

		if (p->context->pvip_ctx_flag & PVIP_MAP_REGS)
			if (!pvip_cntxrestore(softc, p->context)) {
				DEBUGF(1, (CE_CONT,
					"pvip: context restore failed\n"));

				/*
				 * At this point we have no current context.
				 */
				softc->curctx = NULL;

			mutex_exit(&softc->mutex);
			return (-1);
		}

		/*
		 * switch software "context"
		 */
		softc->curctx = p->context;
	}

	ASSERT(p->context->pvip_ctx_pvt);

	for (pvts = p->context->pvip_ctx_pvt; pvts != NULL;
			pvts = pvts->next) {
		if ((err = ddi_mapdev_nointercept(pvts->handle,
			pvts->offset, pvts->len)) != 0) {
			mutex_exit(&softc->mutex);
			return (err);
		}
	}

	mutex_exit(&softc->mutex);
	return (err);
}

/*ARGSUSED*/
static void
pvip_map_free(register ddi_mapdev_handle_t handle,
	register void *pvt)
{
	register struct pvip_map_pvt *p = (struct pvip_map_pvt *)pvt;
	register struct pvip_map_pvt	*pvts;
	register struct pvip_map_pvt	*ppvts;
	register struct pvip_cntxt *ctx = p->context;
	register struct pvip_softc *softc = p->softc;
	register struct pvip_cntxt *shared_ctx = &softc->shared_ctx;

	mutex_enter(&softc->mutex);

	DEBUGF(4, (CE_CONT, "pvip_map_free: cleaning up pid %d\n", ctx->pid));

	/*
	* Remove the pvt data
	*/
	ppvts = NULL;
	for (pvts = ctx->pvip_ctx_pvt; pvts != NULL; pvts = pvts->next) {
		if (pvts == pvt) {
			if (ppvts == NULL) {
				ctx->pvip_ctx_pvt = pvts->next;
			} else {
				ppvts->next = pvts->next;
			}
			kmem_free(pvt, sizeof (struct pvip_map_pvt));
			break;
		}
		ppvts = pvt;
	}

	/*
	 * Remove the context if this is not the shared context and there are
	 * no more associated pvt's
	 */
	if ((ctx != shared_ctx) && (ctx->pvip_ctx_pvt == NULL)) {
		register struct pvip_cntxt *ctxptr;

		if (ctx == softc->curctx)
			softc->curctx = NULL;

		/*
		 * Scan private context list for entry to remove.
		 * Check first to see if it's the head of our list.
		 */
		if (softc->pvt_ctx == ctx) {
			softc->pvt_ctx = ctx->link;
			kmem_free(ctx, sizeof (struct pvip_cntxt));
		} else {
			for (ctxptr = softc->pvt_ctx; ctxptr != NULL;
					ctxptr = ctxptr->link) {
				if (ctxptr->link == ctx) {
					ctxptr->link = ctx->link;
					kmem_free(ctx,
						sizeof (struct pvip_cntxt));
				}
			}
		}
	}

	/*
	* If the curctx is the shared context, and there are no
	* more pvt's for the shared context, set the curctx to
	* NULL to force a context switch on the next device access.
	*/
	if ((softc->curctx == shared_ctx) &&
			(softc->curctx->pvip_ctx_pvt == NULL)) {
		softc->curctx = NULL;
	}

	mutex_exit(&softc->mutex);
}

/*ARGSUSED*/
static int
pvip_map_dup(register ddi_mapdev_handle_t old_handle,
	register void *oldpvt, register ddi_mapdev_handle_t new_handle,
	register void **newpvt)
{
	register struct pvip_map_pvt *p = (struct pvip_map_pvt *)oldpvt;
	register struct pvip_softc *softc = p->softc;
	register struct pvip_map_pvt	*pvt;
	register struct pvip_cntxt		*ctx;

	mutex_enter(&softc->mutex);
	if (p->context != &softc->shared_ctx) {
		ctx = (struct pvip_cntxt *)
			kmem_zalloc(sizeof (struct pvip_cntxt), KM_SLEEP);

		*ctx = *p->context;
		ctx->pvip_ctx_pvt = NULL;
	} else
		ctx = &softc->shared_ctx;

	pvt = pvip_pvt_alloc(ctx, 0, p->len, softc);

	pvt->handle = new_handle;
	*newpvt = pvt;

#if ALWAYS_NO_INTERRUPTS

	if (p->context && (p->context->pvip_ctx_flag & PVIP_MAP_VRT)) {
		softc->vrtflag |= PVIP_VRTCTR;

		if (softc->vrtmaps == 0)
			pvip_int_enable(softc);

		softc->vrtmaps++;
	}

#endif /* ALWAYS_NO_INTERRUPTS */

	mutex_exit(&softc->mutex);
	return (0);
}

static void
pvip_set_video(struct pvip_softc *softc, int v)
{
	volatile pvipp_t	pvipreg = &softc->pvip;
	register ddi_acc_handle_t handle = softc->pvip_bt_handle;

	if (v & FBVIDEO_ON)
		ddi_putb(handle, &pvipreg->bt485_comreg0,
			ddi_getb(handle, &pvipreg->bt485_comreg0) & ~
			BT485_POWER_DOWN_ENABLE);
	else
		ddi_putb(handle, &pvipreg->bt485_comreg0,
			ddi_getb(handle, &pvipreg->bt485_comreg0) |
			BT485_POWER_DOWN_ENABLE);
}

static int
pvip_cntxsave(register struct pvip_softc *softc, struct pvip_cntxt *saved)
{
	volatile pvipp_t	pvipreg = &softc->pvip;
	register ddi_acc_handle_t cr_handle = softc->pvip_cr_handle[0];

	DEBUGF(5, (CE_CONT, "saving registers\n"));

	CDELAY(!(ddi_getl(cr_handle, &pvipreg->pvip_status) &
		PVIP_STATUS_QUAD_OR_BUSY), PVIP_WAIT);

	if (ddi_getl(cr_handle,
			&pvipreg->pvip_status) & PVIP_STATUS_QUAD_OR_BUSY) {
		DEBUGF(1, (CE_CONT,
			"pvip: cntxsave: status = %x\n",
		ddi_getl(cr_handle, &pvipreg->pvip_status)));
		return (0);
	}

	/*
	 * start dumping stuff out.
	 */

	saved->pvip_ctx_interrupt =
		ddi_getl(cr_handle, &pvipreg->pvip_interrupt);
	saved->pvip_ctx_interrupt_en =
		ddi_getl(cr_handle, &pvipreg->pvip_interrupt_en);
	saved->pvip_ctx_cindex = ddi_getl(cr_handle, &pvipreg->pvip_cindex);
	saved->pvip_ctx_w_off_xy =
		ddi_getl(cr_handle, (uint32_t *)&pvipreg->pvip_w_off_xy);
	saved->pvip_ctx_fground = ddi_getl(cr_handle, &pvipreg->pvip_fground);
	saved->pvip_ctx_bground = ddi_getl(cr_handle, &pvipreg->pvip_bground);
	saved->pvip_ctx_pmask = ddi_getl(cr_handle, &pvipreg->pvip_pmask);
	saved->pvip_ctx_draw_mode =
		ddi_getl(cr_handle, &pvipreg->pvip_draw_mode);
	saved->pvip_ctx_pat_originx =
		ddi_getl(cr_handle, (uint32_t *)&pvipreg->pvip_pat_originx);
	saved->pvip_ctx_pat_originy =
		ddi_getl(cr_handle, (uint32_t *)&pvipreg->pvip_pat_originy);
	saved->pvip_ctx_raster = ddi_getl(cr_handle, &pvipreg->pvip_raster);
	saved->pvip_ctx_pixel8_reg =
		ddi_getl(cr_handle, &pvipreg->pvip_pixel8_reg);
	saved->pvip_ctx_w_min = ddi_getl(cr_handle,
		(uint32_t *)&pvipreg->pvip_w_min);
	saved->pvip_ctx_w_max = ddi_getl(cr_handle,
		(uint32_t *)&pvipreg->pvip_w_max);
	ddi_rep_getl(cr_handle, (uint32_t *)&saved->pvip_ctx_x0,
		pvipreg->pvip_pattern,
		PVIP_NUM_PAT, DDI_DEV_AUTOINCR);
	saved->pvip_ctx_x0 = ddi_getl(cr_handle, (uint32_t *)&pvipreg->pvip_x0);
	saved->pvip_ctx_y0 = ddi_getl(cr_handle, (uint32_t *)&pvipreg->pvip_y0);
	saved->pvip_ctx_x1 = ddi_getl(cr_handle, (uint32_t *)&pvipreg->pvip_x1);
	saved->pvip_ctx_y1 = ddi_getl(cr_handle, (uint32_t *)&pvipreg->pvip_y1);
	saved->pvip_ctx_x2 = ddi_getl(cr_handle, (uint32_t *)&pvipreg->pvip_x2);
	saved->pvip_ctx_y2 = ddi_getl(cr_handle, (uint32_t *)&pvipreg->pvip_y2);
	saved->pvip_ctx_x3 = ddi_getl(cr_handle, (uint32_t *)&pvipreg->pvip_x3);
	saved->pvip_ctx_y3 = ddi_getl(cr_handle, (uint32_t *)&pvipreg->pvip_y3);
	return (1);
}

static int
pvip_cntxrestore(register struct pvip_softc *softc,
	register struct pvip_cntxt *saved)
{
	volatile pvipp_t	pvipreg = &softc->pvip;
	register ddi_acc_handle_t cr_handle = softc->pvip_cr_handle[0];

	DEBUGF(5, (CE_CONT, "restoring registers\n"));

	CDELAY(!(ddi_getl(cr_handle, &pvipreg->pvip_status) &
		PVIP_STATUS_QUAD_OR_BUSY), PVIP_WAIT);

	if (ddi_getl(cr_handle,
			&pvipreg->pvip_status) & PVIP_STATUS_QUAD_OR_BUSY) {

		DEBUGF(1, (CE_CONT,
			"pvip: cntxrestore: status = %x\n",

		ddi_getl(cr_handle, &pvipreg->pvip_status)));
		return (0);
	}

	/*
	 * start restoring stuff.
	 */
	ddi_putl(cr_handle, &pvipreg->pvip_interrupt,
		saved->pvip_ctx_interrupt | PVIP_INT_VBLANKED_CTRL |
		PVIP_INT_PICKED_CTRL | PVIP_INT_DE_IDLE_CTRL);

	ddi_putl(cr_handle, &pvipreg->pvip_interrupt_en,
		saved->pvip_ctx_interrupt_en | PVIP_INTEN_MEN_CTRL |
		PVIP_INTEN_VBLANKED_EN_CTRL | PVIP_INTEN_PICKED_EN_CTRL |
		PVIP_INTEN_DE_IDLE_EN_CTRL);

	ddi_putl(cr_handle, &pvipreg->pvip_cindex, saved->pvip_ctx_cindex);
	ddi_putl(cr_handle, (uint32_t *)&pvipreg->pvip_w_off_xy,
		saved->pvip_ctx_w_off_xy);
	ddi_putl(cr_handle, &pvipreg->pvip_fground, saved->pvip_ctx_fground);
	ddi_putl(cr_handle, &pvipreg->pvip_bground, saved->pvip_ctx_bground);
	ddi_putl(cr_handle, &pvipreg->pvip_pmask, saved->pvip_ctx_pmask);
	ddi_putl(cr_handle,
		&pvipreg->pvip_draw_mode, saved->pvip_ctx_draw_mode);
	ddi_putl(cr_handle, (uint32_t *)&pvipreg->pvip_pat_originx,
		saved->pvip_ctx_pat_originx);
	ddi_putl(cr_handle, (uint32_t *)&pvipreg->pvip_pat_originy,
		saved->pvip_ctx_pat_originy);
	ddi_putl(cr_handle, &pvipreg->pvip_raster, saved->pvip_ctx_raster);
	ddi_putl(cr_handle,
		&pvipreg->pvip_pixel8_reg, saved->pvip_ctx_pixel8_reg);
	ddi_putl(cr_handle, (uint32_t *)&pvipreg->pvip_w_min,
		saved->pvip_ctx_w_min);
	ddi_putl(cr_handle, (uint32_t *)&pvipreg->pvip_w_max,
		saved->pvip_ctx_w_max);
	ddi_rep_putl(cr_handle,
		saved->pvip_ctx_pattern, (uint32_t *)&pvipreg->pvip_pattern,
		PVIP_NUM_PAT, DDI_DEV_AUTOINCR);
	ddi_putl(cr_handle, (uint32_t *)&pvipreg->pvip_x0, saved->pvip_ctx_x0);
	ddi_putl(cr_handle, (uint32_t *)&pvipreg->pvip_y0, saved->pvip_ctx_y0);
	ddi_putl(cr_handle, (uint32_t *)&pvipreg->pvip_x1, saved->pvip_ctx_x1);
	ddi_putl(cr_handle, (uint32_t *)&pvipreg->pvip_y1, saved->pvip_ctx_y1);
	ddi_putl(cr_handle, (uint32_t *)&pvipreg->pvip_x2, saved->pvip_ctx_x2);
	ddi_putl(cr_handle, (uint32_t *)&pvipreg->pvip_y2, saved->pvip_ctx_y2);
	ddi_putl(cr_handle, (uint32_t *)&pvipreg->pvip_x3, saved->pvip_ctx_x3);
	ddi_putl(cr_handle, (uint32_t *)&pvipreg->pvip_y3, saved->pvip_ctx_y3);
	return (1);
}


/*
 * pvip_ctx_map_insert()
 *
 * Insert a mapping into the mapping list of a private context.First
 * determine if there's an existing context (e.g. one with the same PID
 * as the currentone and that does not already have a mapping of this
 * type yet).If not, allocate a new one.Then insert mapping into this
 * context's list.
 *
 * The softc mutex must be held across calls to this routine.
 */
static
struct pvip_cntxt *
pvip_ctx_map_insert(struct pvip_softc *softc, int maptype)
{
	register struct pvip_cntxt	*ctx;
	register struct pvip_cntxt	*pvt_ctx_list;
	int		curpid = pvip_getpid();

	DEBUGF(4, (CE_CONT, "pvip_ctx_map_insert: maptype=0x%x curpid=%d\n",
		maptype, curpid));

	/*
	 * If this is the first time we're here, then alloc space
	 * for new context and depart.
	 */
	if (softc->pvt_ctx == NULL) {
		ctx = (struct pvip_cntxt *)
		kmem_zalloc(sizeof (struct pvip_cntxt), KM_SLEEP);
		ctx->pid = curpid;
		ctx->link = NULL;
		softc->pvt_ctx = ctx;
		return (ctx);
	}

	pvt_ctx_list = softc->pvt_ctx;

	/*
	 * Find existing context if one exists.  We have a match if
	 * we're the same process *and* there's not already a
	 * mapping of this type assigned.
	 */
	for (ctx = pvt_ctx_list; ctx != NULL; ctx = ctx->link) {
		if (ctx->pid == curpid &&
			(maptype & ctx->pvip_ctx_flag & PVIP_MAP_CTX) == 0)
		break;
	}

	/* no match, create a new one and add to softc list */
	if (ctx == NULL) {
		ctx = (struct pvip_cntxt *)
			kmem_zalloc(sizeof (struct pvip_cntxt), KM_SLEEP);
		ctx->pid = curpid;
		ctx->link = softc->pvt_ctx;
		softc->pvt_ctx = ctx;
	}

	return (ctx);
}

/*
 * pvip_getpid()
 *
 * Simple wrapper around process ID call to drv_getparm(9f).
 */
static int
pvip_getpid(void)
{
	ulong_t	rval;

	if (drv_getparm(PPID, &rval) == -1)
		return (0);
	else
		return ((int)rval);
}


/*
 * Viper specific routines
 */

static int
viper_ioctl(register dev_t dev, register int cmd, register int data,
	register int mode)
{
	register struct pvip_softc *softc = getsoftc(getminor(dev));
	register int	status;

	switch (cmd) {
	case VIPERIO_GET_INIT:
		mutex_enter(&softc->mutex);
		status = ddi_copyout((caddr_t)&softc->viperinit,
				(caddr_t)data, sizeof (viper_init_t),
				mode);
		mutex_exit(&softc->mutex);
		if (status)
			return (EFAULT);
		break;

	case VIPERIO_PUT_INIT:
		mutex_enter(&softc->mutex);
		status = ddi_copyin((caddr_t)data,
				(caddr_t)&softc->viperinit,
				sizeof (viper_init_t), mode);
		if (!status) {
			softc->w = softc->viperinit.vi_width;
			softc->h = softc->viperinit.vi_height;
			softc->size = softc->h *
			PVIP_LINEBYTES(softc->viperinit.vi_sysconfig);
		}
		mutex_exit(&softc->mutex);

		if (status)
			return (EFAULT);
		break;

	case VIPERIO_GET_PORTS:
		status = ddi_copyout((caddr_t)&softc->viperport,
				(caddr_t)data,
				sizeof (viper_port_t),
				mode);
		if (status)
			return (EFAULT);
		break;

	case VIPERIO_GRAPH_MODE:
		mutex_enter(&softc->mutex);
		status = viper_enter_graph_mode(softc);
		mutex_exit(&softc->mutex);
		if (status)
			return (status);
		break;

	case VIPERIO_TEXT_MODE:
		mutex_enter(&softc->mutex);
		status = viper_enter_text_mode(softc);
		mutex_exit(&softc->mutex);
		if (status)
			return (status);
		break;

	case 250:	/* for compatibility */
	case VIPERIO_SET_ICD_CLOCK:
		mutex_enter(&softc->mutex);
		viper_write_ic2061a(softc, data);
		mutex_exit(&softc->mutex);
		break;

	default:
		return (ENOTTY);
	}

	return (0);
}


/*
 * enable/disable/update HW cursor
 */
static void
viper_setcurpos(register struct pvip_softc *softc)
{
	volatile pvipp_t	pvipreg = &softc->pvip;
	register ddi_acc_handle_t handle = softc->pvip_bt_handle;

	if (softc->cur.enable) {
		ddi_putb(handle, &pvipreg->bt485_cursor_x_low,
			(softc->cur.pos.x - softc->cur.hot.x + 32) & 0xff);
		ddi_putb(handle, &pvipreg->bt485_cursor_x_high,
			(softc->cur.pos.x - softc->cur.hot.x + 32) >> 8);
		ddi_putb(handle, &pvipreg->bt485_cursor_y_low,
			(softc->cur.pos.y - softc->cur.hot.y + 32) & 0xff);
		ddi_putb(handle, &pvipreg->bt485_cursor_y_high,
			(softc->cur.pos.y - softc->cur.hot.y + 32) >> 8);
	} else {
		ddi_putb(handle, &pvipreg->bt485_cursor_x_low,
			BT485_CURSOR_OFFPOS & 0xff);
		ddi_putb(handle, &pvipreg->bt485_cursor_x_high,
			BT485_CURSOR_OFFPOS >> 8);
		ddi_putb(handle, &pvipreg->bt485_cursor_y_low,
			BT485_CURSOR_OFFPOS & 0xff);
		ddi_putb(handle, &pvipreg->bt485_cursor_y_high,
			BT485_CURSOR_OFFPOS >> 8);
	}
}

/*
 * load HW cursor bitmaps
 */
static void
viper_setcurshape(register struct pvip_softc *softc)
{
	volatile pvipp_t	pvipreg = &softc->pvip;
	register ddi_acc_handle_t handle = softc->pvip_bt_handle;
	register uint32_t	edge = 0xffffffffUL;
	register uint32_t	*image, *mask;
	register int	i;
	uint32_t	tmp;

	/* compute right edge mask */
	if (softc->cur.size.x < 32)
		edge = (1 << softc->cur.size.x) - 1;

	ddi_putb(handle, &pvipreg->bt485_ram_write, 0);

	for (i = 0, image = softc->cur.image; i < 32; i++, image++) {
		tmp = *image;
		ddi_putb(handle, &pvipreg->bt485_cursor_data,
			pvip_bitflip[tmp & 0xff]);
		ddi_putb(handle, &pvipreg->bt485_cursor_data,
			pvip_bitflip[(tmp >> 8) & 0xff]);
		ddi_putb(handle, &pvipreg->bt485_cursor_data,
			pvip_bitflip[(tmp >> 16) & 0xff]);
		ddi_putb(handle, &pvipreg->bt485_cursor_data,
			pvip_bitflip[(tmp >> 24) & 0xff]);
	}

	for (i = 0, mask = softc->cur.mask; i < 32; i++, mask++) {
		tmp = *mask & edge;
		ddi_putb(handle, &pvipreg->bt485_cursor_data,
			pvip_bitflip[tmp & 0xff]);
		ddi_putb(handle, &pvipreg->bt485_cursor_data,
			pvip_bitflip[(tmp >> 8) & 0xff]);
		ddi_putb(handle, &pvipreg->bt485_cursor_data,
			pvip_bitflip[(tmp >> 16) & 0xff]);
		ddi_putb(handle, &pvipreg->bt485_cursor_data,
			pvip_bitflip[(tmp >> 24) & 0xff]);
	}
}


static void
viper_setcurcolor(register struct pvip_softc *softc)
{
	volatile pvipp_t	pvipreg = &softc->pvip;
	register ddi_acc_handle_t handle = softc->pvip_bt_handle;

	ddi_putb(handle, &pvipreg->bt485_color_write, BT485_CURSOR1_COLOR);

	ddi_rep_putb(handle, (u_char *) softc->omap,
		&pvipreg->bt485_color_data, sizeof (softc->omap),
		DDI_DEV_NO_AUTOINCR);
}


static void
viper_setpalet(register struct pvip_softc *softc)
{
	volatile pvipp_t	pvipreg = &softc->pvip;
	register ddi_acc_handle_t handle = softc->pvip_vga_handle;

	ddi_putb(handle, &pvipreg->bt485_ram_write, softc->cmap_index);

	ddi_rep_putb(handle, (u_char *) &softc->cmap[softc->cmap_index],
		&pvipreg->bt485_palet_data, softc->cmap_count * 3,
		DDI_DEV_NO_AUTOINCR);
}


static int
viper_enter_graph_mode(register struct pvip_softc *softc)
{
	volatile pvipp_t	pvipreg = &softc->pvip;
	register ddi_acc_handle_t bt_handle =
		softc->pvip_bt_handle;
	register ddi_acc_handle_t seq_handle = softc->pvip_vga_handle;
	register ddi_acc_handle_t cr_handle = softc->pvip_cr_handle[0];
	register u_int	value;

	if (!softc->viperinit.vi_valid)
		return (EIO);

	/* disable vga video */
	ddi_putb(seq_handle,
		softc->viperport.vp_seq_index_port, VIPER_SEQ_CLOCK_INDEX);

	ddi_putb(seq_handle, softc->viperport.vp_seq_data_port,
		ddi_getb(seq_handle, softc->viperport.vp_seq_data_port) &
		VIPER_CLOCK_SCREEN_OFF);

	ddi_putb(bt_handle, &pvipreg->bt485_comreg0,
		ddi_getb(bt_handle, &pvipreg->bt485_comreg0) |
		BT485_POWER_DOWN_ENABLE);

	softc->chiprev =
		ddi_getl(cr_handle, &pvipreg->pvip_sysconfig) &
		PVIP_SYSCONF_VERSION_MASK;

	/* enable command register 3 & 8bit DAC */
	ddi_putb(bt_handle, &pvipreg->bt485_comreg0,
		BT485_COMMAND_REGISTER_3_ENABLE | BT485_DAC_8_BIT_RESOLUTION);

	ddi_putb(bt_handle, &pvipreg->bt485_ram_write, BT485_COMREG3_SELECT);

	/* set clock multiplier */
	ddi_putb(bt_handle, &pvipreg->bt485_comreg3,
		softc->viperinit.vi_vclk & VIPER_INIT_VCLK_DOUBLER ?
		BT485_2X_CLOCK_MULTIPLIER_ENABLED : 0);

	/* select viper video clock, x windows cursor */
	ddi_putb(bt_handle, &pvipreg->bt485_comreg2,
		BT485_PORTSEL_UNMASKED | BT485_PCLK1_SELECTED |
		BT485_TWO_COLOR_X_WINDOWS_CURSOR);

	switch (softc->viperinit.vi_depth) {
	case 4:	/* 4 bit pixels, use palette */
		return (-1);

	case 8:	/* 8 bit pixels, use palette */
	default:
		ddi_putb(bt_handle,
			&pvipreg->bt485_comreg1, BT485_8_BIT_PIXELS);
		break;

	case 16:	/* 16 bit (5:5:5) pixels, bypass palette */
		ddi_putb(bt_handle,
			&pvipreg->bt485_comreg1,
			BT485_16_BIT_PIXELS | BT485_TRUE_COLOR_ENABLE);
		break;

	case 24:	/* 24 bit (8:8:8) pixels, bypass palette */
		ddi_putb(bt_handle, &pvipreg->bt485_comreg1,
			BT485_24_BIT_PIXELS | BT485_TRUE_COLOR_ENABLE);
		break;

	case 32:	/* 32 bit (8:8:8) pixels, bypass palette */
		ddi_putb(bt_handle, &pvipreg->bt485_comreg1,
			BT485_24_BIT_PIXELS | BT485_TRUE_COLOR_ENABLE);
		break;
	}

	/* pixel read mask */
	ddi_putb(bt_handle, &pvipreg->bt485_pixel_mask, 0xff);

	ddi_putb(bt_handle, &pvipreg->bt485_cursor_x_low,
		BT485_CURSOR_OFFPOS & 0xff);
	ddi_putb(bt_handle, &pvipreg->bt485_cursor_x_high,
		BT485_CURSOR_OFFPOS >> 8);
	ddi_putb(bt_handle, &pvipreg->bt485_cursor_y_low,
		BT485_CURSOR_OFFPOS & 0xff);
	ddi_putb(bt_handle, &pvipreg->bt485_cursor_y_high,
		BT485_CURSOR_OFFPOS >> 8);

	(void) viper_unlock_regs(softc);

	/* select output reg */
	ddi_putb(seq_handle, softc->viperport.vp_seq_index_port,
		VIPER_SEQ_OUTCTRL_INDEX);

	value = ddi_getb(seq_handle, softc->viperport.vp_seq_data_port);

	value |= VIPER_OUTCTRL_PVIP_VIDEO_ENABLE;

	ddi_putb(seq_handle, softc->viperport.vp_seq_data_port, value);

	(void) viper_lock_regs(softc);

	viper_write_ic2061a(softc, softc->viperinit.vi_vclk >>
			VIPER_INIT_VCLK_DOT_FREQ_SHIFT);

	viper_write_ic2061a(softc, softc->viperinit.vi_memspeed);

	/* note: this needs to be at least 20ms */
	delay(drv_usectohz(20000));

	softc->pu_config = ddi_getl(cr_handle, &pvipreg->pvip_pu_config);

	ddi_putl(cr_handle, &pvipreg->pvip_sysconfig,
		softc->viperinit.vi_sysconfig);

	/* INTERRUPT_EN = disabled */
	ddi_putl(cr_handle, &pvipreg->pvip_interrupt_en,
		PVIP_INTEN_MEN_CTRL | PVIP_INTEN_VBLANKED_EN_CTRL |
		PVIP_INTEN_PICKED_EN_CTRL | PVIP_INTEN_DE_IDLE_EN_CTRL);

	ddi_putl(cr_handle, &pvipreg->pvip_interrupt,
		PVIP_INT_VBLANKED_CTRL | PVIP_INT_PICKED_CTRL |
		PVIP_INT_DE_IDLE_CTRL);

	ddi_putl(cr_handle, &pvipreg->pvip_prehrzc,
		softc->viperinit.vi_prehrzc);
	ddi_putl(cr_handle, &pvipreg->pvip_prevrtc,
		softc->viperinit.vi_prevrtc);
	ddi_putl(cr_handle, &pvipreg->pvip_hrzsr, softc->viperinit.vi_hrzsr);
	ddi_putl(cr_handle, &pvipreg->pvip_hrzbr, softc->viperinit.vi_hrzbr);
	ddi_putl(cr_handle, &pvipreg->pvip_hrzbf, softc->viperinit.vi_hrzbf);
	ddi_putl(cr_handle, &pvipreg->pvip_hrzt, softc->viperinit.vi_hrzt);
	ddi_putl(cr_handle, &pvipreg->pvip_vrtsr, softc->viperinit.vi_vrtsr);
	ddi_putl(cr_handle, &pvipreg->pvip_vrtbr, softc->viperinit.vi_vrtbr);
	ddi_putl(cr_handle, &pvipreg->pvip_vrtbf, softc->viperinit.vi_vrtbf);
	ddi_putl(cr_handle, &pvipreg->pvip_vrtt, softc->viperinit.vi_vrtt);
	ddi_putl(cr_handle, &pvipreg->pvip_rfperiod, 0x186);
	ddi_putl(cr_handle, &pvipreg->pvip_rlmax, 0xfa);

	ddi_putl(cr_handle, &pvipreg->pvip_mem_config,
		softc->viperinit.vi_memconfig);

	value = ddi_getb(seq_handle, softc->viperport.vp_miscin);

	value &= ~(VIPER_MISC_VSYNC_POLARITY | VIPER_MISC_HSYNC_POLARITY);

	if (softc->viperinit.vi_vclk & VIPER_INIT_VCLK_HSYNC)
		value |= VIPER_MISC_HSYNC_POLARITY;
	if (softc->viperinit.vi_vclk & VIPER_INIT_VCLK_VSYNC)
		value |= VIPER_MISC_VSYNC_POLARITY;

	ddi_putb(seq_handle, softc->viperport.vp_miscout, value);

	/* enable video in the video controller */
	ddi_putl(cr_handle, &pvipreg->pvip_srtctl,
		softc->viperinit.vi_srtctl);

	ddi_putb(bt_handle, &pvipreg->bt485_comreg0,
		ddi_getb(bt_handle, &pvipreg->bt485_comreg0) &
		~(BT485_POWER_DOWN_ENABLE | BT485_COMMAND_REGISTER_3_ENABLE));

	return (0);
}


static int
viper_enter_text_mode(register struct pvip_softc *softc)
{
	register pvipp_t	pvipreg = &softc->pvip;
	register ddi_acc_handle_t bt_handle = softc->pvip_bt_handle;
	register ddi_acc_handle_t seq_handle = softc->pvip_vga_handle;
	register ddi_acc_handle_t cr_handle = softc->pvip_cr_handle[0];
	register int		value;

	/* disable video in the video controller */
	ddi_putl(cr_handle, &pvipreg->pvip_srtctl,
		PVIP_SRTCTL_INTERNAL_VSYNC | PVIP_SRTCTL_INTERNAL_HSYNC);

	/* put card in emulation mode */
	(void) viper_unlock_regs(softc);

	/* select output reg */
	ddi_putb(seq_handle, softc->viperport.vp_seq_index_port,
		VIPER_SEQ_OUTCTRL_INDEX);

	value = ddi_getb(seq_handle, softc->viperport.vp_seq_data_port);
	value &= ~VIPER_OUTCTRL_PVIP_VIDEO_ENABLE;
	ddi_putb(seq_handle, softc->viperport.vp_seq_data_port, value);

	(void) viper_lock_regs(softc);

	/* put VGA in 720x400 pixel mode, enable CPU access, color */
	ddi_putb(seq_handle, softc->viperport.vp_miscout,
		VIPER_MISC_HSYNC_POLARITY | VIPER_MISC_PAGE_BIT |
		VIPER_MISC_CLOCK | VIPER_MISC_ENABLE_RAM |
		VIPER_MISC_IO_ADDR);

	/* let ICD to settle note: this needs to be at least 20ms */
	delay(drv_usectohz(20000));

	/* enable Command Reg3 */
	ddi_putb(bt_handle, &pvipreg->bt485_comreg0,
		BT485_COMMAND_REGISTER_3_ENABLE);

	ddi_putb(bt_handle, &pvipreg->bt485_ram_write,
		BT485_COMREG3_SELECT);

	/* disable RAMDAC advance features */
	ddi_putb(bt_handle, &pvipreg->bt485_comreg3, 0);

	ddi_putb(bt_handle, &pvipreg->bt485_comreg2, 0);

	ddi_putb(bt_handle, &pvipreg->bt485_comreg1, 0);

	ddi_putb(bt_handle, &pvipreg->bt485_pixel_mask, 0xff);

	ddi_putb(bt_handle, &pvipreg->bt485_cursor_x_low,
		BT485_CURSOR_OFFPOS & 0xff);
	ddi_putb(bt_handle, &pvipreg->bt485_cursor_x_high,
		BT485_CURSOR_OFFPOS >> 8);
	ddi_putb(bt_handle, &pvipreg->bt485_cursor_y_low,
		BT485_CURSOR_OFFPOS & 0xff);
	ddi_putb(bt_handle, &pvipreg->bt485_cursor_y_high,
		BT485_CURSOR_OFFPOS >> 8);

	/* reenable vga video, just in case it got turned off */
	ddi_putb(seq_handle,
		softc->viperport.vp_seq_index_port, VIPER_SEQ_CLOCK_INDEX);

	ddi_putb(seq_handle, softc->viperport.vp_seq_data_port,
		ddi_getb(seq_handle, softc->viperport.vp_seq_data_port) &
		~VIPER_CLOCK_SCREEN_OFF);


	mutex_enter(&(softc)->interlock);

	pvip_int_disable_intr(softc);

	mutex_exit(&(softc)->interlock);

	return (0);
}


#define	PVIP_UNLOCK_DID_GET_CLOCK	0x0001
#define	PVIP_UNLOCK_SUCCESS		0x0002

static int
viper_unlock_regs(register struct pvip_softc *softc)
{
	register ddi_acc_handle_t seq_handle = softc->pvip_vga_handle;
	register u_char	oldseq;
	register u_char	oldmisc;
	register u_char	oldclockmode;
	register u_char	newmisc;
	register int	status = 0;

	oldseq = ddi_getb(seq_handle, softc->viperport.vp_seq_index_port);

	ddi_putb(seq_handle, softc->viperport.vp_seq_index_port,
		VIPER_SEQ_CLOCK_INDEX);
	if ((ddi_getb(seq_handle, softc->viperport.vp_seq_index_port) &
			VIPER_SEQ_MASK) != 0x01) {
		DEBUGF(2, (CE_CONT,
			" setting clockmode seq failed\n"));
		goto failure;
	}

	oldclockmode = ddi_getb(seq_handle, softc->viperport.vp_seq_data_port);
	status |= PVIP_UNLOCK_DID_GET_CLOCK;

	/* select misc reg */

	ddi_putb(seq_handle,
		softc->viperport.vp_seq_index_port, VIPER_SEQ_MISC_INDEX);

	if ((ddi_getb(seq_handle, softc->viperport.vp_seq_index_port) &
			VIPER_SEQ_MASK) !=
			(VIPER_SEQ_MISC_INDEX & VIPER_SEQ_MASK)) {
		DEBUGF(2, (CE_CONT,
			" setting viper seq misc failed\n"));
		goto failure;
	}

	/* save misc reg in case its unlocked already */
	oldmisc = ddi_getb(seq_handle, softc->viperport.vp_seq_data_port);

	/* unlock misc reg via 2 writes */

	ddi_putb(seq_handle, softc->viperport.vp_seq_data_port, oldmisc);
	ddi_putb(seq_handle, softc->viperport.vp_seq_data_port, oldmisc);

	/* get the value */

	newmisc = ddi_getb(seq_handle, softc->viperport.vp_seq_data_port);
	newmisc &= ~VIPER_MISC_CRLOCK;

	/* clear bit 5 to unlock extended regs */

	ddi_putb(seq_handle, softc->viperport.vp_seq_data_port, newmisc);
	if (ddi_getb(seq_handle, softc->viperport.vp_seq_data_port) !=
			newmisc) {
		DEBUGF(2, (CE_CONT,
			" setting viper misc failed\n"));
		ddi_putb(seq_handle, softc->viperport.vp_seq_data_port,
			oldmisc);
		goto failure;
	}

	status |= PVIP_UNLOCK_SUCCESS;

failure:

	if (status & PVIP_UNLOCK_DID_GET_CLOCK) {
		ddi_putb(seq_handle, softc->viperport.vp_seq_index_port,
			VIPER_SEQ_CLOCK_INDEX);
		ddi_putb(seq_handle,
			softc->viperport.vp_seq_data_port, oldclockmode);
	}

	ddi_putb(seq_handle, softc->viperport.vp_seq_index_port, oldseq);

	return ((status & PVIP_UNLOCK_SUCCESS) != 0);
}


#define	PVIP_LOCK_DID_GET_CLOCK		0x0001
#define	PVIP_LOCK_SUCCESS		0x0002

static int
viper_lock_regs(register struct pvip_softc *softc)
{
	register ddi_acc_handle_t seq_handle = softc->pvip_vga_handle;
	register u_char	oldseq;
	register u_char	oldmisc;
	register u_char	oldclockmode;
	register int	status = 0;

	oldseq = ddi_getb(seq_handle, softc->viperport.vp_seq_index_port);

	ddi_putb(seq_handle, softc->viperport.vp_seq_index_port,
		VIPER_SEQ_CLOCK_INDEX);
	if ((ddi_getb(seq_handle, softc->viperport.vp_seq_index_port) &
			VIPER_SEQ_MASK) != VIPER_SEQ_CLOCK_INDEX) {
		DEBUGF(2, (CE_CONT,
			" setting clockmode seq failed\n"));
		goto failure;
	}

	oldclockmode =
		ddi_getb(seq_handle, softc->viperport.vp_seq_data_port);
	status |= PVIP_LOCK_DID_GET_CLOCK;

	/* select misc reg */

	ddi_putb(seq_handle,
		softc->viperport.vp_seq_index_port, VIPER_SEQ_MISC_INDEX);

	if ((ddi_getb(seq_handle, softc->viperport.vp_seq_index_port) &
			VIPER_SEQ_MASK) !=
			(VIPER_SEQ_MISC_INDEX & VIPER_SEQ_MASK)) {
		DEBUGF(2, (CE_CONT,
			" setting viper seq misc failed\n"));
		goto failure;
	}

	/* get the value */
	oldmisc = ddi_getb(seq_handle, softc->viperport.vp_seq_data_port);

	/* set bit 5 to lock extended regs */

	ddi_putb(seq_handle, softc->viperport.vp_seq_data_port,
		oldmisc | VIPER_MISC_CRLOCK);

	status |= PVIP_LOCK_SUCCESS;

failure:

	if (status & PVIP_LOCK_DID_GET_CLOCK) {
		ddi_putb(seq_handle, softc->viperport.vp_seq_index_port,
			VIPER_SEQ_CLOCK_INDEX);
		ddi_putb(seq_handle, softc->viperport.vp_seq_data_port,
			oldclockmode);
	}

	ddi_putb(seq_handle, softc->viperport.vp_seq_index_port, oldseq);

	return ((status & PVIP_LOCK_SUCCESS) != 0);
}


static void
viper_write_ic2061a(register struct pvip_softc *softc,
	register uint32_t clockval)
{
	register ddi_acc_handle_t seq_handle = softc->pvip_vga_handle;
	register uint32_t	data;
	register int	i;
	register u_char	oldstate;
	register u_int flags;

	flags = ddi_enter_critical();

	oldstate = ddi_getb(seq_handle, softc->viperport.vp_miscin) &
		~(VIPER_MISC_DATA | VIPER_MISC_CLOCK);

	/*
	 * the icd2061a requires a recovery time of at least 30ns
	 * (setup 20ns and hold time 10ns) between write accesses,
	 * and must be less than the timeout specification 2ms.  Hence we
	 * will sprinkle in calls to delay 1us.  In reality, the current
	 * implementation of drv_usecwait will delay 10us.
	 */

	/* first, send the "unlock sequence" to the clock chip */

	/* raise the data bit */

	ddi_putb(seq_handle,
		softc->viperport.vp_miscout, oldstate | VIPER_MISC_DATA);
	drv_usecwait(1);

	/* send at least 5 unlock bits */

	for (i = 0; i < 5; i++) {

		/* Hold the data on while lowering and rasing the clock */

		ddi_putb(seq_handle, softc->viperport.vp_miscout,
			oldstate | VIPER_MISC_DATA);

		drv_usecwait(1);

		ddi_putb(seq_handle, softc->viperport.vp_miscout,
			oldstate | VIPER_MISC_DATA | VIPER_MISC_CLOCK);

		drv_usecwait(1);
	}

	/* turn the data and clock off */

	ddi_putb(seq_handle, softc->viperport.vp_miscout, oldstate);
	drv_usecwait(1);

	/* turn the clock on one more time */

	ddi_putb(seq_handle,
		softc->viperport.vp_miscout, oldstate | VIPER_MISC_CLOCK);
	drv_usecwait(1);

	/* Now send start bit */

	/* leave data off, and lower the clock */

	ddi_putb(seq_handle, softc->viperport.vp_miscout, oldstate);
	drv_usecwait(1);

	/* leave data off, and raise the clock */

	ddi_putb(seq_handle,
		softc->viperport.vp_miscout, oldstate | VIPER_MISC_CLOCK);
	drv_usecwait(1);

	/* next, send the 24 data bits */

	for (i = 0, data = clockval << 3; i < 24; i++, data >>= 1) {

		/* leave the clock high, raise the inverse of the data bit. */

		ddi_putb(seq_handle, softc->viperport.vp_miscout,
			oldstate | VIPER_MISC_CLOCK |
			(~data & VIPER_MISC_DATA));
		drv_usecwait(1);

		/* leave the inverse data in place, lower the clock */

		ddi_putb(seq_handle, softc->viperport.vp_miscout,
			oldstate | (~data & VIPER_MISC_DATA));
		drv_usecwait(1);

		/* leave the clock low, raise the data bit */

		ddi_putb(seq_handle, softc->viperport.vp_miscout,
			oldstate | (data & VIPER_MISC_DATA));
		drv_usecwait(1);

		/* leave the data bit in place, raise the clock */

		ddi_putb(seq_handle, softc->viperport.vp_miscout,
			oldstate | VIPER_MISC_CLOCK |
			(data & VIPER_MISC_DATA));
		drv_usecwait(1);

		/* get the next bit of data */
	}

	/* leave the clock high, raise the data bit */

	ddi_putb(seq_handle, softc->viperport.vp_miscout,
		oldstate | VIPER_MISC_DATA | VIPER_MISC_CLOCK);
	drv_usecwait(1);

	/* leave the data high, clock the clock low, then high again */

	ddi_putb(seq_handle,
		softc->viperport.vp_miscout, oldstate | VIPER_MISC_DATA);
	drv_usecwait(1);

	ddi_putb(seq_handle, softc->viperport.vp_miscout,
		oldstate | VIPER_MISC_DATA | VIPER_MISC_CLOCK);
	drv_usecwait(1);

	/*
	 * select both serial control bits to 1 will
	 * select register 2 frequency.
	 */
	ddi_putb(seq_handle, softc->viperport.vp_miscout,
		oldstate | VIPER_MISC_DATA | VIPER_MISC_CLOCK);

	ddi_exit_critical(flags);

	/*
	 * clock settings become effective when the watchdog
	 * timer times out, 2ms.
	 */
	delay(drv_usectohz(2000));
}
